## Sid Meier's Civilization 4
## Copyright Firaxis Games 2005

from CvPythonExtensions import *
import CvUtil
import Popup as PyPopup
import PyHelpers

# globals
gc = CyGlobalContext()
PyPlayer = PyHelpers.PyPlayer

t4Units = [ 'UNITCLASS_HEAVY_CROSSBOWMAN','UNITCLASS_FLURRY','UNITCLASS_MARKSMAN','UNITCLASS_ARCHMAGE','UNITCLASS_SUMMONER','UNITCLASS_HIGH_PRIEST','UNITCLASS_INQUISITOR','UNITCLASS_MONK','UNITCLASS_PALADIN','UNITCLASS_WAR_ELEPHANT','UNITCLASS_CAMEL_ARCHER','UNITCLASS_KNIGHT','UNITCLASS_WAR_CHARIOT','UNITCLASS_SHIELD_WALL','UNITCLASS_IMMORTAL','UNITCLASS_SPARTIATOI','UNITCLASS_BERSERKER' ]

class CustomFunctions:

	def FFHAddBonus(self, iBonus, iNum, sIcon):
		listPlots = []
		for i in range (CyMap().numPlots()):
			pPlot = CyMap().plotByIndex(i)
			if (pPlot.canHaveBonus(CvUtil.getInfoTypeForStringCached(iBonus),True) and pPlot.getBonusType(-1) == -1 and pPlot.isCity() == False):
				listPlots.append(i)
		if len(listPlots) > 0:
			for i in range (iNum):
				iRnd = CyGame().getSorenRandNum(len(listPlots), "Add Bonus")
				pPlot = CyMap().plotByIndex(listPlots[iRnd])
				pPlot.setBonusType(CvUtil.getInfoTypeForStringCached(iBonus))
				if sIcon != -1:
					iActivePlayer = CyGame().getActivePlayer()
					CyInterface().addMessage(iActivePlayer,True,25,'Resource Discovered','AS2D_DISCOVERBONUS',1,sIcon,ColorTypes(8),pPlot.getX(),pPlot.getY(),True,True)

	def FFHAddImprovements(self):
		iBarrow = CvUtil.getInfoTypeForStringCached('IMPROVEMENT_BARROW')
		iBarrowChance = gc.getDefineINT('BARROW_CHANCE')
		iRuins = CvUtil.getInfoTypeForStringCached('IMPROVEMENT_RUINS')
		iRuinsChance = gc.getDefineINT('RUINS_CHANCE')
		iTower = CvUtil.getInfoTypeForStringCached('IMPROVEMENT_TOWER')
		iTowerChance = gc.getDefineINT('TOWER_CHANCE')
		for i in range (CyMap().numPlots()):
			if CyGame().getSorenRandNum(10000, "Add Barrows") <= iBarrowChance:
				pPlot = CyMap().plotByIndex(i)
				if (pPlot.canHaveImprovement(iBarrow,-1,True) and pPlot.getImprovementType() == -1 and pPlot.isCity() == False and pPlot.getFeatureType() == -1):
					pPlot.setImprovementType(iBarrow)
			if CyGame().getSorenRandNum(10000, "Add Ruins") <= iRuinsChance:
				pPlot = CyMap().plotByIndex(i)
				if (pPlot.canHaveImprovement(iRuins,-1,True) and pPlot.getImprovementType() == -1 and pPlot.isCity() == False):
					pPlot.setImprovementType(iRuins)
			if CyGame().getSorenRandNum(10000, "Add Tower") <= iTowerChance:
				pPlot = CyMap().plotByIndex(i)
				if (pPlot.canHaveImprovement(iTower,-1,True) and pPlot.getImprovementType() == -1 and pPlot.isCity() == False):
					pPlot.setImprovementType(iTower)

	def FFHAddPopup(self, szText, sDDS):
		szTitle = self.szGameDate = CyGameTextMgr().getTimeStr(CyGame().getGameTurn(), false)
		popup = PyPopup.PyPopup(-1)
		popup.addDDS(sDDS, 0, 0, 128, 384)
		popup.addSeparator()
		popup.setHeaderString(szTitle)
		popup.setBodyString(szText)
		popup.launch(true, PopupStates.POPUPSTATE_IMMEDIATE)

	def FFHAddUnit(self, iUnit):
		pBestPlot = -1
		iBestPlot = -1
		for i in range (CyMap().numPlots()):
			pPlot = CyMap().plotByIndex(i)
			iPlot = -1
			if pPlot.isWater() == False:
				if pPlot.getNumUnits() == 0:
					if pPlot.isCity() == False:
						if pPlot.isImpassable() == False:
							iPlot = CyGame().getSorenRandNum(500, "Add Unit")
							iPlot = iPlot + (pPlot.area().getNumTiles() * 10)
							if pPlot.isBarbarian():
								iPlot = iPlot + 200
			if iPlot > iBestPlot:
				iBestPlot = iPlot
				pBestPlot = pPlot
		if iBestPlot != -1:
			bPlayer = gc.getPlayer(gc.getBARBARIAN_PLAYER())
			newUnit = bPlayer.initUnit(iUnit, pBestPlot.getX(), pBestPlot.getY(), UnitAITypes.NO_UNITAI)
			if iUnit == CvUtil.getInfoTypeForStringCached('UNIT_ORTHUS'):
				newUnit.setHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_ORTHUSS_AXE'),True)

	def FFHApocalypse(self):
		iPercent = gc.getDefineINT('APOCALYPSE_KILL_CHANCE')
		for iPlayer in range(gc.getMAX_PLAYERS()):
			pPlayer = gc.getPlayer(iPlayer)
			if (pPlayer.isAlive()):
				if pPlayer.hasTrait(CvUtil.getInfoTypeForStringCached('TRAIT_FALLOW')) == False:
					for i in range(pPlayer.getNumCities()):
						pCity = pPlayer.getCity(i)
						iPop = pCity.getPopulation()
						iPop = int(iPop / 2)
						if iPop == 0:
							iPop = 1
						pCity.setPopulation(iPop)
					pyPlayer = PyPlayer(iPlayer)
					apUnitList = pyPlayer.getUnitList()
					for pUnit in apUnitList:
						if (CyGame().getSorenRandNum(100, "Apocalypse") <= iPercent):
							if pUnit.isAlive():
								pUnit.kill(False,0)

	def FFHBlight(self):
		for i in range (CyMap().numPlots()):
			pPlot = CyMap().plotByIndex(i)
			if (pPlot.getImprovementType() == CvUtil.getInfoTypeForStringCached('IMPROVEMENT_FARM') or pPlot.getImprovementType() == CvUtil.getInfoTypeForStringCached('IMPROVEMENT_PLANTATION') or pPlot.getImprovementType() == CvUtil.getInfoTypeForStringCached('IMPROVEMENT_PASTURE') or pPlot.getImprovementType() == CvUtil.getInfoTypeForStringCached('IMPROVEMENT_CAMP') or pPlot.getImprovementType() == CvUtil.getInfoTypeForStringCached('IMPROVEMENT_WINERY')):
				pPlot.setImprovementType(-1)
			if (pPlot.getBonusType(-1) == CvUtil.getInfoTypeForStringCached('BONUS_WHEAT') or pPlot.getBonusType(-1) == CvUtil.getInfoTypeForStringCached('BONUS_RICE') or pPlot.getBonusType(-1) == CvUtil.getInfoTypeForStringCached('BONUS_CORN') or pPlot.getBonusType(-1) == CvUtil.getInfoTypeForStringCached('BONUS_COW') or pPlot.getBonusType(-1) == CvUtil.getInfoTypeForStringCached('BONUS_DEER') or pPlot.getBonusType(-1) == CvUtil.getInfoTypeForStringCached('BONUS_PIG') or pPlot.getBonusType(-1) == CvUtil.getInfoTypeForStringCached('BONUS_BANANA') or pPlot.getBonusType(-1) == CvUtil.getInfoTypeForStringCached('BONUS_SHEEP') or pPlot.getBonusType(-1) == CvUtil.getInfoTypeForStringCached('BONUS_SUGAR') or pPlot.getBonusType(-1) == CvUtil.getInfoTypeForStringCached('BONUS_WINE')):
				iRand = CyGame().getSorenRandNum(100, "Blight")
				if (iRand <= 25):
					pPlot.setBonusType(-1)
			if pPlot.getFeatureType() == CvUtil.getInfoTypeForStringCached('FEATURE_ICE'):
				iRand = CyGame().getSorenRandNum(100, "Blight")
				if (iRand <= 50):
					pPlot.setFeatureType(-1, -1)
			if pPlot.getTerrainType() == CvUtil.getInfoTypeForStringCached('TERRAIN_PLAINS'):
				iRand = CyGame().getSorenRandNum(100, "Blight")
				if (iRand <= 25):
					pPlot.setTerrainType(CvUtil.getInfoTypeForStringCached('TERRAIN_DESERT'),True,True)

	def FFHChangeArmageddonCounter(self, iChange, iPlayer):
		iMaxArmCount = CyGame().getMaxArmageddonCounter()
		if iChange > 0:
			CyAudioGame().Play2DSound("AS2D_ARMAGEDDON_INCREASE")
		if iChange < 0:
			CyAudioGame().Play2DSound("AS2D_ARMAGEDDON_DECREASE")
		CyGame().changeArmageddonCounter(iChange, iPlayer)
		iArmCount = CyGame().getArmageddonCounter()
		if (iArmCount >= 1 and iMaxArmCount < 1):
			self.FFHAddPopup(CyTranslator().getText("TXT_KEY_POPUP_ARMAGEDDON_COUNTER",()), 'art/interface/popups/Abandon.dds')
		if (iArmCount >= 10 and iMaxArmCount < 10):
			self.FFHAddPopup(CyTranslator().getText("TXT_KEY_POPUP_ARMAGEDDON",()), 'art/interface/popups/Armageddon.dds')
		if (iArmCount >= 40 and iMaxArmCount < 40):
			self.FFHAddPopup(CyTranslator().getText("TXT_KEY_POPUP_BLIGHT",()), 'art/interface/popups/Blight.dds')
			self.FFHBlight()
		if (iArmCount >= 60 and iMaxArmCount < 60):
			self.FFHAddPopup(CyTranslator().getText("TXT_KEY_POPUP_STEPHANOS",()), 'art/interface/popups/Stephanos.dds')
			self.FFHAddUnit(CvUtil.getInfoTypeForStringCached('UNIT_STEPHANOS'))
		if (iArmCount >= 63 and iMaxArmCount < 63):
			self.FFHAddPopup(CyTranslator().getText("TXT_KEY_POPUP_BUBOES",()), 'art/interface/popups/Buboes.dds')
			self.FFHAddUnit(CvUtil.getInfoTypeForStringCached('UNIT_BUBOES'))
		if (iArmCount >= 66 and iMaxArmCount < 66):
			self.FFHAddPopup(CyTranslator().getText("TXT_KEY_POPUP_YERSINIA",()), 'art/interface/popups/Yersinia.dds')
			self.FFHAddUnit(CvUtil.getInfoTypeForStringCached('UNIT_YERSINIA'))
		if (iArmCount >= 69 and iMaxArmCount < 69):
			self.FFHAddPopup(CyTranslator().getText("TXT_KEY_POPUP_ARS",()), 'art/interface/popups/Ars Moriendi.dds')
			self.FFHAddUnit(CvUtil.getInfoTypeForStringCached('UNIT_ARS'))
		if (iArmCount >= 70 and iMaxArmCount < 70):
			self.FFHAddPopup(CyTranslator().getText("TXT_KEY_POPUP_HELLFIRE",()), 'art/interface/popups/Hellfire.dds')
			self.FFHHellfire()
		if (iArmCount >= 90 and iMaxArmCount < 90):
			self.FFHAddPopup(CyTranslator().getText("TXT_KEY_POPUP_WRATH",()), 'art/interface/popups/Avatar of Wrath.dds')
			self.FFHWrath()
		if (iArmCount >= 100 and iMaxArmCount < 100):
			self.FFHAddPopup(CyTranslator().getText("TXT_KEY_POPUP_APOCALYPSE",()), 'art/interface/popups/Apocalypse.dds')
			self.FFHApocalypse()

	def FFHConvert(self, pAttacker, pOldUnit, unitType, bPromotions):
		if pOldUnit.isHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_LOYALTY')):
			pOldUnit.kill(True,0)
		else:
			pPlayer = gc.getPlayer(pAttacker.getOwner())
			pPlot = pAttacker.plot()
			pClearPlot = self.FFHFindClearPlot(pOldUnit, -1)
			if pOldUnit.plot().getNumUnits() == 1 and pClearPlot != -1:
				pPlot = pOldUnit.plot()
				pOldUnit.setXY(pClearPlot.getX(), pClearPlot.getY())
			newUnit = pPlayer.initUnit(unitType, pPlot.getX(), pPlot.getY(), UnitAITypes.NO_UNITAI)
			iDmg = pOldUnit.getDamage()
			if iDmg > gc.getMAX_HIT_POINTS()*3/4:  # Smarter Orcs fix.
				iDmg = gc.getMAX_HIT_POINTS()*3/4  # Smarter Orcs fix.
			pOldUnit.setDamage(iDmg, False)
			if bPromotions == False:
				pOldUnit.setLevel(0)
				pOldUnit.setExperience(0,-1)
				for iCount in range(gc.getNumPromotionInfos()):
					pOldUnit.setHasPromotion(iCount, False)
			newUnit.convert(pOldUnit)
			newUnit.finishMoves()

	def FFHDoTurnKhazad(self, iPlayer):
		pPlayer = gc.getPlayer(iPlayer)
		if pPlayer.getNumCities() > 0:
			iVault1 = CvUtil.getInfoTypeForStringCached('BUILDING_DWARVEN_VAULT_EMPTY')
			iVault2 = CvUtil.getInfoTypeForStringCached('BUILDING_DWARVEN_VAULT_LOW')
			iVault3 = CvUtil.getInfoTypeForStringCached('BUILDING_DWARVEN_VAULT')
			iVault4 = CvUtil.getInfoTypeForStringCached('BUILDING_DWARVEN_VAULT_STOCKED')
			iVault5 = CvUtil.getInfoTypeForStringCached('BUILDING_DWARVEN_VAULT_ABUNDANT')
			iVault6 = CvUtil.getInfoTypeForStringCached('BUILDING_DWARVEN_VAULT_FULL')
			iVault7 = CvUtil.getInfoTypeForStringCached('BUILDING_DWARVEN_VAULT_OVERFLOWING')
			iGold = pPlayer.getGold() / pPlayer.getNumCities()
			if iGold <= 49:
				iNewVault = iVault1
			if (iGold >= 50 and iGold <= 99):
				iNewVault = iVault2
			if (iGold >= 100 and iGold <= 149):
				iNewVault = iVault3
			if (iGold >= 150 and iGold <= 199):
				iNewVault = iVault4
			if (iGold >= 200 and iGold <= 299):
				iNewVault = iVault5
			if (iGold >= 300 and iGold <= 499):
				iNewVault = iVault6
			if iGold >= 500:
				iNewVault = iVault7
			for i in range(pPlayer.getNumCities()):
				pCity = pPlayer.getCity(i)
				pCity.setHasRealBuilding(iVault1, False)
				pCity.setHasRealBuilding(iVault2, False)
				pCity.setHasRealBuilding(iVault3, False)
				pCity.setHasRealBuilding(iVault4, False)
				pCity.setHasRealBuilding(iVault5, False)
				pCity.setHasRealBuilding(iVault6, False)
				pCity.setHasRealBuilding(iVault7, False)
				pCity.setHasRealBuilding(iNewVault, True)

	def FFHDoTurnLuchuirp(self, iPlayer):
		pPlayer = gc.getPlayer(iPlayer)
		py = PyPlayer(iPlayer)
		pBarnaxus = -1
		iBarnaxus = CvUtil.getInfoTypeForStringCached('UNITCLASS_BARNAXUS')
		iCombat1 = CvUtil.getInfoTypeForStringCached('PROMOTION_COMBAT1')
		iCombat2 = CvUtil.getInfoTypeForStringCached('PROMOTION_COMBAT2')
		iCombat3 = CvUtil.getInfoTypeForStringCached('PROMOTION_COMBAT3')
		iCombat4 = CvUtil.getInfoTypeForStringCached('PROMOTION_COMBAT4')
		iCombat5 = CvUtil.getInfoTypeForStringCached('PROMOTION_COMBAT5')
		iEmpower1 = CvUtil.getInfoTypeForStringCached('PROMOTION_EMPOWER1')
		iEmpower2 = CvUtil.getInfoTypeForStringCached('PROMOTION_EMPOWER2')
		iEmpower3 = CvUtil.getInfoTypeForStringCached('PROMOTION_EMPOWER3')
		iEmpower4 = CvUtil.getInfoTypeForStringCached('PROMOTION_EMPOWER4')
		iEmpower5 = CvUtil.getInfoTypeForStringCached('PROMOTION_EMPOWER5')
		iGolem = CvUtil.getInfoTypeForStringCached('PROMOTION_GOLEM')
		if pPlayer.getUnitClassCount(iBarnaxus) > 0:
			for pUnit in py.getUnitList():
				if pUnit.getUnitClassType() == iBarnaxus:
					pBarnaxus = pUnit
		for pUnit in py.getUnitList():
			if pUnit.isHasPromotion(iGolem):
				pUnit.setHasPromotion(iEmpower1, False)
				pUnit.setHasPromotion(iEmpower2, False)
				pUnit.setHasPromotion(iEmpower3, False)
				pUnit.setHasPromotion(iEmpower4, False)
				pUnit.setHasPromotion(iEmpower5, False)
				if (pBarnaxus != -1 and pUnit.getUnitClassType() != iBarnaxus):
					if pBarnaxus.isHasPromotion(iCombat1):
						pUnit.setHasPromotion(iEmpower1, True)
					if pBarnaxus.isHasPromotion(iCombat2):
						pUnit.setHasPromotion(iEmpower2, True)
					if pBarnaxus.isHasPromotion(iCombat3):
						pUnit.setHasPromotion(iEmpower3, True)
					if pBarnaxus.isHasPromotion(iCombat4):
						pUnit.setHasPromotion(iEmpower4, True)
					if pBarnaxus.isHasPromotion(iCombat5):
						pUnit.setHasPromotion(iEmpower5, True)

	def FFHEventCarnivalEscape(self, pCity, plot):
		bPlayer = gc.getPlayer(gc.getBARBARIAN_PLAYER())
		iCommando = CvUtil.getInfoTypeForStringCached('PROMOTION_COMMANDO')
		if pCity.hasBuilding(CvUtil.getInfoTypeForStringCached('BUILDING_WOLF_PEN')):
			CyInterface().addMessage(pCity.getOwner(),True,25,'A Wolf has escaped your carnival!','AS2D_CITY_REVOLT',1,'Art/Interface/Buttons/Units/Wolf.dds',ColorTypes(8),pCity.getX(),pCity.getY(),True,True)
			newUnit = bPlayer.initUnit(CvUtil.getInfoTypeForStringCached('UNIT_WOLF'), plot.getX(), plot.getY(), UnitAITypes.NO_UNITAI)
			newUnit.setHasPromotion(iCommando, True)
			pCity.setHasRealBuilding(CvUtil.getInfoTypeForStringCached('BUILDING_WOLF_PEN'), False)
		if pCity.hasBuilding(CvUtil.getInfoTypeForStringCached('BUILDING_LION_CAGE')):
			CyInterface().addMessage(pCity.getOwner(),True,25,'A Lion has escaped your carnival!','AS2D_CITY_REVOLT',1,'Art/Interface/Buttons/Units/Lion.dds',ColorTypes(8),pCity.getX(),pCity.getY(),True,True)
			newUnit2 = bPlayer.initUnit(CvUtil.getInfoTypeForStringCached('UNIT_LION'), plot.getX(), plot.getY(), UnitAITypes.NO_UNITAI)
			newUnit2.setHasPromotion(iCommando, True)
			pCity.setHasRealBuilding(CvUtil.getInfoTypeForStringCached('BUILDING_LION_CAGE'), False)
		if pCity.hasBuilding(CvUtil.getInfoTypeForStringCached('BUILDING_TIGER_CAGE')):
			CyInterface().addMessage(pCity.getOwner(),True,25,'A Tiger has escaped your carnival!','AS2D_CITY_REVOLT',1,'Art/Interface/Buttons/Units/Tiger.dds',ColorTypes(8),pCity.getX(),pCity.getY(),True,True)
			newUnit4 = bPlayer.initUnit(CvUtil.getInfoTypeForStringCached('UNIT_TIGER'), plot.getX(), plot.getY(), UnitAITypes.NO_UNITAI)
			newUnit4.setHasPromotion(iCommando, True)
			pCity.setHasRealBuilding(CvUtil.getInfoTypeForStringCached('BUILDING_TIGER_CAGE'), False)
		if pCity.hasBuilding(CvUtil.getInfoTypeForStringCached('BUILDING_DANCING_BEAR')):
			CyInterface().addMessage(pCity.getOwner(),True,25,'A Dancing Bear has escaped!','AS2D_CITY_REVOLT',1,'Art/Interface/Buttons/Units/Bear.dds',ColorTypes(8),pCity.getX(),pCity.getY(),True,True)
			newUnit5 = bPlayer.initUnit(CvUtil.getInfoTypeForStringCached('UNIT_BEAR'), plot.getX(), plot.getY(), UnitAITypes.NO_UNITAI)
			newUnit5.setHasPromotion(iCommando, True)
			pCity.setHasRealBuilding(CvUtil.getInfoTypeForStringCached('BUILDING_DANCING_BEAR'), False)
		if pCity.hasBuilding(CvUtil.getInfoTypeForStringCached('BUILDING_SPIDER_PEN')):
			CyInterface().addMessage(pCity.getOwner(),True,25,'A Giant Spider has escaped!','AS2D_CITY_REVOLT',1,'Art/Interface/Buttons/Units/Giantspider.dds',ColorTypes(8),pCity.getX(),pCity.getY(),True,True)
			newUnit6 = bPlayer.initUnit(CvUtil.getInfoTypeForStringCached('UNIT_GIANT_SPIDER'), plot.getX(), plot.getY(), UnitAITypes.NO_UNITAI)
			newUnit6.setHasPromotion(iCommando, True)
			pCity.setHasRealBuilding(CvUtil.getInfoTypeForStringCached('BUILDING_SPIDER_PEN'), False)
		if pCity.hasBuilding(CvUtil.getInfoTypeForStringCached('BUILDING_GORILLA_CAGE')):
			CyInterface().addMessage(pCity.getOwner(),True,25,'A Gorilla has escaped!','AS2D_CITY_REVOLT',1,'Art/Interface/Buttons/Units/Gorilla.dds',ColorTypes(8),pCity.getX(),pCity.getY(),True,True)
			newUnit7 = bPlayer.initUnit(CvUtil.getInfoTypeForStringCached('UNIT_GORILLA'), plot.getX(), plot.getY(), UnitAITypes.NO_UNITAI)
			newUnit7.setHasPromotion(iCommando, True)
			pCity.setHasRealBuilding(CvUtil.getInfoTypeForStringCached('BUILDING_GORILLA_CAGE'), False)

	def FFHFindClearPlot(self, pUnit, plot):
		BestPlot = -1
		iBestPlot = 0
		if pUnit == -1:
			iX = plot.getX()
			iY = plot.getY()
			for iiX in range(iX-1, iX+2, 1):
				for iiY in range(iY-1, iY+2, 1):
					iCurrentPlot = 0
					pPlot = CyMap().plot(iiX,iiY)
					if pPlot.getNumUnits() == 0:
						if (pPlot.isWater() == plot.isWater() and pPlot.isPeak() == False and pPlot.isCity() == False):
							iCurrentPlot = iCurrentPlot + 5
					if iCurrentPlot >= 1:
						iCurrentPlot = iCurrentPlot + CyGame().getSorenRandNum(5, "FFHFindClearPlot")
						if iCurrentPlot >= iBestPlot:
							BestPlot = pPlot
							iBestPlot = iCurrentPlot
			return BestPlot
		iX = pUnit.getX()
		iY = pUnit.getY()
		for iiX in range(iX-1, iX+2, 1):
			for iiY in range(iY-1, iY+2, 1):
				iCurrentPlot = 0
				pPlot = CyMap().plot(iiX,iiY)
				if pPlot.getNumUnits() == 0:
					if pUnit.canMoveOrAttackInto(pPlot, False):
						iCurrentPlot = iCurrentPlot + 5
				for i in range(pPlot.getNumUnits()):
					if pPlot.getUnit(i).getOwner() == pUnit.getOwner():
						if pUnit.canMoveOrAttackInto(pPlot, False):
							iCurrentPlot = iCurrentPlot + 15
				if pPlot.isCity():
					if pPlot.getPlotCity().getOwner() == pUnit.getOwner():
						iCurrentPlot = iCurrentPlot + 50
				if (iX == iiX and iY == iiY):
					iCurrentPlot = 0
				if iCurrentPlot >= 1:
					iCurrentPlot = iCurrentPlot + CyGame().getSorenRandNum(5, "FFHFindClearPlot")
					if iCurrentPlot >= iBestPlot:
						BestPlot = pPlot
						iBestPlot = iCurrentPlot
		return BestPlot

	def FFHGenesis(self, pCity):
		iSnow = CvUtil.getInfoTypeForStringCached('TERRAIN_SNOW')
		iTundra = CvUtil.getInfoTypeForStringCached('TERRAIN_TUNDRA')
		iPlains = CvUtil.getInfoTypeForStringCached('TERRAIN_PLAINS')
		iDesert = CvUtil.getInfoTypeForStringCached('TERRAIN_DESERT')
		iGrass = CvUtil.getInfoTypeForStringCached('TERRAIN_GRASS')
		iOasis = CvUtil.getInfoTypeForStringCached('FEATURE_OASIS')
		iForestAncient = CvUtil.getInfoTypeForStringCached('FEATURE_FOREST_ANCIENT')
		iForest = CvUtil.getInfoTypeForStringCached('FEATURE_FOREST')
		for i in range (CyMap().numPlots()):
			pPlot = CyMap().plotByIndex(i)
			if pPlot.getOwner() == pCity.getOwner():
				if(pPlot.getTerrainType() == iSnow):
					pPlot.setTerrainType(iTundra,True,True)
				elif(pPlot.getTerrainType() == iTundra):
					pPlot.setTerrainType(iPlains,True,True)
				elif(pPlot.getTerrainType() == iDesert and pPlot.getFeatureType() != iOasis):
					pPlot.setTerrainType(iPlains,True,True)
				elif(pPlot.getTerrainType() == iPlains):
					pPlot.setTerrainType(iGrass,True,True)
				elif(pPlot.getTerrainType() == iGrass and pPlot.getImprovementType() == -1 and pPlot.getFeatureType() != iForestAncient and pPlot.isPeak() == False):
					pPlot.setFeatureType(iForest, 0)

	def FFHGetHero(self, pPlayer):
		iHero = -1
		if pPlayer.getCivilizationType() == CvUtil.getInfoTypeForStringCached('CIVILIZATION_BANNOR'):
			iHero = CvUtil.getInfoTypeForStringCached('UNITCLASS_DONAL')
		if pPlayer.getCivilizationType() == CvUtil.getInfoTypeForStringCached('CIVILIZATION_MALAKIM'):
			iHero = CvUtil.getInfoTypeForStringCached('UNITCLASS_CHALID')
		if pPlayer.getCivilizationType() == CvUtil.getInfoTypeForStringCached('CIVILIZATION_ELOHIM'):
			iHero = CvUtil.getInfoTypeForStringCached('UNITCLASS_CORLINDALE')
		if pPlayer.getCivilizationType() == CvUtil.getInfoTypeForStringCached('CIVILIZATION_MERCURIANS'):
			iHero = CvUtil.getInfoTypeForStringCached('UNITCLASS_BASIUM')
		if pPlayer.getCivilizationType() == CvUtil.getInfoTypeForStringCached('CIVILIZATION_LANUN'):
			iHero = CvUtil.getInfoTypeForStringCached('UNITCLASS_GUYBRUSH')
		if pPlayer.getCivilizationType() == CvUtil.getInfoTypeForStringCached('CIVILIZATION_KURIOTATES'):
			iHero = CvUtil.getInfoTypeForStringCached('UNITCLASS_EURABATRES')
		if pPlayer.getCivilizationType() == CvUtil.getInfoTypeForStringCached('CIVILIZATION_LJOSALFAR'):
			iHero = CvUtil.getInfoTypeForStringCached('UNITCLASS_GILDEN')
		if pPlayer.getCivilizationType() == CvUtil.getInfoTypeForStringCached('CIVILIZATION_KHAZAD'):
			iHero = CvUtil.getInfoTypeForStringCached('UNITCLASS_MAROS')
		if pPlayer.getCivilizationType() == CvUtil.getInfoTypeForStringCached('CIVILIZATION_SIDAR'):
			iHero = CvUtil.getInfoTypeForStringCached('UNITCLASS_RATHUS')
		if pPlayer.getCivilizationType() == CvUtil.getInfoTypeForStringCached('CIVILIZATION_HIPPUS'):
			iHero = CvUtil.getInfoTypeForStringCached('UNITCLASS_MAGNADINE')
		if pPlayer.getCivilizationType() == CvUtil.getInfoTypeForStringCached('CIVILIZATION_AMURITES'):
			iHero = CvUtil.getInfoTypeForStringCached('UNITCLASS_GOVANNON')
		if pPlayer.getCivilizationType() == CvUtil.getInfoTypeForStringCached('CIVILIZATION_BALSERAPHS'):
			iHero = CvUtil.getInfoTypeForStringCached('UNITCLASS_LOKI')
		if pPlayer.getCivilizationType() == CvUtil.getInfoTypeForStringCached('CIVILIZATION_CLAN_OF_EMBERS'):
			iHero = CvUtil.getInfoTypeForStringCached('UNITCLASS_RANTINE')
		if pPlayer.getCivilizationType() == CvUtil.getInfoTypeForStringCached('CIVILIZATION_SVARTALFAR'):
			iHero = CvUtil.getInfoTypeForStringCached('UNITCLASS_KAEL')
		if pPlayer.getCivilizationType() == CvUtil.getInfoTypeForStringCached('CIVILIZATION_CALABIM'):
			iHero = CvUtil.getInfoTypeForStringCached('UNITCLASS_LOSHA')
		if pPlayer.getCivilizationType() == CvUtil.getInfoTypeForStringCached('CIVILIZATION_SHEAIM'):
			iHero = CvUtil.getInfoTypeForStringCached('UNITCLASS_ABASHI')
		return iHero

	def FFHGetMaxSprawlingCities(self):
		i = gc.getDefineINT('SPRAWLING_NUMBER_OF_CITIES')
		if CyMap().getWorldSize() == CvUtil.getInfoTypeForStringCached('WORLDSIZE_DUEL'):
			i = i - 1
		if CyMap().getWorldSize() == CvUtil.getInfoTypeForStringCached('WORLDSIZE_LARGE'):
			i = i + 1
		if CyMap().getWorldSize() == CvUtil.getInfoTypeForStringCached('WORLDSIZE_HUGE'):
			i = i + 2
		return i

	def FFHGetOpenPlayer(self):
		i = -1
		for iPlayer in range(gc.getMAX_PLAYERS()):
			pPlayer = gc.getPlayer(iPlayer)
			if (pPlayer.isEverAlive() == False and i == -1):
				i = iPlayer
		return i

	def FFHGetUnitStrString(self, pUnit, bDetail):
		if (gc.getUnitInfo(pUnit.getUnitType()).getAirCombat() > 0):
			iStrength = pUnitTypeInfo.getAirCombat()
			szText = u"%i "%iStrength + u"%c" % CyGame().getSymbolID( FontSymbols.STRENGTH_CHAR )
			return szText
		iAttack = pUnit.baseCombatStr(true)
		iDefense = pUnit.baseCombatStr(false)
		if (pUnit.isFighting()):
			szText = u"?/%d" %(iAttack)
		else:
			if bDetail == True:
				iAttack = iAttack - pUnit.getCombatCold() - pUnit.getCombatDeath() - pUnit.getCombatFire() - pUnit.getCombatHoly() - pUnit.getCombatLightning() - pUnit.getCombatPoison() - pUnit.getCombatUnholy()
				iDefense = iDefense - pUnit.getCombatCold() - pUnit.getCombatDeath() - pUnit.getCombatFire() - pUnit.getCombatHoly() - pUnit.getCombatLightning() - pUnit.getCombatPoison() - pUnit.getCombatUnholy()
			if (pUnit.isHurt()):
				if iAttack == iDefense:
					szText = u"%.1f/%d" %(((float(iAttack * pUnit.currHitPoints())) / (float(pUnit.maxHitPoints()))), iAttack)
				else:
					szText = u"%.1f/%.1f" %(((float(iAttack * pUnit.currHitPoints())) / (float(pUnit.maxHitPoints()))), ((float(iDefense * pUnit.currHitPoints())) / (float(pUnit.maxHitPoints()))))
			else:
				if iAttack == iDefense:
					szText = u"%d" %(iAttack)
				else:
					szText = u"%d/%d" %(iAttack, iDefense)
			if bDetail == True:
				if pUnit.getCombatCold() != 0:
					szText += u" (+%i %s)" %(pUnit.getCombatCold(),CyTranslator().getText("TXT_KEY_DAMAGE_COLD_TAG",()))
				if pUnit.getCombatDeath() != 0:
					szText += u" (+%i %s)" %(pUnit.getCombatDeath(),CyTranslator().getText("TXT_KEY_DAMAGE_DEATH_TAG",()))
				if pUnit.getCombatFire() != 0:
					szText += u" (+%i %s)" %(pUnit.getCombatFire(),CyTranslator().getText("TXT_KEY_DAMAGE_FIRE_TAG",()))
				if pUnit.getCombatHoly() != 0:
					szText += u" (+%i %s)" %(pUnit.getCombatHoly(),CyTranslator().getText("TXT_KEY_DAMAGE_HOLY_TAG",()))
				if pUnit.getCombatLightning() != 0:
					szText += u" (+%i %s)" %(pUnit.getCombatLightning(),CyTranslator().getText("TXT_KEY_DAMAGE_LIGHTNING_TAG",()))
				if pUnit.getCombatPoison() != 0:
					szText += u" (+%i %s)" %(pUnit.getCombatPoison(),CyTranslator().getText("TXT_KEY_DAMAGE_POISON_TAG",()))
				if pUnit.getCombatUnholy() != 0:
					szText += u" (+%i %s)" %(pUnit.getCombatUnholy(),CyTranslator().getText("TXT_KEY_DAMAGE_UNHOLY_TAG",()))
		szText += u"%c" %(CyGame().getSymbolID(FontSymbols.STRENGTH_CHAR))
		return szText

	def FFHGiveExperience(self, caster):
		i = -1
		pPlayer = gc.getPlayer(caster.getOwner())
		if caster.isHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_CHANNELING1')):
			i = 25 - caster.getExperience()
			if pPlayer.hasTrait(CvUtil.getInfoTypeForStringCached('TRAIT_ARCANE')):
				i = i + 5
			if i <= 0:
				i = 1
		if caster.isHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_CHANNELING2')):
			i = 35 - (caster.getExperience() / 2)
			if pPlayer.hasTrait(CvUtil.getInfoTypeForStringCached('TRAIT_ARCANE')):
				i = i + 10
			if i <= 1:
				i = 2
		if caster.isHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_CHANNELING3')):
			i = 40 - (caster.getExperience() / 3)
			if pPlayer.hasTrait(CvUtil.getInfoTypeForStringCached('TRAIT_ARCANE')):
				i = i + 15
			if i <= 2:
				i = 3
		if caster.isHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_UNHOLY_TAINT')):
			i = i + 10
		iRnd = CyGame().getSorenRandNum(gc.getDefineINT('CASTER_XP_PROBABILITY'), "Bob")
		if iRnd <= i:
			caster.changeExperience(1,-1)

	def FFHGiftUnit(self, iUnit, iCivilization, iXP, pFromPlot, iFromPlayer):
		iAngel = CvUtil.getInfoTypeForStringCached('UNIT_ANGEL')
		iManes = CvUtil.getInfoTypeForStringCached('UNIT_MANES')
		if (iUnit == iAngel or iUnit == iManes):
			iChance = 100 - (CyGame().countCivPlayersAlive() * 3)
			iChance = iChance + iXP
			if iChance < 5:
				iChance = 5
			if iChance > 95:
				iChance = 95
			if CyGame().getSorenRandNum(100, "Gift Unit") > iChance:
				iUnit = -1
		if iUnit != -1:
			for iPlayer in range(gc.getMAX_PLAYERS()):
				pPlayer = gc.getPlayer(iPlayer)
				if (pPlayer.isAlive()):
					if pPlayer.getCivilizationType() == iCivilization:
						py = PyPlayer(iPlayer)
						if pPlayer.getNumCities() > 0:
							iRnd = CyGame().getSorenRandNum(py.getNumCities(), "Gift Unit")
							pCity = py.getCityList()[iRnd]
							pPlot = pCity.plot()
							newUnit = pPlayer.initUnit(iUnit, pPlot.getX(), pPlot.getY(), UnitAITypes.NO_UNITAI)
							newUnit.changeExperience(iXP, -1)
							if (pFromPlot != -1 and gc.getPlayer(iFromPlayer).isHuman()):
								if iUnit == iManes:
									CyInterface().addMessage(iFromPlayer,True,25,CyTranslator().getText("TXT_KEY_MESSAGE_UNIT_FALLS",()),'AS2D_UNIT_FALLS',1,'Art/Interface/Buttons/Promotions/Demon.dds',ColorTypes(7),pFromPlot.getX(),pFromPlot.getY(),True,True)
								if iUnit == iAngel:
									CyInterface().addMessage(iFromPlayer,True,25,CyTranslator().getText("TXT_KEY_MESSAGE_UNIT_RISES",()),'AS2D_UNIT_FALLS',1,'Art/Interface/Buttons/Promotions/Angel.dds',ColorTypes(7),pFromPlot.getX(),pFromPlot.getY(),True,True)
							if pPlayer.isHuman():
								if iUnit == iManes:
									CyInterface().addMessage(iPlayer,True,25,CyTranslator().getText("TXT_KEY_MESSAGE_ADD_MANES",()),'AS2D_UNIT_FALLS',1,'Art/Interface/Buttons/Promotions/Demon.dds',ColorTypes(7),pPlot.getX(),pPlot.getY(),True,True)
								if iUnit == iAngel:
									CyInterface().addMessage(iPlayer,True,25,CyTranslator().getText("TXT_KEY_MESSAGE_ADD_ANGEL",()),'AS2D_UNIT_FALLS',1,'Art/Interface/Buttons/Promotions/Angel.dds',ColorTypes(7),pPlot.getX(),pPlot.getY(),True,True)
							if (pPlayer.isHuman() == False and iUnit == iManes and pCity != -1):
								if CyGame().getSorenRandNum(100, "Manes") < (100 - (pCity.getPopulation() * 5)):
									pCity.changePopulation(1)
									newUnit.kill(True,0)

	def FFHHasRace(self, pUnit):
		if pUnit.isHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_ANGEL')):
			return True
		if pUnit.isHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_DEMON')):
			return True
		if pUnit.isHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_DRAGON')):
			return True
		if pUnit.isHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_DWARF')):
			return True
		if pUnit.isHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_ELF')):
			return True
		if pUnit.isHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_ELEMENTAL')):
			return True
		if pUnit.isHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_GOLEM')):
			return True
		if pUnit.isHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_ORC')):
			return True
		if pUnit.isHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_UNDEAD')):
			return True
		return False

	def FFHHellfire(self):
		iHellfire = CvUtil.getInfoTypeForStringCached('IMPROVEMENT_HELLFIRE')
		iHellfireChance = gc.getDefineINT('HELLFIRE_CHANCE')
		pPlayer = gc.getPlayer(gc.getBARBARIAN_PLAYER())
		for iPlayer2 in range(gc.getMAX_PLAYERS()):
			pPlayer2 = gc.getPlayer(iPlayer2)
			if (pPlayer2.isAlive()):
				if pPlayer2.getCivilizationType() == CvUtil.getInfoTypeForStringCached('CIVILIZATION_INFERNAL'):
					pPlayer = pPlayer2
		for i in range (CyMap().numPlots()):
			if CyGame().getSorenRandNum(10000, "Hellfire") <= iHellfireChance:
				pPlot = CyMap().plotByIndex(i)
				if (pPlot.getBonusType(-1) == -1 and pPlot.isCity() == False and pPlot.isFlatlands() and pPlot.isWater() == False and pPlot.getNumUnits() == 0):
					pPlot.setImprovementType(iHellfire)
					newUnit = pPlayer.initUnit(CvUtil.getInfoTypeForStringCached('UNIT_SECT_OF_FLIES'), pPlot.getX(), pPlot.getY(), UnitAITypes.NO_UNITAI)

	def FFHImmortal(self, pUnit):
		if pUnit.getOwner() != gc.getBARBARIAN_PLAYER():
			pPlayer = gc.getPlayer(pUnit.getOwner())
			if pPlayer.getNumCities() > 0:
				pUnit.setHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_IMMORTAL'), false)
				pCity = pPlayer.getCapitalCity()
				pUnit.setDamage(gc.getMAX_HIT_POINTS()/2, False) # Smarter Orcs defect fix.
				newUnit = pPlayer.initUnit(pUnit.getUnitType(), pCity.getX(), pCity.getY(), UnitAITypes.NO_UNITAI)
				newUnit.convert(pUnit)
				newUnit.finishMoves()

	def FFHInsane(self, iPlayer):
		pPlayer = gc.getPlayer(iPlayer)
		for i in range(gc.getNumTraitInfos()):
			if (pPlayer.hasTrait(i) and i != CvUtil.getInfoTypeForStringCached('TRAIT_INSANE')):
				CyMessageControl().sendApplyEvent(5012, EventContextTypes.EVENTCONTEXT_ALL, (iPlayer,i,False))
		Traits = [ 'TRAIT_AGGRESSIVE','TRAIT_ARCANE','TRAIT_CREATIVE','TRAIT_DEFENDER','TRAIT_EXPANSIVE','TRAIT_FINANCIAL','TRAIT_INDUSTRIOUS','TRAIT_ORGANIZED','TRAIT_MAGIC_RESISTANT','TRAIT_PHILOSOPHICAL','TRAIT_RAIDERS','TRAIT_SPIRITUAL' ]
		iRnd1 = CyGame().getSorenRandNum(len(Traits), "Insane")
		iRnd2 = CyGame().getSorenRandNum(len(Traits), "Insane")
		while iRnd2 == iRnd1:
			iRnd2 = CyGame().getSorenRandNum(len(Traits), "Insane")
		iRnd3 = CyGame().getSorenRandNum(len(Traits), "Insane")
		while iRnd3 == iRnd1 or iRnd3 == iRnd2:
			iRnd3 = CyGame().getSorenRandNum(len(Traits), "Insane")
		CyMessageControl().sendApplyEvent(5012, EventContextTypes.EVENTCONTEXT_ALL, (iPlayer,CvUtil.getInfoTypeForStringCached(Traits[iRnd1]),True))
		CyMessageControl().sendApplyEvent(5012, EventContextTypes.EVENTCONTEXT_ALL, (iPlayer,CvUtil.getInfoTypeForStringCached(Traits[iRnd2]),True))
		CyMessageControl().sendApplyEvent(5012, EventContextTypes.EVENTCONTEXT_ALL, (iPlayer,CvUtil.getInfoTypeForStringCached(Traits[iRnd3]),True))
		if pPlayer.isHuman() and gc.getGame().getGameTurnYear() >= 5:
			popup = PyPopup.PyPopup(-1)
			popup.addDDS('art/interface/popups/Perpentach.dds', 0, 0, 128, 384)
			popup.addSeparator()
			popup.setHeaderString(CyTranslator().getText("TXT_KEY_POPUP_INSANE_HEADER",()))
			popup.setBodyString(CyTranslator().getText("TXT_KEY_POPUP_INSANE",()))
			popup.launch(true, PopupStates.POPUPSTATE_QUEUED)

	def FFHMutate(self, pUnit):
		if pUnit.isHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_MUTATED')) == False:
			pUnit.setHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_MUTATED'),True)
			iRnd = CyGame().getSorenRandNum(100, "Mutation")
			if iRnd <= 15:
				pUnit.changeExtraCombatStr(-1)
			if iRnd >= 60:
				pUnit.changeExtraCombatStr(1)
			if iRnd >= 85:
				pUnit.changeExtraCombatStr(1)
			if CyGame().getSorenRandNum(100, "Mutation") <= 15:
				pUnit.setHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_ENRAGED'),True)
			if CyGame().getSorenRandNum(100, "Mutation") <= 15:
				pUnit.setHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_CRAZED'),True)
			if CyGame().getSorenRandNum(100, "Mutation") <= 10:
				pUnit.setHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_DISEASED'),True)
			if CyGame().getSorenRandNum(100, "Mutation") <= 10:
				pUnit.setHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_VULNERABLE_TO_FIRE'),True)
			if CyGame().getSorenRandNum(100, "Mutation") <= 5:
				pUnit.setHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_MAGIC_RESISTANCE'),True)
			if CyGame().getSorenRandNum(100, "Mutation") <= 5:
				pUnit.setHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_REGENERATION'),True)
			if CyGame().getSorenRandNum(100, "Mutation") <= 5:
				pUnit.setHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_WITHERED'),True)

	def FFHNewCity(self, pCity):
		iPlayer = pCity.getOwner()
		pPlayer = gc.getPlayer(iPlayer)
		if pPlayer.isSprawling():
			if pPlayer.getBuildingClassCount(CvUtil.getInfoTypeForStringCached("BUILDINGCLASS_CITY")) >= self.FFHGetMaxSprawlingCities():
				pCity.setHasRealBuilding(CvUtil.getInfoTypeForStringCached('BUILDING_SETTLEMENT'), True)
			else:
				pCity.setHasRealBuilding(CvUtil.getInfoTypeForStringCached('BUILDING_CITY'), True)
				if (pPlayer.isHuman() and pPlayer.getBuildingClassCount(CvUtil.getInfoTypeForStringCached("BUILDINGCLASS_CITY")) > 1 and CyGame().isNetworkMultiPlayer() == False):
					popupInfo = CyPopupInfo()
					popupInfo.setButtonPopupType(ButtonPopupTypes.BUTTONPOPUP_PYTHON)
					popupInfo.setData1(iPlayer)
					popupInfo.setData2(pCity.getID())
					popupInfo.setText(CyTranslator().getText("TXT_KEY_POPUP_SETTLEMENT_OR_CITY",()))
					popupInfo.setOnClickedPythonCallback("selectSettlement")
					popupInfo.addPythonButton(CyTranslator().getText("TXT_KEY_BUILDING_CITY", ()), "")
					popupInfo.addPythonButton(CyTranslator().getText("TXT_KEY_BUILDING_SETTLEMENT", ()), "")
					popupInfo.addPopup(iPlayer)

	def FFHPromoteAll(self, iPromotion, iPlayer):
		player = PyPlayer(iPlayer) 
		apUnitList = player.getUnitList() 
		for pUnit in apUnitList:
			if not pUnit.isHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_SUMMONED')):
				pUnit.setHasPromotion(iPromotion, True)

	def FFHSelectTrait(self, iPlayer):
		pPlayer = gc.getPlayer(iPlayer)
		iAgnostic = CvUtil.getInfoTypeForStringCached('TRAIT_AGNOSTIC')
		iSprawling = CvUtil.getInfoTypeForStringCached('TRAIT_SPRAWLING')
		iAdaptive = CvUtil.getInfoTypeForStringCached('TRAIT_ADAPTIVE')
		if pPlayer.getLeaderType() == CvUtil.getInfoTypeForStringCached('LEADER_VARN'):
			iPermTrait = CvUtil.getInfoTypeForStringCached('TRAIT_CREATIVE')
		if pPlayer.getLeaderType() == CvUtil.getInfoTypeForStringCached('LEADER_CARDITH'):
			iPermTrait = CvUtil.getInfoTypeForStringCached('TRAIT_EXPANSIVE')
		if pPlayer.getLeaderType() == CvUtil.getInfoTypeForStringCached('LEADER_CASSIEL'):
			iPermTrait = CvUtil.getInfoTypeForStringCached('TRAIT_PHILOSOPHICAL')
		for i in range(gc.getNumTraitInfos()):
			if pPlayer.hasTrait(i):
				if (i != iPermTrait and i != iSprawling and i != iAgnostic and i != iAdaptive):
					CyMessageControl().sendApplyEvent(5012, EventContextTypes.EVENTCONTEXT_ALL, (iPlayer,i,False))
	
		popupInfo = CyPopupInfo()
		popupInfo.setButtonPopupType(ButtonPopupTypes.BUTTONPOPUP_PYTHON)
		popupInfo.setText(CyTranslator().getText("TXT_KEY_POPUP_SELECT_TRAIT",()))
		popupInfo.setData1(iPlayer)
		popupInfo.setOnClickedPythonCallback("selectTrait")

		szText = CyTranslator().getText("TXT_KEY_TRAIT_AGGRESSIVE", ())
		if iPermTrait == CvUtil.getInfoTypeForStringCached('TRAIT_AGGRESSIVE'):
			szText = szText + CyTranslator().getText("TXT_KEY_TRAIT_PERMANENT", ())
		popupInfo.addPythonButton(szText, "")
		szText = CyTranslator().getText("TXT_KEY_TRAIT_ARCANE", ())
		if iPermTrait == CvUtil.getInfoTypeForStringCached('TRAIT_ARCANE'):
			szText = szText + CyTranslator().getText("TXT_KEY_TRAIT_PERMANENT", ())
		popupInfo.addPythonButton(szText, "")
		szText = CyTranslator().getText("TXT_KEY_TRAIT_CREATIVE", ())
		if iPermTrait == CvUtil.getInfoTypeForStringCached('TRAIT_CREATIVE'):
			szText = szText + CyTranslator().getText("TXT_KEY_TRAIT_PERMANENT", ())
		popupInfo.addPythonButton(szText, "")
		szText = CyTranslator().getText("TXT_KEY_TRAIT_DEFENDER", ())
		if iPermTrait == CvUtil.getInfoTypeForStringCached('TRAIT_DEFENDER'):
			szText = szText + CyTranslator().getText("TXT_KEY_TRAIT_PERMANENT", ())
		popupInfo.addPythonButton(szText, "")
		szText = CyTranslator().getText("TXT_KEY_TRAIT_EXPANSIVE", ())
		if iPermTrait == CvUtil.getInfoTypeForStringCached('TRAIT_EXPANSIVE'):
			szText = szText + CyTranslator().getText("TXT_KEY_TRAIT_PERMANENT", ())
		popupInfo.addPythonButton(szText, "")
		szText = CyTranslator().getText("TXT_KEY_TRAIT_FINANCIAL", ())
		if iPermTrait == CvUtil.getInfoTypeForStringCached('TRAIT_FINANCIAL'):
			szText = szText + CyTranslator().getText("TXT_KEY_TRAIT_PERMANENT", ())
		popupInfo.addPythonButton(szText, "")
		szText = CyTranslator().getText("TXT_KEY_TRAIT_INDUSTRIOUS", ())
		if iPermTrait == CvUtil.getInfoTypeForStringCached('TRAIT_INDUSTRIOUS'):
			szText = szText + CyTranslator().getText("TXT_KEY_TRAIT_PERMANENT", ())
		popupInfo.addPythonButton(szText, "")
		szText = CyTranslator().getText("TXT_KEY_TRAIT_ORGANIZED", ())
		if iPermTrait == CvUtil.getInfoTypeForStringCached('TRAIT_ORGANIZED'):
			szText = szText + CyTranslator().getText("TXT_KEY_TRAIT_PERMANENT", ())
		popupInfo.addPythonButton(szText, "")
		szText = CyTranslator().getText("TXT_KEY_TRAIT_MAGIC_RESISTANT", ())
		if iPermTrait == CvUtil.getInfoTypeForStringCached('TRAIT_MAGIC_RESISTANT'):
			szText = szText + CyTranslator().getText("TXT_KEY_TRAIT_PERMANENT", ())
		popupInfo.addPythonButton(szText, "")
		szText = CyTranslator().getText("TXT_KEY_TRAIT_PHILOSOPHICAL", ())
		if iPermTrait == CvUtil.getInfoTypeForStringCached('TRAIT_PHILOSOPHICAL'):
			szText = szText + CyTranslator().getText("TXT_KEY_TRAIT_PERMANENT", ())
		popupInfo.addPythonButton(szText, "")
		szText = CyTranslator().getText("TXT_KEY_TRAIT_RAIDERS", ())
		if iPermTrait == CvUtil.getInfoTypeForStringCached('TRAIT_RAIDERS'):
			szText = szText + CyTranslator().getText("TXT_KEY_TRAIT_PERMANENT", ())
		popupInfo.addPythonButton(szText, "")
		szText = CyTranslator().getText("TXT_KEY_TRAIT_SPIRITUAL", ())
		if iPermTrait == CvUtil.getInfoTypeForStringCached('TRAIT_SPIRITUAL'):
			szText = szText + CyTranslator().getText("TXT_KEY_TRAIT_PERMANENT", ())
		popupInfo.addPythonButton(szText, "")
		popupInfo.addPopup(iPlayer)

	def FFHStartWar(self, iPlayer, i2Player): #Smarter Orcs - plan ahead!
		iTeam = gc.getPlayer(iPlayer).getTeam()
		i2Team = gc.getPlayer(i2Player).getTeam()
		eTeam = gc.getTeam(iTeam)
		if (eTeam.isAtWar(i2Team) == False and iTeam != i2Team and eTeam.isHasMet(i2Team)):
			eTeam.AI_setWarPlan(i2Team, WarPlanTypes.WARPLAN_TOTAL) #Smarter Orcs - plan ahead!

	def FFHT4Units(self, pUnit):
		for i in t4Units:
			if pUnit.getUnitClassType() == CvUtil.getInfoTypeForStringCached(i):
				return True
		return False

	def FFHWarScript(self, iPlayer):
		pPlayer = gc.getPlayer(iPlayer)
		iEnemy = -1
		for iPlayer2 in range(gc.getMAX_PLAYERS()):
			pPlayer2 = gc.getPlayer(iPlayer2)
			if pPlayer2.isAlive():
				iTeam = gc.getPlayer(iPlayer).getTeam()
				iTeam2 = gc.getPlayer(iPlayer2).getTeam()
				eTeam = gc.getTeam(iTeam)
				if eTeam.isAtWar(iTeam2):
					if CyGame().getSorenRandNum(100, "War Script") < 5:
						self.FFHDogpile(iPlayer, iPlayer2)
				if self.FFHWarScriptAllow(iPlayer, iPlayer2):
					if pPlayer2.getBuildingClassMaking(CvUtil.getInfoTypeForStringCached('BUILDINGCLASS_TOWER_OF_MASTERY')) > 0:
						if eTeam.getAtWarCount(True) == 0:
							self.FFHStartWar(iPlayer, iPlayer2)
					if (pPlayer2.getBuildingClassMaking(CvUtil.getInfoTypeForStringCached('BUILDINGCLASS_ALTAR_OF_THE_LUONNOTAR_DIVINE')) > 0 or pPlayer2.getBuildingClassMaking(CvUtil.getInfoTypeForStringCached('BUILDINGCLASS_ALTAR_OF_THE_LUONNOTAR_EXALTED')) > 0):
						if pPlayer.getAlignment() == -1:
							if eTeam.getAtWarCount(True) == 0:
								self.FFHStartWar(iPlayer, iPlayer2)
					if pPlayer.getCivilizationType() == CvUtil.getInfoTypeForStringCached('CIVILIZATION_MERCURIANS'):
						if pPlayer2.getStateReligion() == CvUtil.getInfoTypeForStringCached('RELIGION_THE_ASHEN_VEIL'):
							self.FFHStartWar(iPlayer, iPlayer2)
					if CyGame().getArmageddonCounter() > 20:
						if pPlayer.getCivilizationType() == CvUtil.getInfoTypeForStringCached('CIVILIZATION_SVARTALFAR'):
							if (pPlayer2.getCivilizationType() == CvUtil.getInfoTypeForStringCached('CIVILIZATION_LJOSALFAR') and CyGame().getPlayerRank(iPlayer) > CyGame().getPlayerRank(iPlayer2)):
								self.FFHStartWar(iPlayer, iPlayer2)
						if pPlayer.getCivilizationType() == CvUtil.getInfoTypeForStringCached('CIVILIZATION_LJOSALFAR'):
							if (pPlayer2.getCivilizationType() == CvUtil.getInfoTypeForStringCached('CIVILIZATION_SVARTALFAR') and CyGame().getPlayerRank(iPlayer) > CyGame().getPlayerRank(iPlayer2)):
								self.FFHStartWar(iPlayer, iPlayer2)
					if (CyGame().getArmageddonCounter() > 40 or pPlayer.getCivilizationType() == CvUtil.getInfoTypeForStringCached('CIVILIZATION_INFERNAL') or pPlayer.getCivilizationType() == CvUtil.getInfoTypeForStringCached('CIVILIZATION_DOVIELLO')):
						if pPlayer.getAlignment() == -1:
							if (eTeam.getAtWarCount(True) == 0 and CyGame().getPlayerRank(iPlayer2) > CyGame().getPlayerRank(iPlayer)):
								if (iEnemy == -1 or CyGame().getPlayerRank(iPlayer2) > CyGame().getPlayerRank(iEnemy)):
									iEnemy = iPlayer2
		if iEnemy != -1:
			if CyGame().getPlayerRank(iPlayer) > CyGame().getPlayerRank(iEnemy):
				self.FFHStartWar(iPlayer, iEnemy)

	def FFHWarScriptAllow(self, iPlayer, iPlayer2):
		pPlayer = gc.getPlayer(iPlayer)
		pPlayer2 = gc.getPlayer(iPlayer2)
		iTeam = gc.getPlayer(iPlayer).getTeam()
		iTeam2 = gc.getPlayer(iPlayer2).getTeam()
		eTeam = gc.getTeam(iTeam)
		if iPlayer == gc.getBARBARIAN_PLAYER():
			return False
		if eTeam.isHasMet(iTeam2) == False:
			return False
		if eTeam.AI_getAtPeaceCounter(iTeam2) < 20:
			return False
#		if pPlayer.AI_getAttitude(iPlayer2) <= gc.getLeaderHeadInfo(pPlayer.getLeaderType()).getDeclareWarRefuseAttitudeThreshold():
#			return False
		if eTeam.isAtWar(iTeam2):
			return False
		if pPlayer.getCivilizationType() == CvUtil.getInfoTypeForStringCached('CIVILIZATION_INFERNAL'):
			if pPlayer2.getStateReligion() == CvUtil.getInfoTypeForStringCached('RELIGION_THE_ASHEN_VEIL'):
				return False
		return True

	def FFHDogpile(self, iPlayer, iVictim):
		pPlayer = gc.getPlayer(iPlayer)
		for iPlayer2 in range(gc.getMAX_PLAYERS()):
			pPlayer2 = gc.getPlayer(iPlayer2)
			iChance = -1
			if pPlayer2.isAlive():
				if (self.FFHDogPileAllow(iPlayer, iPlayer2) and self.FFHWarScriptAllow(iPlayer2, iVictim)):
					iChance = pPlayer2.AI_getAttitude(iPlayer) * 5
					if iChance > 0:
						iChance = iChance - (pPlayer2.AI_getAttitude(iVictim) * 5) - 10
						if CyGame().isOption(CvUtil.getInfoTypeForStringCached('GAMEOPTION_AGGRESSIVE_AI')) == False:
							iChance = iChance - 10
						if iChance > 0:
							iChance = iChance + (CyGame().getArmageddonCounter() / 3)
							if pPlayer2.getCivilizationType() == CvUtil.getInfoTypeForStringCached('CIVILIZATION_BALSERAPHS'):
								iChance = CyGame().getSorenRandNum(50, "Dogpile")
							if CyGame().getSorenRandNum(100, "Dogpile") < iChance:
								self.FFHStartWar(iPlayer2, iVictim)

	def FFHDogPileAllow(self, iPlayer, iPlayer2):
		pPlayer = gc.getPlayer(iPlayer)
		pPlayer2 = gc.getPlayer(iPlayer2)
		iTeam = gc.getPlayer(iPlayer).getTeam()
		iTeam2 = gc.getPlayer(iPlayer2).getTeam()
		if iPlayer == iPlayer2:
			return False
		if iTeam == iTeam2:
			return False
		if pPlayer2.isHuman():
			return False
		if pPlayer2.getCivilizationType() == CvUtil.getInfoTypeForStringCached('CIVILIZATION_ELOHIM'):
			return False
		return True

	def FFHWarn(self, iPlayer, szText, pPlot):
		pPlayer = gc.getPlayer(iPlayer)
		for iPlayer2 in range(gc.getMAX_PLAYERS()):
			pPlayer2 = gc.getPlayer(iPlayer2)
			if (pPlayer2.isAlive() and iPlayer != iPlayer2):
				if pPlayer2.isHuman():
					popupInfo = CyPopupInfo()
					popupInfo.setButtonPopupType(ButtonPopupTypes.BUTTONPOPUP_PYTHON)
					popupInfo.setText(szText)
					popupInfo.setOnClickedPythonCallback("selectWarn")
					popupInfo.addPythonButton(CyTranslator().getText("TXT_KEY_MAIN_MENU_OK",()), "")
					popupInfo.addPopup(iPlayer2)
				if pPlot != -1:
					CyInterface().addMessage(iPlayer2,True,25,CyTranslator().getText("TXT_KEY_MESSAGE_ALTAR_OF_THE_LUONNOTAR",()),'AS2D_DISCOVERBONUS',1,'Art/Interface/Buttons/Buildings/AltaroftheLuonnotar.dds',ColorTypes(7),pPlot.getX(),pPlot.getY(),True,True)

	def FFHWrath(self):
		self.FFHAddUnit(CvUtil.getInfoTypeForStringCached('UNIT_WRATH'))
		bPlayer = gc.getPlayer(gc.getBARBARIAN_PLAYER())
		py = PyPlayer(gc.getBARBARIAN_PLAYER())
		for pUnit in py.getUnitList():
			if pUnit.getUnitType() == CvUtil.getInfoTypeForStringCached('UNIT_WRATH'):
				pPlot = pUnit.plot()
		for iPlayer in range(gc.getMAX_PLAYERS()):
			pPlayer = gc.getPlayer(iPlayer)
			if (pPlayer.isAlive() and iPlayer != gc.getBARBARIAN_PLAYER()):
				py = PyPlayer(iPlayer)
				for pUnit in py.getUnitList():
					iRand = CyGame().getSorenRandNum(100, "Bob")
					if (iRand <= gc.getDefineINT('WRATH_CONVERT_CHANCE') and isWorldUnitClass(pUnit.getUnitClassType()) == False):
						if pUnit.getDomainType() == CvUtil.getInfoTypeForStringCached('DOMAIN_LAND'):
							newUnit = bPlayer.initUnit(pUnit.getUnitType(), pPlot.getX(), pPlot.getY(), UnitAITypes.NO_UNITAI)
							newUnit.setExperience(pUnit.getExperience(), -1)
							newUnit.setLevel(pUnit.getLevel())
							for iCount in range(gc.getNumPromotionInfos()):
								if (pUnit.isHasPromotion(iCount)):
									newUnit.setHasPromotion(iCount, True)
							pUnit.kill(False,0)
