## Sid Meier's Civilization 4
## Copyright Firaxis Games 2005
## 
## CvEventManager
## This class is passed an argsList from CvAppInterface.onEvent
## The argsList can contain anything from mouse location to key info
## The EVENTLIST that are being notified can be found 


from CvPythonExtensions import *
import CvUtil
import CvScreensInterface
import CvDebugTools
import CvWBPopups
import PyHelpers
import Popup as PyPopup
import CvCameraControls
import CvTopCivs
import sys
import CvWorldBuilderScreen
import CvAdvisorUtils
import CvTechChooser
import CvIntroMovieScreen
import Trophies

import CustomFunctions


gc = CyGlobalContext()

cf = CustomFunctions.CustomFunctions()
localText = CyTranslator()
PyPlayer = PyHelpers.PyPlayer
PyInfo = PyHelpers.PyInfo

manaTypes = [ 'BONUS_MANA_AIR','BONUS_MANA_BODY','BONUS_MANA_CHAOS','BONUS_MANA_DEATH','BONUS_MANA_DIMENSIONAL','BONUS_MANA_EARTH','BONUS_MANA_ENCHANTMENT','BONUS_MANA_ENTROPY','BONUS_MANA_FIRE','BONUS_MANA_LAW','BONUS_MANA_LIFE','BONUS_MANA_MIND','BONUS_MANA_NATURE','BONUS_MANA_SPIRIT','BONUS_MANA_WATER' ]

minBonusTypes = [ 'BONUS_REAGENTS','BONUS_IRON','BONUS_GUNPOWDER','BONUS_COPPER','BONUS_MITHRIL','BONUS_INCENSE','BONUS_GEMS' ]

# globals
###################################################
class CvEventManager:
	def __init__(self):
		#################### ON EVENT MAP ######################
		#print "EVENTMANAGER INIT"
		
		self.bCtrl = False
		self.bShift = False
		self.bAlt = False
		self.bAllowCheats = False
		
		# OnEvent Enums
		self.EventLButtonDown=1
		self.EventLcButtonDblClick=2
		self.EventRButtonDown=3
		self.EventBack=4
		self.EventForward=5
		self.EventKeyDown=6
		self.EventKeyUp=7
	
		self.__LOG_MOVEMENT = 0
		self.__LOG_BUILDING = 0
		self.__LOG_COMBAT = 0
		self.__LOG_CONTACT = 0
		self.__LOG_IMPROVEMENT = 0
		self.__LOG_CITYLOST = 0
		self.__LOG_CITYBUILDING = 0
		self.__LOG_TECH = 0
		self.__LOG_UNITBUILD = 0
		self.__LOG_UNITKILLED = 0
		self.__LOG_UNITLOST = 0
		self.__LOG_UNITPROMOTED = 0
		self.__LOG_UNITSELECTED = 0
		self.__LOG_GOODYRECEIVED = 0
		self.__LOG_GREATPERSON = 0
		self.__LOG_RELIGION = 0
		self.__LOG_RELIGIONSPREAD = 0
		self.__LOG_GOLDENAGE = 0
		self.__LOG_ENDGOLDENAGE = 0
		self.__LOG_WARPEACE = 0

		## EVENTLIST
		self.EventHandlerMap = {
			'mouseEvent'			: self.onMouseEvent,
			'kbdEvent' 				: self.onKbdEvent,
			'Init'					: self.onInit,
			'Update'				: self.onUpdate,
			'UnInit'				: self.onUnInit,
			'OnSave'				: self.onSaveGame,
			'OnPreSave'				: self.onPreSave,
			'OnLoad'				: self.onLoadGame,
			'GameStart'				: self.onGameStart,
			'GameEnd'				: self.onGameEnd,
			'plotRevealed' 			: self.onPlotRevealed,
			'BeginGameTurn'			: self.onBeginGameTurn,
			'EndGameTurn'			: self.onEndGameTurn,
			'BeginPlayerTurn'		: self.onBeginPlayerTurn,
			'EndPlayerTurn'			: self.onEndPlayerTurn,
			'endTurnReady'			: self.onEndTurnReady,
			'combatResult' 			: self.onCombatResult,
			'combatLogCalc'	 		: self.onCombatLogCalc,
			'combatLogHit'			: self.onCombatLogHit,
			'improvementBuilt' 		: self.onImprovementBuilt,
			'routeBuilt' 			: self.onRouteBuilt,
			'firstContact' 			: self.onFirstContact,
			'cityBuilt' 			: self.onCityBuilt,
			'cityRazed'				: self.onCityRazed,
			'cityAcquired' 			: self.onCityAcquired,
			'cityLost'				: self.onCityLost,
			'cultureExpansion' 		: self.onCultureExpansion,
			'cityGrowth' 			: self.onCityGrowth,
			'cityDoTurn' 			: self.onCityDoTurn,
			'cityBuildingUnit'		: self.onCityBuildingUnit,
			'cityBuildingBuilding'	: self.onCityBuildingBuilding,
			'cityRename'			: self.onCityRename,
			'unitMove' 				: self.onUnitMove,
			'unitSetXY' 			: self.onUnitSetXY,
			'unitCreated' 			: self.onUnitCreated,
			'unitBuilt' 			: self.onUnitBuilt,
			'unitKilled'			: self.onUnitKilled,
			'unitLost'				: self.onUnitLost,
			'unitPromoted'			: self.onUnitPromoted,
			'unitSelected'			: self.onUnitSelected, 
			'UnitRename'			: self.onUnitRename,
			'goodyReceived'        	: self.onGoodyReceived,
			'greatPersonBorn'      	: self.onGreatPersonBorn,
			'buildingBuilt' 		: self.onBuildingBuilt,
			'projectBuilt' 			: self.onProjectBuilt,
			'techAcquired'			: self.onTechAcquired,
			'techSelected'			: self.onTechSelected,
			'religionFounded'		: self.onReligionFounded,
			'religionSpread'		: self.onReligionSpread, 
			'goldenAge'				: self.onGoldenAge,
			'endGoldenAge'			: self.onEndGoldenAge,
			'chat' 					: self.onChat,
			'victory'				: self.onVictory,
			'changeWar'				: self.onChangeWar,
			'setPlayerAlive'		: self.onSetPlayerAlive,
			'windowActivation'		: self.onWindowActivation,
			'gameUpdate'			: self.onGameUpdate,		# sample generic event
		}

		################## Events List ###############################
		#
		# Dictionary of Events, indexed by EventID (also used at popup context id)
		#   entries have name, beginFunction, applyFunction [, randomization weight...]
		#
		# Normal events first, random events after
		#	
		################## Events List ###############################
		self.Events={
			CvUtil.EventEditCityName : ('EditCityName', self.__eventEditCityNameApply, self.__eventEditCityNameBegin),
			CvUtil.EventEditCity : ('EditCity', self.__eventEditCityApply, self.__eventEditCityBegin),
			CvUtil.EventPlaceObject : ('PlaceObject', self.__eventPlaceObjectApply, self.__eventPlaceObjectBegin),
			CvUtil.EventAwardTechsAndGold: ('AwardTechsAndGold', self.__eventAwardTechsAndGoldApply, self.__eventAwardTechsAndGoldBegin),
			CvUtil.EventEditUnitName : ('EditUnitName', self.__eventEditUnitNameApply, self.__eventEditUnitNameBegin),
			CvUtil.EventWBAllPlotsPopup : ('WBAllPlotsPopup', self.__eventWBAllPlotsPopupApply, self.__eventWBAllPlotsPopupBegin),
			CvUtil.EventWBLandmarkPopup : ('WBLandmarkPopup', self.__eventWBLandmarkPopupApply, self.__eventWBLandmarkPopupBegin),
			CvUtil.EventWBScriptPopup : ('WBScriptPopup', self.__eventWBScriptPopupApply, self.__eventWBScriptPopupBegin),
			CvUtil.EventWBStartYearPopup : ('WBStartYearPopup', self.__eventWBStartYearPopupApply, self.__eventWBStartYearPopupBegin),
			CvUtil.EventHasTrait : ('HasTrait', self.__eventHasTraitApply, self.__eventHasTraitBegin),
			CvUtil.EventChangePlayer : ('ChangePlayer', self.__eventChangePlayerApply, self.__eventChangePlayerBegin),
			CvUtil.EventReassignPlayer : ('ReassignPlayer', self.__eventReassignPlayerApply, self.__eventReassignPlayerBegin),
			CvUtil.EventCreatePlayer : ('CreatePlayer', self.__eventCreatePlayerApply, self.__eventCreatePlayerBegin),
		}	

		#################### EVENT STARTERS ######################
	def handleEvent(self, argsList):
		'EventMgr entry point'
		# extract the last 6 args in the list, the first arg has already been consumed
		self.origArgsList = argsList	# point to original
		tag = argsList[0]				# event type string
		idx = len(argsList)-6
		bDummy = false
		self.bDbg, bDummy, self.bAlt, self.bCtrl, self.bShift, self.bAllowCheats = argsList[idx:]
		ret = 0
		if self.EventHandlerMap.has_key(tag):
			fxn = self.EventHandlerMap[tag]
			ret = fxn(argsList[1:idx])
		return ret
		
#################### EVENT APPLY ######################	
	def beginEvent( self, context, argsList=-1 ):
		'Begin Event'
		entry = self.Events[context]
		return entry[2]( argsList )
	
	def applyEvent( self, argsList ):
		'Apply the effects of an event '
		context, playerID, netUserData, popupReturn = argsList
		
		if context == CvUtil.PopupTypeEffectViewer:
			return CvDebugTools.g_CvDebugTools.applyEffectViewer( playerID, netUserData, popupReturn )
		
		entry = self.Events[context]
				
		if ( context not in CvUtil.SilentEvents ):
			self.reportEvent(entry, context, (playerID, netUserData, popupReturn) )
		return entry[1]( playerID, netUserData, popupReturn )   # the apply function

	def reportEvent(self, entry, context, argsList):
		'Report an Event to Events.log '
		message = "DEBUG Event: %s (%s)" %(entry[0], gc.getActivePlayer().getName())
		CyInterface().addImmediateMessage(message,"")
		CvUtil.pyPrint(message)
		return 0
		
#################### ON EVENTS ######################
	def onKbdEvent(self, argsList):
		'keypress handler - return 1 if the event was consumed'

		eventType,key,mx,my,px,py = argsList
		game = CyGame()
		
		if (self.bAllowCheats):
			# notify debug tools of input to allow it to override the control
			argsList = (eventType,key,self.bCtrl,self.bShift,self.bAlt,mx,my,px,py,CyGame().isNetworkMultiPlayer())
			if ( CvDebugTools.g_CvDebugTools.notifyInput(argsList) ):
				return 0
		
		if ( eventType == self.EventKeyDown ):
			theKey=int(key)
			
			CvCameraControls.g_CameraControls.handleInput( theKey )
						
			# Shift - T (Debug - No MP)
			if (self.bAllowCheats and theKey == int(InputTypes.KB_T)):
				if ( self.bShift ):
					self.beginEvent(CvUtil.EventAwardTechsAndGold)
					#self.beginEvent(CvUtil.EventCameraControlPopup)
					return 1
						
			# Shift - ] (Debug - currently mouse-overd unit, health += 10
			elif (self.bAllowCheats and theKey == int(InputTypes.KB_LBRACKET) and self.bShift ):
				unit = CyMap().plot(px, py).getUnit(0)
				if ( not unit.isNone() ):
					d = min( unit.maxHitPoints()-1, unit.getDamage() + gc.getMAX_HIT_POINTS()/10 ) # Smarter Orcs fix.
					unit.setDamage( d, PlayerTypes.NO_PLAYER )
				
			# Shift - [ (Debug - currently mouse-overd unit, health -= 10
			elif (self.bAllowCheats and theKey == int(InputTypes.KB_RBRACKET) and self.bShift ):
				unit = CyMap().plot(px, py).getUnit(0)
				if ( not unit.isNone() ):
					d = max( 0, unit.getDamage() - gc.getMAX_HIT_POINTS()/10 ) # Smarter Orcs fix.
					unit.setDamage( d, PlayerTypes.NO_PLAYER )
				
			elif (self.bAllowCheats and theKey == int(InputTypes.KB_F1)):
				if ( self.bShift ):
					CvScreensInterface.replayScreen.showScreen(False)
					return 1
				# don't return 1 unless you want the input consumed
			
			elif (self.bAllowCheats and theKey == int(InputTypes.KB_F2)):
				if ( self.bShift ):
					import CvDebugInfoScreen
 					CvScreensInterface.showDebugInfoScreen()
					return 1
			
			elif (self.bAllowCheats and theKey == int(InputTypes.KB_F3)):
				if ( self.bShift ):
					CvScreensInterface.showDanQuayleScreen(())
					return 1
					
			elif (self.bAllowCheats and theKey == int(InputTypes.KB_F4)):
				if ( self.bShift ):
					CvScreensInterface.showUnVictoryScreen(())
					return 1
					
		return 0

	def onInit(self, argsList):
		'Called when Civ starts up'
		CvUtil.pyPrint( 'OnInit' )
		
	def onUpdate(self, argsList):
		'Called every frame'
		fDeltaTime = argsList[0]

		# allow camera to be updated
		CvCameraControls.g_CameraControls.onUpdate( fDeltaTime )
		
	def onWindowActivation(self, argsList):
		'Called when the game window activates or deactivates'
		bActive = argsList[0]
		
	def onUnInit(self, argsList):
		'Called when Civ shuts down'
		CvUtil.pyPrint('OnUnInit')
	
	def onPreSave(self, argsList):
		"called before a game is actually saved"
		CvUtil.pyPrint('OnPreSave')
	
	def onSaveGame(self, argsList):
		"return the string to be saved - Must be a string"
		return ""

	def onLoadGame(self, argsList):
		return 0

	def onGameStart(self, argsList):
		'Called at the start of the game'

		introMovie = CvIntroMovieScreen.CvIntroMovieScreen()
		introMovie.interfaceScreen()

		iPlayerNum = 0
		iBrokenLands = CvUtil.getInfoTypeForStringCached('TERRAIN_BROKEN_LANDS')
		iBurningSands = CvUtil.getInfoTypeForStringCached('TERRAIN_BURNING_SANDS')
		iFieldsofPerdition = CvUtil.getInfoTypeForStringCached('TERRAIN_FIELDS_OF_PERDITION')
		iSettler = CvUtil.getInfoTypeForStringCached('UNITCLASS_SETTLER')
		iMobility1 = CvUtil.getInfoTypeForStringCached('PROMOTION_MOBILITY1')
		iMobility2 = CvUtil.getInfoTypeForStringCached('PROMOTION_MOBILITY2')
		iSentry3 = CvUtil.getInfoTypeForStringCached('PROMOTION_SENTRY3')
		for iPlayer in range(gc.getMAX_PLAYERS()):
			player = gc.getPlayer(iPlayer)
			if player.isAlive():
				iPlayerNum = iPlayerNum + 1
				if player.hasTrait(CvUtil.getInfoTypeForStringCached('TRAIT_INSANE')):
					cf.FFHInsane(iPlayer)
				if player.hasTrait(CvUtil.getInfoTypeForStringCached('TRAIT_BARBARIAN')):
					eTeam = gc.getTeam(player.getTeam())
					iTeam = gc.getPlayer(gc.getBARBARIAN_PLAYER()).getTeam()
					eTeam.makePeace(iTeam)
				if player.hasTrait(CvUtil.getInfoTypeForStringCached('TRAIT_INGENUITY')):
					player.setGold(player.getGold() + 50)
				py = PyPlayer(iPlayer)
				for pUnit in py.getUnitList():
					if pUnit.getUnitClassType() == iSettler:
						pUnit.setHasPromotion(iMobility1, True)
						pUnit.setHasPromotion(iMobility2, True)
						pUnit.setHasPromotion(iSentry3, True)
				if player.isHuman():
					popupInfo = CyPopupInfo()
					popupInfo.setButtonPopupType(ButtonPopupTypes.BUTTONPOPUP_PYTHON_SCREEN)
					popupInfo.setText(u"showDawnOfMan")
					popupInfo.addPopup(iPlayer)
					
		if gc.getGame().isPbem():
			for iPlayer in range(gc.getMAX_PLAYERS()):
				player = gc.getPlayer(iPlayer)
				if (player.isAlive() and player.isHuman()):
					popupInfo = CyPopupInfo()
					popupInfo.setButtonPopupType(ButtonPopupTypes.BUTTONPOPUP_DETAILS)
					popupInfo.setOption1(true)
					popupInfo.addPopup(iPlayer)

		for iBonus in minBonusTypes:
			if CyMap().getNumBonuses(CvUtil.getInfoTypeForStringCached(iBonus)) < 4:
				cf.FFHAddBonus(iBonus, 4 - CyMap().getNumBonuses(CvUtil.getInfoTypeForStringCached(iBonus)), -1)

		if CyMap().getNumBonuses(CvUtil.getInfoTypeForStringCached('BONUS_MANA')) < iPlayerNum:
			cf.FFHAddBonus('BONUS_MANA', iPlayerNum - CyMap().getNumBonuses(CvUtil.getInfoTypeForStringCached('BONUS_MANA')), -1)
			
		if CyGame().isOption(CvUtil.getInfoTypeForStringCached('GAMEOPTION_NO_BARBARIANS')) == False:
			if CyGame().isOption(CvUtil.getInfoTypeForStringCached('GAMEOPTION_NO_LAIRS')) == False:
				cf.FFHAddImprovements()
			listPlots = []
			for i in range (CyMap().numPlots()):
				pPlot = CyMap().plotByIndex(i)
				pArea = pPlot.area()
				if (pArea.isWater() and pArea.getNumTiles() > 20 and pPlot.getFeatureType() == -1):
					listPlots.append(i)
			if len(listPlots) > 0:
				iRnd = CyGame().getSorenRandNum(len(listPlots), "Sailors Dirge")
				pPlot = CyMap().plotByIndex(listPlots[iRnd])
				bPlayer = gc.getPlayer(gc.getBARBARIAN_PLAYER())
				newUnit = bPlayer.initUnit(CvUtil.getInfoTypeForStringCached('UNIT_SAILORS_DIRGE'), pPlot.getX(), pPlot.getY(), UnitAITypes.NO_UNITAI)

		for i in range (CyMap().numPlots()):
			pPlot = CyMap().plotByIndex(i)
			if (pPlot.getTerrainType() == iBurningSands or pPlot.getTerrainType() == iBrokenLands or pPlot.getTerrainType() == iFieldsofPerdition):
				pPlot.changeHellCount(50)

		CyGame().changeArmageddonCounterLimit((-4 + iPlayerNum) * 5)
		CyEngine().setDirty(EngineDirtyBits.MinimapTexture_DIRTY_BIT, True)
		CyInterface().setDirty(InterfaceDirtyBits.MinimapSection_DIRTY_BIT, True)

	def onGameEnd(self, argsList):
		'Called at the End of the game'
		print("Game is ending")
		return

	def onBeginGameTurn(self, argsList):
		'Called at the beginning of the end of each turn'
		iGameTurn = argsList[0]
		iOrthusTurn = gc.getDefineINT('ORTHUS_TURN')

		if CyGame().isUnitClassMaxedOut(CvUtil.getInfoTypeForStringCached('UNITCLASS_ORTHUS'), 0) == False:
			if CyGame().isOption(CvUtil.getInfoTypeForStringCached('GAMEOPTION_NO_BARBARIANS')) == False:
				bOrthus = False
				if CyGame().getGameSpeedType() == CvUtil.getInfoTypeForStringCached('GAMESPEED_QUICK'):
					if iGameTurn >= iOrthusTurn / 3 * 2:
						bOrthus = True
				if CyGame().getGameSpeedType() == CvUtil.getInfoTypeForStringCached('GAMESPEED_NORMAL'):
					if iGameTurn >= iOrthusTurn:
						bOrthus = True
				if CyGame().getGameSpeedType() == CvUtil.getInfoTypeForStringCached('GAMESPEED_EPIC'):
					if iGameTurn >= iOrthusTurn * 3 / 2:
						bOrthus = True
				if CyGame().getGameSpeedType() == CvUtil.getInfoTypeForStringCached('GAMESPEED_MARATHON'):
					if iGameTurn >= iOrthusTurn * 3:
						bOrthus = True
				if bOrthus:
					cf.FFHAddUnit(CvUtil.getInfoTypeForStringCached('UNIT_ORTHUS'))
					cf.FFHAddPopup(CyTranslator().getText("TXT_KEY_POPUP_ORTHUS_CREATION",()), 'art/interface/popups/Orthus.dds')

		CvTopCivs.CvTopCivs().turnChecker(iGameTurn)

	def onEndGameTurn(self, argsList):
		'Called at the end of the end of each turn'
		iGameTurn = argsList[0]

	def onBeginPlayerTurn(self, argsList):
		'Called at the beginning of a players turn'
		iGameTurn, iPlayer = argsList
		pPlayer = gc.getPlayer(iPlayer)
		py = PyPlayer(iPlayer)

		if pPlayer.isHuman() == False:
			cf.FFHWarScript(iPlayer)

		pPlayer.setFeatAccomplished(FeatTypes.FEAT_POPULATION_2_BILLION, False)

		if (pPlayer.hasTrait(CvUtil.getInfoTypeForStringCached('TRAIT_ADAPTIVE')) and pPlayer.isHuman()):
			iCycle = 100
			if CyGame().getGameSpeedType() == CvUtil.getInfoTypeForStringCached('GAMESPEED_QUICK'):
				iCycle = 75
			if CyGame().getGameSpeedType() == CvUtil.getInfoTypeForStringCached('GAMESPEED_EPIC'):
				iCycle = 150
			if CyGame().getGameSpeedType() == CvUtil.getInfoTypeForStringCached('GAMESPEED_MARATHON'):
				iCycle = 300
			if (iGameTurn - 5 == 0 or iGameTurn - 5 == iCycle or iGameTurn - 5 == iCycle * 2 or iGameTurn - 5 == iCycle * 3 or iGameTurn - 5 == iCycle * 4):
				cf.FFHSelectTrait(iPlayer)

		if pPlayer.hasTrait(CvUtil.getInfoTypeForStringCached('TRAIT_INSANE')):
			if CyGame().getSorenRandNum(10000, "Bob") <= gc.getDefineINT('INSANE_CHANCE'):
				cf.FFHInsane(iPlayer)

		if pPlayer.getStateReligion() == CvUtil.getInfoTypeForStringCached('RELIGION_THE_ASHEN_VEIL'):
			if pPlayer.getCurrentEra() != CvUtil.getInfoTypeForStringCached('ERA_VEIL'):
				pPlayer.setCurrentEra(CvUtil.getInfoTypeForStringCached('ERA_VEIL'))

		if pPlayer.getStateReligion() == CvUtil.getInfoTypeForStringCached('RELIGION_THE_ORDER'):
			if pPlayer.getCurrentEra() != CvUtil.getInfoTypeForStringCached('ERA_ORDE'):
				pPlayer.setCurrentEra(CvUtil.getInfoTypeForStringCached('ERA_ORDE'))

		if pPlayer.getStateReligion() == CvUtil.getInfoTypeForStringCached('RELIGION_FELLOWSHIP_OF_LEAVES'):
			if pPlayer.getCurrentEra() != CvUtil.getInfoTypeForStringCached('ERA_LEAF'):
				pPlayer.setCurrentEra(CvUtil.getInfoTypeForStringCached('ERA_LEAF'))

		if pPlayer.getStateReligion() == CvUtil.getInfoTypeForStringCached('RELIGION_RUNES_OF_KILMORPH'):
			if pPlayer.getCurrentEra() != CvUtil.getInfoTypeForStringCached('ERA_RUNE'):
				pPlayer.setCurrentEra(CvUtil.getInfoTypeForStringCached('ERA_RUNE'))

		if pPlayer.getStateReligion() == CvUtil.getInfoTypeForStringCached('RELIGION_OCTOPUS_OVERLORDS'):
			if pPlayer.getCurrentEra() != CvUtil.getInfoTypeForStringCached('ERA_OCTO'):
				pPlayer.setCurrentEra(CvUtil.getInfoTypeForStringCached('ERA_OCTO'))

		if pPlayer.getCivilizationType() == CvUtil.getInfoTypeForStringCached('CIVILIZATION_KHAZAD'):
			cf.FFHDoTurnKhazad(iPlayer)

		if pPlayer.getCivilizationType() == CvUtil.getInfoTypeForStringCached('CIVILIZATION_LUCHUIRP'):
			cf.FFHDoTurnLuchuirp(iPlayer)

		if pPlayer.hasTrait(CvUtil.getInfoTypeForStringCached('TRAIT_BARBARIAN')):
			eTeam = gc.getTeam(gc.getPlayer(gc.getBARBARIAN_PLAYER()).getTeam())
			iTeam = pPlayer.getTeam()
			if eTeam.isAtWar(iTeam) == False:
				if 2 * CyGame().getPlayerScore(iPlayer) >= 3 * CyGame().getPlayerScore(CyGame().getRankPlayer(1)):
					if gc.getGame().getGameTurnYear() >= 20:
						eTeam.declareWar(iTeam, False)
						if (iPlayer == CyGame().getActivePlayer() and pPlayer.isHuman()):
							cf.FFHAddPopup(CyTranslator().getText("TXT_KEY_POPUP_BARBARIAN_DECLARE_WAR",()), 'art/interface/popups/Barbarian.dds')

	def onEndPlayerTurn(self, argsList):
		'Called at the end of a players turn'
		iGameTurn, iPlayer = argsList

		if (gc.getGame().getElapsedGameTurns() == 1):
			if (gc.getPlayer(iPlayer).isHuman()):
				if (gc.getPlayer(iPlayer).canRevolution(0)):
					popupInfo = CyPopupInfo()
					popupInfo.setButtonPopupType(ButtonPopupTypes.BUTTONPOPUP_CHANGECIVIC)
					popupInfo.addPopup(iPlayer)

		CvAdvisorUtils.resetAdvisorNags()
		CvAdvisorUtils.endTurnFeats(iPlayer)

	def onEndTurnReady(self, argsList):
		iGameTurn = argsList[0]

	def onFirstContact(self, argsList):
		'Contact'
		iTeamX,iHasMetTeamY = argsList
		if (not self.__LOG_CONTACT):
			return
		CvUtil.pyPrint('Team %d has met Team %d' %(iTeamX, iHasMetTeamY))

	def onCombatResult(self, argsList):
		'Combat Result'
		pWinner,pLoser = argsList
		playerX = PyPlayer(pWinner.getOwner())
		unitX = PyInfo.UnitInfo(pWinner.getUnitType())
		playerY = PyPlayer(pLoser.getOwner())
		unitY = PyInfo.UnitInfo(pLoser.getUnitType())
		pPlayer = gc.getPlayer(pWinner.getOwner())
		bUncaptured = True

		if pPlayer.isHuman():
			if pLoser.getUnitType() == CvUtil.getInfoTypeForStringCached('UNIT_ACHERON'):
				Trophies.addTrophy("TROPHY_DEFEATED_ACHERON")
			if pLoser.getUnitType() == CvUtil.getInfoTypeForStringCached('UNIT_BASIUM'):
				Trophies.addTrophy("TROPHY_DEFEATED_BASIUM")
			if pLoser.getUnitType() == CvUtil.getInfoTypeForStringCached('UNIT_HYBOREM'):
				Trophies.addTrophy("TROPHY_DEFEATED_HYBOREM")
			if pLoser.getUnitType() == CvUtil.getInfoTypeForStringCached('UNIT_ORTHUS'):
				Trophies.addTrophy("TROPHY_DEFEATED_ORTHUS")

		if pWinner.isNoCapture():
			bUncaptured = False

		if pLoser.getCaptureUnitType(pPlayer.getCivilizationType()) != -1:
			bUncaptured = False

		if (pLoser.isAlive() and pWinner.getUnitType() == CvUtil.getInfoTypeForStringCached('UNIT_GIANT_SPIDER')):
			bUncaptured = False
			newUnit = pPlayer.initUnit(CvUtil.getInfoTypeForStringCached('UNIT_BABY_SPIDER'), pWinner.getX(), pWinner.getY(), UnitAITypes.NO_UNITAI)
			newUnit.setMadeAttack(True)
			newUnit.finishMoves()
			newUnit.setDamage(gc.getMAX_HIT_POINTS()/2, -1)# Smarter Orcs defect fix.
			CyInterface().addMessage(pWinner.getOwner(),True,25,CyTranslator().getText("TXT_KEY_MESSAGE_BABY_SPIDER",()),'AS2D_DISCOVERBONUS',1,'Art/Interface/Buttons/Units/GiantSpider.dds',ColorTypes(8),pWinner.getX(),pWinner.getY(),True,True)
			if pPlayer.getCivilizationType() == CvUtil.getInfoTypeForStringCached('CIVILIZATION_SVARTALFAR'):
				newUnit = pPlayer.initUnit(CvUtil.getInfoTypeForStringCached('UNIT_BABY_SPIDER'), pWinner.getX(), pWinner.getY(), UnitAITypes.NO_UNITAI)
				newUnit.setMadeAttack(True)
				newUnit.finishMoves()

		if (pPlayer.getCivics(CvUtil.getInfoTypeForStringCached('CIVICOPTION_LABOR')) == CvUtil.getInfoTypeForStringCached('CIVIC_SLAVERY') or pWinner.isHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_TASKMASTER'))):
			if (bUncaptured and pLoser.isAlive() and pLoser.getUnitCombatType() != CvUtil.getInfoTypeForStringCached('UNITCOMBAT_ANIMAL') and pWinner.getSpecialUnitType() != CvUtil.getInfoTypeForStringCached('SPECIALUNIT_SPELL')):
				if CyGame().getSorenRandNum(100, "Bob") <= gc.getDefineINT('SLAVERY_CHANCE'):
					CyInterface().addMessage(pWinner.getOwner(),True,25,'Slave captured.','AS2D_DISCOVERBONUS',1,'Art/Interface/Buttons/Units/Slave.dds',ColorTypes(8),pLoser.getX(),pLoser.getY(),True,True)
					unitType = CvUtil.getInfoTypeForStringCached('UNIT_SLAVE')
					if pLoser.isHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_DWARF')):
						unitType = CvUtil.getInfoTypeForStringCached('UNIT_SLAVE_DWARF')
					if pLoser.isHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_ELF')):
						unitType = CvUtil.getInfoTypeForStringCached('UNIT_SLAVE_ELF')
					if pLoser.isHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_ORC')):
						unitType = CvUtil.getInfoTypeForStringCached('UNIT_SLAVE_ORC')
					bUncaptured = False
					cf.FFHConvert(pWinner, pLoser, unitType, False)
			
		if (pWinner.isHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_COMMAND')) and bUncaptured):
			if pLoser.isAlive():
				if CyGame().getSorenRandNum(100, "Bob") <= 50:
					bUncaptured = False
					cf.FFHConvert(pWinner, pLoser, pLoser.getUnitType(), True)

		if (pWinner.getUnitType() == CvUtil.getInfoTypeForStringCached('UNIT_GOBLIN') and bUncaptured):
			if (pLoser.getUnitType() == CvUtil.getInfoTypeForStringCached('UNIT_WOLF') or pLoser.getUnitType() == CvUtil.getInfoTypeForStringCached('UNIT_WOLF_PACK')):
				bUncaptured = False
				cf.FFHConvert(pWinner, pWinner, CvUtil.getInfoTypeForStringCached('UNIT_WOLF_RIDER'), True)
				pWinner.setDamage(gc.getMAX_HIT_POINTS(), True) # Smarter Orcs - defect fix.

		if (pWinner.isHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_SUBDUE_ANIMAL')) and bUncaptured):
			if pLoser.isAnimal() and pLoser.getUnitCombatType() == CvUtil.getInfoTypeForStringCached('UNITCOMBAT_ANIMAL'):
				bUncaptured = False
				cf.FFHConvert(pWinner, pLoser, pLoser.getUnitType(), True)

		if (pWinner.isHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_WEREWOLF')) and pLoser.isAlive() and pLoser.getUnitCombatType() != CvUtil.getInfoTypeForStringCached('UNITCOMBAT_ANIMAL') and pLoser.getUnitCombatType() != CvUtil.getInfoTypeForStringCached('UNITCOMBAT_BEAST') and bUncaptured):
			iNumWW = pPlayer.getUnitClassCount(CvUtil.getInfoTypeForStringCached('UNITCLASS_RAVENOUS_WEREWOLF')) + pPlayer.getUnitClassCount(CvUtil.getInfoTypeForStringCached('UNITCLASS_WEREWOLF')) + pPlayer.getUnitClassCount(CvUtil.getInfoTypeForStringCached('UNITCLASS_GREATER_WEREWOLF'))
			iRand = CyGame().getSorenRandNum(100, "Bob")
			if (iRand <= gc.getDefineINT('WEREWOLF_BASE_CHANCE') - (gc.getDefineINT('WEREWOLF_MINUS_PER_WEREWOLF') * iNumWW)):
				CyInterface().addMessage(pWinner.getOwner(),True,25,'Lycanthropy has spread.','AS2D_WOLF',1,'Art/Interface/Buttons/Units/Ravenouswerewolf.dds',ColorTypes(8),pWinner.getX(),pWinner.getY(),True,True)
				bUncaptured = False
				cf.FFHConvert(pWinner, pLoser, CvUtil.getInfoTypeForStringCached('UNIT_RAVENOUS_WEREWOLF'), False)

		if (pWinner.getUnitType() == CvUtil.getInfoTypeForStringCached('UNIT_RAVENOUS_WEREWOLF') and pLoser.isAlive()):
			CyInterface().addMessage(pWinner.getOwner(),True,25,'A Werewolf has become stronger.','AS2D_WOLF',1,'Art/Interface/Buttons/Units/Werewolf.dds',ColorTypes(8),pWinner.getX(),pWinner.getY(),True,True)
			newUnit = pPlayer.initUnit(CvUtil.getInfoTypeForStringCached('UNIT_WEREWOLF'), pWinner.getX(), pWinner.getY(), UnitAITypes.NO_UNITAI)
			newUnit.finishMoves()
			newUnit.setExperience(pWinner.getExperience(), -1)
			newUnit.setLevel(pWinner.getLevel())
			newUnit.setDamage(pWinner.getDamage(), False)
			for iCount in range(gc.getNumPromotionInfos()):
				if (pWinner.isHasPromotion(iCount)):
					newUnit.setHasPromotion(iCount, True)
			newUnit.setHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_ENRAGED'), False)
			pWinner.setDamage(gc.getMAX_HIT_POINTS(), True) # Smarter Orcs - defect fix.

		if (pWinner.getUnitType() == CvUtil.getInfoTypeForStringCached('UNIT_WEREWOLF') and pLoser.isAlive()):
			iRand = CyGame().getSorenRandNum(100, "Bob")
			if (iRand <= 5):
				CyInterface().addMessage(pWinner.getOwner(),True,25,'A Werewolf has become stronger.','AS2D_WOLF',1,'Art/Interface/Buttons/Units/Greaterwerewolf.dds',ColorTypes(8),pWinner.getX(),pWinner.getY(),True,True)
				newUnit = pPlayer.initUnit(CvUtil.getInfoTypeForStringCached('UNIT_GREATER_WEREWOLF'), pWinner.getX(), pWinner.getY(), UnitAITypes.NO_UNITAI)
				newUnit.finishMoves()
				newUnit.setExperience(pWinner.getExperience(), -1)
				newUnit.setLevel(pWinner.getLevel())
				newUnit.setDamage(pWinner.getDamage(), False)
				for iCount in range(gc.getNumPromotionInfos()):
					if (pWinner.isHasPromotion(iCount)):
						newUnit.setHasPromotion(iCount, True)
				pWinner.setDamage(GC.getMAX_HIT_POINTS(), True) # Smarter Orcs - defect fix.

		if (pWinner.isHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_BOARDING')) and pLoser.getUnitCombatType() == CvUtil.getInfoTypeForStringCached('UNITCOMBAT_NAVAL') and bUncaptured):
			CyInterface().addMessage(pWinner.getOwner(),True,25,'Ship Captured.','AS2D_GOODY_GOLD',1,'Art/Interface/Buttons/Promotions/Boarding.dds',ColorTypes(8),pLoser.getX(),pLoser.getY(),True,True)
			bUncaptured = False
			cf.FFHConvert(pWinner, pLoser, pLoser.getUnitType(), True)

		if (pLoser.getUnitType() == CvUtil.getInfoTypeForStringCached('UNIT_BASIUM') or pLoser.getUnitType() == CvUtil.getInfoTypeForStringCached('UNIT_HYBOREM')):
			if pLoser.isHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_IMMORTAL')) == False:
				if pLoser.getUnitType() == CvUtil.getInfoTypeForStringCached('UNIT_BASIUM'):
					CyInterface().addMessage(pLoser.getOwner(),True,25,'Basium has Died!','AS2D_DISCOVERBONUS',1,'Art/Interface/Buttons/Units/Basium.dds',ColorTypes(8),pLoser.getX(),pLoser.getY(),True,True)
				if pLoser.getUnitType() == CvUtil.getInfoTypeForStringCached('UNIT_HYBOREM'):
					CyInterface().addMessage(pLoser.getOwner(),True,25,'Hyborem has Died!','AS2D_DISCOVERBONUS',1,'Art/Interface/Buttons/Units/Hyborem.dds',ColorTypes(8),pLoser.getX(),pLoser.getY(),True,True)
				p2Player = gc.getPlayer(pLoser.getOwner())
				for iTrait in range(gc.getNumTraitInfos()):
					if iTrait != CvUtil.getInfoTypeForStringCached('TRAIT_FALLOW'):
						CyMessageControl().sendApplyEvent(5012, EventContextTypes.EVENTCONTEXT_ALL, (pLoser.getOwner(),iTrait,False))

		if (pWinner.getUnitType() == CvUtil.getInfoTypeForStringCached('UNIT_MAGNADINE') and bUncaptured):
			if (pLoser.isAlive() and pLoser.getOwner() == gc.getBARBARIAN_PLAYER() and pLoser.getUnitCombatType() != CvUtil.getInfoTypeForStringCached('UNITCOMBAT_ANIMAL')):
				bUncaptured = False
				cf.FFHConvert(pWinner, pLoser, pLoser.getUnitType(), True)
		
		if pWinner.getUnitType() == CvUtil.getInfoTypeForStringCached('UNIT_ARS'):
			if (pLoser.isAlive() and bUncaptured):
				bUncaptured = False
				cf.FFHConvert(pWinner, pLoser, CvUtil.getInfoTypeForStringCached('UNIT_WRAITH'), False)

		if (pLoser.getUnitClassType() == CvUtil.getInfoTypeForStringCached('UNITCLASS_TAR_DEMON') and bUncaptured):
			bUncaptured = False
			iStr = pLoser.baseCombatStr(false) - 2
			if iStr > 1:
				pPlayer2 = gc.getPlayer(pLoser.getOwner())
				newUnit = pPlayer2.initUnit(CvUtil.getInfoTypeForStringCached('UNIT_TAR_DEMON'), pLoser.getX(), pLoser.getY(), UnitAITypes.NO_UNITAI)
				newUnit.setDamage(gc.getMAX_HIT_POINTS()/2, -1)# Smarter Orcs defect fix.
				newUnit.setMadeAttack(True)
				newUnit.changeExtraCombatDefense(iStr - newUnit.baseCombatStr(false))
				newUnit = pPlayer2.initUnit(CvUtil.getInfoTypeForStringCached('UNIT_TAR_DEMON'), pLoser.getX(), pLoser.getY(), UnitAITypes.NO_UNITAI)
				newUnit.changeExtraCombatDefense(iStr - newUnit.baseCombatStr(false))
				newUnit.setDamage(gc.getMAX_HIT_POINTS()/2, -1)# Smarter Orcs defect fix.
				newUnit.setMadeAttack(True)
				CyInterface().addMessage(pWinner.getOwner(),True,25,CyTranslator().getText("TXT_KEY_MESSAGE_TAR_DEMON_SPLIT",()),'AS2D_DISCOVERBONUS',1,'Art/Interface/Buttons/Units/Tar Demon.dds',ColorTypes(7),pLoser.getX(),pLoser.getY(),True,True)
				CyInterface().addMessage(pLoser.getOwner(),True,25,CyTranslator().getText("TXT_KEY_MESSAGE_TAR_DEMON_SPLIT",()),'AS2D_DISCOVERBONUS',1,'Art/Interface/Buttons/Units/Tar Demon.dds',ColorTypes(8),pLoser.getX(),pLoser.getY(),True,True)

		if pLoser.isHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_IMMORTAL')):
			if bUncaptured == False:
				pLoser.setHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_IMMORTAL'), False)

		if (not self.__LOG_COMBAT):
			return
		if playerX and playerX and unitX and playerY:
			CvUtil.pyPrint('Player %d Civilization %s Unit %s has defeated Player %d Civilization %s Unit %s' 
				%(playerX.getID(), playerX.getCivilizationName(), unitX.getDescription(), 
				playerY.getID(), playerY.getCivilizationName(), unitY.getDescription()))

	def onCombatLogCalc(self, argsList):
		'Combat Result'	
		genericArgs = argsList[0][0]
		cdAttacker = genericArgs[0]
		cdDefender = genericArgs[1]
		iCombatOdds = genericArgs[2]
		CvUtil.combatMessageBuilder(cdAttacker, cdDefender, iCombatOdds)
		
	def onCombatLogHit(self, argsList):
		'Combat Message'
		global gCombatMessages, gCombatLog
		genericArgs = argsList[0][0]
		cdAttacker = genericArgs[0]
		cdDefender = genericArgs[1]
		iIsAttacker = genericArgs[2]
		iDamage = genericArgs[3]
		
		if (iIsAttacker == 0):
			combatMessage = localText.getText("TXT_KEY_COMBAT_MESSAGE_HIT", (gc.getPlayer(cdDefender.eOwner).getNameKey(), cdDefender.sUnitName, iDamage, cdDefender.iCurrHitPoints, cdDefender.iMaxHitPoints))
			CyInterface().addCombatMessage(cdAttacker.eOwner,combatMessage)
			CyInterface().addCombatMessage(cdDefender.eOwner,combatMessage)
			if (cdDefender.iCurrHitPoints <= 0):
				combatMessage = localText.getText("TXT_KEY_COMBAT_MESSAGE_DEFEATED", (gc.getPlayer(cdAttacker.eOwner).getNameKey(), cdAttacker.sUnitName, gc.getPlayer(cdDefender.eOwner).getNameKey(), cdDefender.sUnitName))
				CyInterface().addCombatMessage(cdAttacker.eOwner,combatMessage)
				CyInterface().addCombatMessage(cdDefender.eOwner,combatMessage)
		elif (iIsAttacker == 1):
			combatMessage = localText.getText("TXT_KEY_COMBAT_MESSAGE_HIT", (gc.getPlayer(cdAttacker.eOwner).getNameKey(), cdAttacker.sUnitName, iDamage, cdAttacker.iCurrHitPoints, cdAttacker.iMaxHitPoints))
			CyInterface().addCombatMessage(cdAttacker.eOwner,combatMessage)
			CyInterface().addCombatMessage(cdDefender.eOwner,combatMessage)
			if (cdAttacker.iCurrHitPoints <= 0):
				combatMessage = localText.getText("TXT_KEY_COMBAT_MESSAGE_DEFEATED", (gc.getPlayer(cdDefender.eOwner).getNameKey(), cdDefender.sUnitName, gc.getPlayer(cdAttacker.eOwner).getNameKey(), cdAttacker.sUnitName))
				CyInterface().addCombatMessage(cdAttacker.eOwner,combatMessage)
				CyInterface().addCombatMessage(cdDefender.eOwner,combatMessage)

	def onImprovementBuilt(self, argsList):
		'Improvement Built'
		iImprovement, iX, iY = argsList
		pPlot = CyMap().plot(iX,iY)
		iDesert = CvUtil.getInfoTypeForStringCached('TERRAIN_DESERT')
		iFallout = CvUtil.getInfoTypeForStringCached('FEATURE_FALLOUT')
		iForestAncient = CvUtil.getInfoTypeForStringCached('FEATURE_FOREST_ANCIENT')
		iForest = CvUtil.getInfoTypeForStringCached('FEATURE_FOREST')

		if iImprovement == CvUtil.getInfoTypeForStringCached('IMPROVEMENT_MANA_AIR'):
			pPlot.setBonusType(CvUtil.getInfoTypeForStringCached('BONUS_MANA_AIR'))
		if iImprovement == CvUtil.getInfoTypeForStringCached('IMPROVEMENT_MANA_BODY'):
			pPlot.setBonusType(CvUtil.getInfoTypeForStringCached('BONUS_MANA_BODY'))
		if iImprovement == CvUtil.getInfoTypeForStringCached('IMPROVEMENT_MANA_CHAOS'):
			pPlot.setBonusType(CvUtil.getInfoTypeForStringCached('BONUS_MANA_CHAOS'))
		if iImprovement == CvUtil.getInfoTypeForStringCached('IMPROVEMENT_MANA_DEATH'):
			pPlot.setBonusType(CvUtil.getInfoTypeForStringCached('BONUS_MANA_DEATH'))
		if iImprovement == CvUtil.getInfoTypeForStringCached('IMPROVEMENT_MANA_DIMENSIONAL'):
			pPlot.setBonusType(CvUtil.getInfoTypeForStringCached('BONUS_MANA_DIMENSIONAL'))
		if iImprovement == CvUtil.getInfoTypeForStringCached('IMPROVEMENT_MANA_EARTH'):
			pPlot.setBonusType(CvUtil.getInfoTypeForStringCached('BONUS_MANA_EARTH'))
		if iImprovement == CvUtil.getInfoTypeForStringCached('IMPROVEMENT_MANA_ENCHANTMENT'):
			pPlot.setBonusType(CvUtil.getInfoTypeForStringCached('BONUS_MANA_ENCHANTMENT'))
		if iImprovement == CvUtil.getInfoTypeForStringCached('IMPROVEMENT_MANA_ENTROPY'):
			pPlot.setBonusType(CvUtil.getInfoTypeForStringCached('BONUS_MANA_ENTROPY'))
			cf.FFHChangeArmageddonCounter(1, pPlot.getOwner())
		if iImprovement == CvUtil.getInfoTypeForStringCached('IMPROVEMENT_MANA_FIRE'):
			pPlot.setBonusType(CvUtil.getInfoTypeForStringCached('BONUS_MANA_FIRE'))
		if iImprovement == CvUtil.getInfoTypeForStringCached('IMPROVEMENT_MANA_LAW'):
			pPlot.setBonusType(CvUtil.getInfoTypeForStringCached('BONUS_MANA_LAW'))
		if iImprovement == CvUtil.getInfoTypeForStringCached('IMPROVEMENT_MANA_LIFE'):
			pPlot.setBonusType(CvUtil.getInfoTypeForStringCached('BONUS_MANA_LIFE'))
			cf.FFHChangeArmageddonCounter(-1, pPlot.getOwner())
		if iImprovement == CvUtil.getInfoTypeForStringCached('IMPROVEMENT_MANA_MIND'):
			pPlot.setBonusType(CvUtil.getInfoTypeForStringCached('BONUS_MANA_MIND'))
		if iImprovement == CvUtil.getInfoTypeForStringCached('IMPROVEMENT_MANA_NATURE'):
			pPlot.setBonusType(CvUtil.getInfoTypeForStringCached('BONUS_MANA_NATURE'))
		if iImprovement == CvUtil.getInfoTypeForStringCached('IMPROVEMENT_MANA_SPIRIT'):
			pPlot.setBonusType(CvUtil.getInfoTypeForStringCached('BONUS_MANA_SPIRIT'))
		if iImprovement == CvUtil.getInfoTypeForStringCached('IMPROVEMENT_MANA_WATER'):
			pPlot.setBonusType(CvUtil.getInfoTypeForStringCached('BONUS_MANA_WATER'))

		if (not self.__LOG_IMPROVEMENT):
			return
		CvUtil.pyPrint('Improvement %s was built at %d, %d'
			%(PyInfo.ImprovementInfo(iImprovement).getDescription(), iX, iY))

	def onRouteBuilt(self, argsList):
		'Route Built'
		iRoute, iX, iY = argsList
		if (not self.__LOG_IMPROVEMENT):
			return
		CvUtil.pyPrint('Route %s was built at %d, %d'
			%(gc.getRouteInfo(iRoute).getDescription(), iX, iY))

	def onPlotRevealed(self, argsList):
		'Plot Revealed'
		pPlot = argsList[0]
		iTeam = argsList[1]

	def onBuildingBuilt(self, argsList):
		'Building Completed'
		pCity, iBuildingType = argsList
		game = CyGame()
		player = pCity.getOwner()
		pPlayer = gc.getPlayer(player)
		pPlot = pCity.plot()
		if ((not CyGame().isNetworkMultiPlayer()) and (pCity.getOwner() == CyGame().getActivePlayer()) and isWorldWonderClass(gc.getBuildingInfo(iBuildingType).getBuildingClassType())):
			if gc.getBuildingInfo(iBuildingType).getMovie():
		  		popupInfo = CyPopupInfo()
				popupInfo.setButtonPopupType(ButtonPopupTypes.BUTTONPOPUP_PYTHON_SCREEN)
				popupInfo.setData1(iBuildingType)
				popupInfo.setData2(pCity.getID())
				popupInfo.setData3(0)
				popupInfo.setText(u"showWonderMovie")
				popupInfo.addPopup(pCity.getOwner())

		if iBuildingType == CvUtil.getInfoTypeForStringCached('BUILDING_INFERNAL_GRIMOIRE'):
			if CyGame().getSorenRandNum(100, "Bob") <= 20:
				pPlot2 = cf.FFHFindClearPlot(-1, pPlot)
				if pPlot2 != -1:
					bPlayer = gc.getPlayer(gc.getBARBARIAN_PLAYER())
					newUnit = bPlayer.initUnit(CvUtil.getInfoTypeForStringCached('UNIT_BALOR'), pPlot2.getX(), pPlot2.getY(), UnitAITypes.NO_UNITAI)
					CyInterface().addMessage(pCity.getOwner(),True,25,CyTranslator().getText("TXT_KEY_MESSAGE_INFERNAL_GRIMOIRE_BALOR",()),'AS2D_BALOR',1,'Art/Interface/Buttons/Units/Balor.dds',ColorTypes(7),newUnit.getX(),newUnit.getY(),True,True)
					if pCity.getOwner() == CyGame().getActivePlayer():
						cf.FFHAddPopup(CyTranslator().getText("TXT_KEY_POPUP_INFERNAL_GRIMOIRE_BALOR",()), 'art/interface/popups/Balor.dds')

		if iBuildingType == CvUtil.getInfoTypeForStringCached('BUILDING_PACT_OF_THE_NILHORN'):
			newUnit1 = pPlayer.initUnit(CvUtil.getInfoTypeForStringCached('UNIT_HILL_GIANT'), pCity.getX(), pCity.getY(), UnitAITypes.NO_UNITAI)
			newUnit1.setName("Larry")
			newUnit2 = pPlayer.initUnit(CvUtil.getInfoTypeForStringCached('UNIT_HILL_GIANT'), pCity.getX(), pCity.getY(), UnitAITypes.NO_UNITAI)
			newUnit2.setName("Curly")
			newUnit3 = pPlayer.initUnit(CvUtil.getInfoTypeForStringCached('UNIT_HILL_GIANT'), pCity.getX(), pCity.getY(), UnitAITypes.NO_UNITAI)
			newUnit3.setName("Moe")

		if iBuildingType == CvUtil.getInfoTypeForStringCached('BUILDING_CITY'):
			pCity.setHasRealBuilding(CvUtil.getInfoTypeForStringCached('BUILDING_SETTLEMENT'), False)

		if iBuildingType == CvUtil.getInfoTypeForStringCached('BUILDING_STIGMATA_ON_THE_UNBORN'):
			cf.FFHChangeArmageddonCounter(5, pCity.getOwner())

		if iBuildingType == CvUtil.getInfoTypeForStringCached('BUILDING_ALTAR_OF_THE_LUONNOTAR_FINAL'):
			pCity.setHasRealBuilding(CvUtil.getInfoTypeForStringCached('BUILDING_ALTAR_OF_THE_LUONNOTAR_EXALTED'), False)
		if iBuildingType == CvUtil.getInfoTypeForStringCached('BUILDING_ALTAR_OF_THE_LUONNOTAR_EXALTED'):
			pCity.setHasRealBuilding(CvUtil.getInfoTypeForStringCached('BUILDING_ALTAR_OF_THE_LUONNOTAR_DIVINE'), False)
			cf.FFHWarn(player,pPlayer.getName() + " " + CyTranslator().getText("TXT_KEY_POPUP_WARN_ALTAR_OF_THE_LUONNOTAR_EXALTED",()), pPlot)
		if iBuildingType == CvUtil.getInfoTypeForStringCached('BUILDING_ALTAR_OF_THE_LUONNOTAR_DIVINE'):
			pCity.setHasRealBuilding(CvUtil.getInfoTypeForStringCached('BUILDING_ALTAR_OF_THE_LUONNOTAR_CONSECRATED'), False)
			cf.FFHWarn(player,pPlayer.getName() + " " + CyTranslator().getText("TXT_KEY_POPUP_WARN_ALTAR_OF_THE_LUONNOTAR_DIVINE",()), pPlot)
		if iBuildingType == CvUtil.getInfoTypeForStringCached('BUILDING_ALTAR_OF_THE_LUONNOTAR_CONSECRATED'):
			pCity.setHasRealBuilding(CvUtil.getInfoTypeForStringCached('BUILDING_ALTAR_OF_THE_LUONNOTAR_BLESSED'), False)
		if iBuildingType == CvUtil.getInfoTypeForStringCached('BUILDING_ALTAR_OF_THE_LUONNOTAR_BLESSED'):
			pCity.setHasRealBuilding(CvUtil.getInfoTypeForStringCached('BUILDING_ALTAR_OF_THE_LUONNOTAR_ANOINTED'), False)
		if iBuildingType == CvUtil.getInfoTypeForStringCached('BUILDING_ALTAR_OF_THE_LUONNOTAR_ANOINTED'):
			pCity.setHasRealBuilding(CvUtil.getInfoTypeForStringCached('BUILDING_ALTAR_OF_THE_LUONNOTAR'), False)

		if (iBuildingType == CvUtil.getInfoTypeForStringCached('BUILDING_TOWER_OF_ALTERATION') or iBuildingType == CvUtil.getInfoTypeForStringCached('BUILDING_TOWER_OF_DIVINATION') or iBuildingType == CvUtil.getInfoTypeForStringCached('BUILDING_TOWER_OF_THE_ELEMENTS') or iBuildingType == CvUtil.getInfoTypeForStringCached('BUILDING_TOWER_OF_NECROMANCY')):
			i = 0
			if pPlayer.getBuildingClassCount(CvUtil.getInfoTypeForStringCached('BUILDINGCLASS_TOWER_OF_ALTERATION')) > 0:
				i = i + 1
			if pPlayer.getBuildingClassCount(CvUtil.getInfoTypeForStringCached('BUILDINGCLASS_TOWER_OF_DIVINATION')) > 0:
				i = i + 1
			if pPlayer.getBuildingClassCount(CvUtil.getInfoTypeForStringCached('BUILDINGCLASS_TOWER_OF_THE_ELEMENTS')) > 0:
				i = i + 1
			if pPlayer.getBuildingClassCount(CvUtil.getInfoTypeForStringCached('BUILDINGCLASS_TOWER_OF_NECROMANCY')) > 0:
				i = i + 1
			if i == 4:
				cf.FFHWarn(player,pPlayer.getName() + " " + CyTranslator().getText("TXT_KEY_POPUP_WARN_TOWER_OF_MASTERY",()),-1)

		if iBuildingType == CvUtil.getInfoTypeForStringCached('BUILDING_MERCURIAN_GATE'):
			iPlayer = cf.FFHGetOpenPlayer()
			iCivType = CvUtil.getInfoTypeForStringCached('CIVILIZATION_MERCURIANS')
			iLeader = CvUtil.getInfoTypeForStringCached('LEADER_BASIUM')
			iTeam = pPlayer.getTeam()
			bDoGive = true
			pPlot2 = cf.FFHFindClearPlot(-1, pCity.plot())
			if (iPlayer != -1 and pPlot2 != -1):
				for i in range(pPlayer.getNumUnits(), -1, -1):
					pUnit = pPlot.getUnit(i)
					pUnit.setXY(pPlot2.getX(), pPlot2.getY())
				pPlot3 = cf.FFHFindClearPlot(-1, pPlot)
#				CyMessageControl().sendApplyEvent(5015, EventContextTypes.EVENTCONTEXT_ALL, (iPlayer, iCivType, iLeader, iTeam, bDoGive, pPlot.getX(), pPlot.getY()))
				CyGame().createNewPlayer(iPlayer, iCivType, iLeader, iTeam, bDoGive, pPlot3.getX(), pPlot3.getY())
				newUnit1 = gc.getPlayer(iPlayer).initUnit(CvUtil.getInfoTypeForStringCached('UNIT_BASIUM'), pPlot.getX(), pPlot.getY(), UnitAITypes.NO_UNITAI)
				newUnit2 = gc.getPlayer(iPlayer).initUnit(CvUtil.getInfoTypeForStringCached('UNIT_ANGEL'), pPlot.getX(), pPlot.getY(), UnitAITypes.NO_UNITAI)
				newUnit3 = gc.getPlayer(iPlayer).initUnit(CvUtil.getInfoTypeForStringCached('UNIT_ANGEL'), pPlot.getX(), pPlot.getY(), UnitAITypes.NO_UNITAI)
				newUnit4 = gc.getPlayer(iPlayer).initUnit(CvUtil.getInfoTypeForStringCached('UNIT_ANGEL'), pPlot.getX(), pPlot.getY(), UnitAITypes.NO_UNITAI)
				newUnit5 = gc.getPlayer(iPlayer).initUnit(CvUtil.getInfoTypeForStringCached('UNIT_ANGEL'), pPlot.getX(), pPlot.getY(), UnitAITypes.NO_UNITAI)
				newUnit6 = gc.getPlayer(iPlayer).initUnit(CvUtil.getInfoTypeForStringCached('UNIT_ANGEL'), pPlot.getX(), pPlot.getY(), UnitAITypes.NO_UNITAI)
				newUnit7 = gc.getPlayer(iPlayer).initUnit(CvUtil.getInfoTypeForStringCached('UNIT_ANGEL'), pPlot.getX(), pPlot.getY(), UnitAITypes.NO_UNITAI)
				if pPlayer.isHuman():
					popupInfo = CyPopupInfo()
					popupInfo.setButtonPopupType(ButtonPopupTypes.BUTTONPOPUP_PYTHON)
					popupInfo.setText(CyTranslator().getText("TXT_KEY_POPUP_CONTROL_MERCURIANS",()))
					popupInfo.setData1(player)
					popupInfo.setData2(iPlayer)
					popupInfo.addPythonButton(CyTranslator().getText("TXT_KEY_POPUP_YES", ()), "")
					popupInfo.addPythonButton(CyTranslator().getText("TXT_KEY_POPUP_NO", ()), "")
					popupInfo.setOnClickedPythonCallback("reassignPlayer")
					popupInfo.addPopup(player)

		CvAdvisorUtils.buildingBuiltFeats(pCity, iBuildingType)

		if (not self.__LOG_BUILDING):
			return
		CvUtil.pyPrint('%s was finished by Player %d Civilization %s' 
			%(PyInfo.BuildingInfo(iBuildingType).getDescription(), pCity.getOwner(), gc.getPlayer(pCity.getOwner()).getCivilizationDescription(0)))

	def onProjectBuilt(self, argsList):
		'Project Completed'
		pCity, iProjectType = argsList
		game = CyGame()
		if ((not CyGame().isNetworkMultiPlayer()) and (pCity.getOwner() == CyGame().getActivePlayer())):
			if gc.getProjectInfo(iProjectType).getMovieArtDef():
				popupInfo = CyPopupInfo()
				popupInfo.setButtonPopupType(ButtonPopupTypes.BUTTONPOPUP_PYTHON_SCREEN)
				popupInfo.setData1(iProjectType)
				popupInfo.setData2(pCity.getID())
				popupInfo.setData3(2)
				popupInfo.setText(u"showWonderMovie")
				popupInfo.addPopup(pCity.getOwner())

		if iProjectType == CvUtil.getInfoTypeForStringCached('PROJECT_BANE_DIVINE'):
			listPlayers = []
			for iPlayer in range(gc.getMAX_PLAYERS()):
				pPlayer = gc.getPlayer(iPlayer)
				if (pPlayer.isAlive()):
					listPlayers.append(iPlayer)
			for iPlayer in range(len(listPlayers)):
				pPlayer = gc.getPlayer(listPlayers[iPlayer])
				apUnitList = PyPlayer(iPlayer).getUnitList()
				for pUnit in apUnitList:
					if (pUnit.getUnitCombatType() == CvUtil.getInfoTypeForStringCached('UNITCOMBAT_DISCIPLE') and pUnit.getUnitType() != CvUtil.getInfoTypeForStringCached('UNIT_PROPHET2')):
						newUnit = pPlayer.initUnit(CvUtil.getInfoTypeForStringCached('UNIT_PROPHET2'), pUnit.getX(), pUnit.getY(), UnitAITypes.NO_UNITAI)
						pUnit.kill(False,0)

		if iProjectType == CvUtil.getInfoTypeForStringCached('PROJECT_GENESIS'):
			cf.FFHGenesis(pCity)

		if iProjectType == CvUtil.getInfoTypeForStringCached('PROJECT_GLORY_EVERLASTING'):
			for iPlayer in range(gc.getMAX_PLAYERS()):
				pPlayer = gc.getPlayer(iPlayer)
				player = PyPlayer(iPlayer) 
				if pPlayer.isAlive():
					apUnitList = player.getUnitList() 
					for pUnit in apUnitList:
						if (pUnit.isHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_DEMON')) or pUnit.isHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_UNDEAD'))):
							pUnit.kill(False,0)

		if iProjectType == CvUtil.getInfoTypeForStringCached('PROJECT_RITES_OF_OGHMA'):
			i = 7
			if CyMap().getWorldSize() == CvUtil.getInfoTypeForStringCached('WORLDSIZE_DUEL'):
				i = i - 3
			if CyMap().getWorldSize() == CvUtil.getInfoTypeForStringCached('WORLDSIZE_TINY'):
				i = i - 2
			if CyMap().getWorldSize() == CvUtil.getInfoTypeForStringCached('WORLDSIZE_SMALL'):
				i = i - 1
			if CyMap().getWorldSize() == CvUtil.getInfoTypeForStringCached('WORLDSIZE_LARGE'):
				i = i + 1
			if CyMap().getWorldSize() == CvUtil.getInfoTypeForStringCached('WORLDSIZE_HUGE'):
				i = i + 3
			cf.FFHAddBonus('BONUS_MANA',i,'Art/Interface/Buttons/WorldBuilder/mana_button.dds')

		if iProjectType == CvUtil.getInfoTypeForStringCached('PROJECT_NATURES_REVOLT'):
			bPlayer = gc.getPlayer(gc.getBARBARIAN_PLAYER())
			py = PyPlayer(gc.getBARBARIAN_PLAYER())
			iAxeman = CvUtil.getInfoTypeForStringCached('UNITCLASS_AXEMAN')
			iBear = CvUtil.getInfoTypeForStringCached('UNIT_BEAR')
			iCommando = CvUtil.getInfoTypeForStringCached('PROMOTION_COMMANDO')
			iHunter = CvUtil.getInfoTypeForStringCached('UNITCLASS_HUNTER')
			iLion = CvUtil.getInfoTypeForStringCached('UNIT_LION')
			iScout = CvUtil.getInfoTypeForStringCached('UNITCLASS_SCOUT')
			iTiger = CvUtil.getInfoTypeForStringCached('UNIT_TIGER')
			iWarrior = CvUtil.getInfoTypeForStringCached('UNITCLASS_WARRIOR')
			iWolf = CvUtil.getInfoTypeForStringCached('UNIT_WOLF')
			iWorker = CvUtil.getInfoTypeForStringCached('UNITCLASS_WORKER')
			for pUnit in py.getUnitList():
				bValid = False
				if pUnit.getUnitClassType() == iWorker:
					iNewUnit = iWolf
					bValid = True
				if pUnit.getUnitClassType() == iScout:
					iNewUnit = iLion
					bValid = True
				if pUnit.getUnitClassType() == iWarrior:
					iNewUnit = iLion
					bValid = True
				if pUnit.getUnitClassType() == iHunter:
					iNewUnit = iTiger
					bValid = True
				if pUnit.getUnitClassType() == iAxeman:
					iNewUnit = iBear
					bValid = True
				if bValid:
					newUnit = bPlayer.initUnit(iNewUnit, pUnit.getX(), pUnit.getY(), UnitAITypes.NO_UNITAI)
					newUnit.setHasPromotion(iCommando, True)
					newUnit = bPlayer.initUnit(iNewUnit, pUnit.getX(), pUnit.getY(), UnitAITypes.NO_UNITAI)
					newUnit.setHasPromotion(iCommando, True)
					newUnit = bPlayer.initUnit(iNewUnit, pUnit.getX(), pUnit.getY(), UnitAITypes.NO_UNITAI)
					newUnit.setHasPromotion(iCommando, True)
					pUnit.kill(True,0)
			for iPlayer in range(gc.getMAX_PLAYERS()):
				pPlayer = gc.getPlayer(iPlayer)
				if pPlayer.isAlive():
					py = PyPlayer(iPlayer)
					for pUnit in py.getUnitList():
						if pUnit.isAnimal():
							pUnit.changeExtraCombatStr(1)
							pUnit.changeExtraCombatDefense(1)

		if iProjectType == CvUtil.getInfoTypeForStringCached('PROJECT_ELEGY_OF_THE_SHEAIM'):
			cf.FFHChangeArmageddonCounter(5, pCity.getOwner())

		if iProjectType == CvUtil.getInfoTypeForStringCached('PROJECT_HALLOWING_OF_THE_ELOHIM'):
			cf.FFHChangeArmageddonCounter(-5, pCity.getOwner())

		if iProjectType == CvUtil.getInfoTypeForStringCached('PROJECT_BLOOD_OF_THE_PHOENIX'):
			py = PyPlayer(pCity.getOwner())
			apUnitList = py.getUnitList()
			for pUnit in apUnitList:
				if pUnit.isAlive():
					pUnit.setHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_IMMORTAL'), True)
		if iProjectType == CvUtil.getInfoTypeForStringCached('PROJECT_PURGE_THE_UNFAITHFUL'):
			pPlayer = gc.getPlayer(pCity.getOwner())
			for i in range(pPlayer.getNumCities()):
				iRnd = CyGame().getSorenRandNum(2, "Bob")
				pCity2 = pPlayer.getCity(i)
				StateBelief = pPlayer.getStateReligion()
				if StateBelief == CvUtil.getInfoTypeForStringCached('RELIGION_THE_ORDER'):
					iRnd = iRnd - 1
				for iTarget in range(gc.getNumReligionInfos()):
					if (StateBelief != iTarget and pCity2.isHasReligion(iTarget) and pCity2.isHolyCityByType(iTarget) == False):
						pCity2.setHasReligion(iTarget, False, True, True)
						iRnd = iRnd + 1
						for i in range(gc.getNumBuildingInfos()):
							if gc.getBuildingInfo(i).getPrereqReligion() == iTarget:
								pCity2.setHasRealBuilding(i, False)
				if iRnd > 0:
					pCity2.setOccupationTimer(iRnd)


	def onUnitMove(self, argsList):
		'unit move'
		pPlot,pUnit = argsList
		player = PyPlayer(pUnit.getOwner())
		unitInfo = PyInfo.UnitInfo(pUnit.getUnitType())
		pPlayer = gc.getPlayer(pUnit.getOwner())

		if (not self.__LOG_MOVEMENT):
			return
		if player and unitInfo:
			CvUtil.pyPrint('Player %d Civilization %s unit %s is moving to %d, %d' 
				%(player.getID(), player.getCivilizationName(), unitInfo.getDescription(), 
				pUnit.getX(), pUnit.getY()))

	def onUnitSetXY(self, argsList):
		'units xy coords set manually'
		pPlot,pUnit = argsList
		player = PyPlayer(pUnit.getOwner())
		unitInfo = PyInfo.UnitInfo(pUnit.getUnitType())
		if (not self.__LOG_MOVEMENT):
			return

	def onUnitCreated(self, argsList):
		'Unit Completed'
		unit = argsList[0]
		player = PyPlayer(unit.getOwner())
		pPlayer = gc.getPlayer(unit.getOwner())
		unitInfo = PyInfo.UnitInfo(unit.getUnitType())

		if unit.getUnitCombatType() == CvUtil.getInfoTypeForStringCached('UNITCOMBAT_ANIMAL'):
			unit.changeExtraCombatStr(gc.getDefineINT('STR_MODIFIER_ANIMAL'))
			unit.changeExtraCombatDefense(gc.getDefineINT('STR_MODIFIER_ANIMAL'))

		if pPlayer.getCivilizationType() == CvUtil.getInfoTypeForStringCached('CIVILIZATION_MALAKIM'):
			if unit.isAlive():
				unit.setHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_NOMAD'), True)

		if unit.getOwner() == gc.getBARBARIAN_PLAYER():
			if unit.getUnitCombatType() == CvUtil.getInfoTypeForStringCached('UNITCOMBAT_MELEE'):
				eTeam = gc.getTeam(pPlayer.getTeam())
				if eTeam.isHasTech(CvUtil.getInfoTypeForStringCached('TECH_BRONZE_WORKING')):
					unit.setHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_BRONZE_WEAPONS'), True)
				if eTeam.isHasTech(CvUtil.getInfoTypeForStringCached('TECH_IRON_WORKING')):
					unit.setHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_IRON_WEAPONS'), True)
					unit.setHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_BRONZE_WEAPONS'), False)
				if eTeam.isHasTech(CvUtil.getInfoTypeForStringCached('TECH_MITHRIL_WORKING')):
					unit.setHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_MITHRIL_WEAPONS'), True)
					unit.setHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_IRON_WEAPONS'), False)

		if (not self.__LOG_UNITBUILD):
			return

	def onUnitBuilt(self, argsList):
		'Unit Completed'
		city = argsList[0]
		unit = argsList[1]
		player = PyPlayer(city.getOwner())
		pPlayer = gc.getPlayer(unit.getOwner())

		if city.hasBonus(CvUtil.getInfoTypeForStringCached('BONUS_SHEUT_STONE')):
			if (unit.getUnitClassType() != CvUtil.getInfoTypeForStringCached('UNITCLASS_WORKER') and unit.getUnitClassType() != CvUtil.getInfoTypeForStringCached('UNITCLASS_SETTLER') and unit.getUnitClassType() != CvUtil.getInfoTypeForStringCached('UNITCLASS_WORKBOAT')):
				unit.setHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_SHEUT_STONE'), True)

		if unit.getUnitType() == CvUtil.getInfoTypeForStringCached('UNIT_BEAST_OF_AGARES'):
			iPop = city.getPopulation() - 4
			if iPop <= 1:
				iPop = 1
			city.setPopulation(iPop)
			city.setOccupationTimer(4)

		if city.hasBuilding(CvUtil.getInfoTypeForStringCached('BUILDING_CHANCEL_OF_GUARDIANS')):
			iRnd = CyGame().getSorenRandNum(100, "Bob")
			if iRnd <= 20:
				unit.setHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_DEFENSIVE'), True)

		if city.hasBuilding(CvUtil.getInfoTypeForStringCached('BUILDING_RELIQUARY')):
			unit.setHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_SPIRIT_GUIDE'), True)

		if (city.hasBuilding(CvUtil.getInfoTypeForStringCached('BUILDING_CAVE_OF_ANCESTORS')) and unit.getUnitCombatType() == (CvUtil.getInfoTypeForStringCached('UNITCOMBAT_ADEPT'))):
			i = 0
			for iBonus in manaTypes:
				if city.hasBonus(CvUtil.getInfoTypeForStringCached(iBonus)):
					i = i + 1
			if i >= 1:
				unit.changeExperience(i,-1)

		if unit.getUnitCombatType() == CvUtil.getInfoTypeForStringCached('UNITCOMBAT_ADEPT'):
			iNum = city.getNumBonuses(CvUtil.getInfoTypeForStringCached('BONUS_MANA_AIR'))
			if iNum > 1:
				unit.setHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_AIR1'), True)
				if iNum > 2:
					unit.setHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_AIR2'), True)
			iNum = city.getNumBonuses(CvUtil.getInfoTypeForStringCached('BONUS_MANA_BODY'))
			if iNum > 1:
				unit.setHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_BODY1'), True)
				if iNum > 2:
					unit.setHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_BODY2'), True)
			iNum = city.getNumBonuses(CvUtil.getInfoTypeForStringCached('BONUS_MANA_CHAOS'))
			if iNum > 1:
				unit.setHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_CHAOS1'), True)
				if iNum > 2:
					unit.setHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_CHAOS2'), True)
			iNum = city.getNumBonuses(CvUtil.getInfoTypeForStringCached('BONUS_MANA_DEATH'))
			if iNum > 1:
				unit.setHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_DEATH1'), True)
				if iNum > 2:
					unit.setHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_DEATH2'), True)
			iNum = city.getNumBonuses(CvUtil.getInfoTypeForStringCached('BONUS_MANA_DIMENSIONAL'))
			if iNum > 1:
				unit.setHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_DIMENSIONAL1'), True)
			iNum = city.getNumBonuses(CvUtil.getInfoTypeForStringCached('BONUS_MANA_EARTH'))
			if iNum > 1:
				unit.setHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_EARTH1'), True)
				if iNum > 2:
					unit.setHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_EARTH2'), True)
			iNum = city.getNumBonuses(CvUtil.getInfoTypeForStringCached('BONUS_MANA_ENCHANTMENT'))
			if iNum > 1:
				unit.setHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_ENCHANTMENT1'), True)
				if iNum > 2:
					unit.setHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_ENCHANTMENT2'), True)
			iNum = city.getNumBonuses(CvUtil.getInfoTypeForStringCached('BONUS_MANA_ENTROPY'))
			if iNum > 1:
				unit.setHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_ENTROPY1'), True)
				if iNum > 2:
					unit.setHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_ENTROPY2'), True)
			iNum = city.getNumBonuses(CvUtil.getInfoTypeForStringCached('BONUS_MANA_FIRE'))
			if iNum > 1:
				unit.setHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_FIRE1'), True)
				if iNum > 2:
					unit.setHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_FIRE2'), True)
			iNum = city.getNumBonuses(CvUtil.getInfoTypeForStringCached('BONUS_MANA_LAW'))
			if iNum > 1:
				unit.setHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_LAW1'), True)
				if iNum > 2:
					unit.setHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_LAW2'), True)
			iNum = city.getNumBonuses(CvUtil.getInfoTypeForStringCached('BONUS_MANA_LIFE'))
			if iNum > 1:
				unit.setHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_LIFE1'), True)
				if iNum > 2:
					unit.setHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_LIFE2'), True)
			iNum = city.getNumBonuses(CvUtil.getInfoTypeForStringCached('BONUS_MANA_MIND'))
			if iNum > 1:
				unit.setHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_MIND1'), True)
				if iNum > 2:
					unit.setHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_MIND2'), True)
			iNum = city.getNumBonuses(CvUtil.getInfoTypeForStringCached('BONUS_MANA_NATURE'))
			if iNum > 1:
				unit.setHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_NATURE1'), True)
				if iNum > 2:
					unit.setHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_NATURE2'), True)
			iNum = city.getNumBonuses(CvUtil.getInfoTypeForStringCached('BONUS_MANA_SPIRIT'))
			if iNum > 1:
				unit.setHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_SPIRIT1'), True)
				if iNum > 2:
					unit.setHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_SPIRIT2'), True)
			iNum = city.getNumBonuses(CvUtil.getInfoTypeForStringCached('BONUS_MANA_WATER'))
			if iNum > 1:
				unit.setHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_WATER1'), True)
				if iNum > 2:
					unit.setHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_WATER2'), True)

		if unit.isHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_GOLEM')):
			if city.hasBuilding(CvUtil.getInfoTypeForStringCached('BUILDING_BLASTING_WORKSHOP')):
				unit.setHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_FIRE2'), True)
				unit.setHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_SORCERY'), True)
			if city.hasBuilding(CvUtil.getInfoTypeForStringCached('BUILDING_ARMAMENT_MOLDS')):
				unit.setHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_HEAVY'), True)
			if city.hasBuilding(CvUtil.getInfoTypeForStringCached('BUILDING_VELOX_WORKSHOP')):
				unit.setHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_LIGHT'), True)
			if city.hasBuilding(CvUtil.getInfoTypeForStringCached('BUILDING_GRIDLINES')):
				unit.setHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_CITY_GARRISON1'), True)

		if (city.hasBuilding(CvUtil.getInfoTypeForStringCached('BUILDING_ASYLUM')) and unit.isAlive()):
			if CyGame().getSorenRandNum(100, "Bob") <= 25:
				unit.setHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_CRAZED'), True)
				unit.setHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_ENRAGED'), True)

		if cf.FFHT4Units(unit):
			unit.setName(unit.getName() + " of " + city.getName())

		if unit.getUnitType() == CvUtil.getInfoTypeForStringCached('UNIT_ACHERON'):
			city.setHasRealBuilding(CvUtil.getInfoTypeForStringCached('BUILDING_THE_DRAGONS_HORDE'), True)
			iX = city.getX()
			iY = city.getY()
			for iiX in range(iX-1, iX+2, 1):
				for iiY in range(iY-1, iY+2, 1):
					pPlot = CyMap().plot(iiX,iiY)
					if (pPlot.getFeatureType() == CvUtil.getInfoTypeForStringCached('FEATURE_FOREST') or pPlot.getFeatureType() == CvUtil.getInfoTypeForStringCached('FEATURE_JUNGLE')):
						pPlot.setFeatureType(CvUtil.getInfoTypeForStringCached('FEATURE_FLAMES'), 0)
			cf.FFHAddPopup(CyTranslator().getText("TXT_KEY_POPUP_ACHERON_CREATION",()), 'art/interface/popups/Acheron the Red Dragon.dds')

		if pPlayer.getCivics(CvUtil.getInfoTypeForStringCached('CIVICOPTION_CULTURAL_VALUES')) == CvUtil.getInfoTypeForStringCached('CIVIC_CRUSADE'):
			if unit.getUnitCombatType() == CvUtil.getInfoTypeForStringCached('UNITCOMBAT_DISCIPLE'):
				unit.setHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_EVANGELIST'), True)

		if (pPlayer.getBuildingClassCount(CvUtil.getInfoTypeForStringCached('BUILDINGCLASS_ALTAR_OF_THE_LUONNOTAR_BLESSED')) > 0 or pPlayer.getBuildingClassCount(CvUtil.getInfoTypeForStringCached('BUILDINGCLASS_ALTAR_OF_THE_LUONNOTAR_CONSECRATED')) or pPlayer.getBuildingClassCount(CvUtil.getInfoTypeForStringCached('BUILDINGCLASS_ALTAR_OF_THE_LUONNOTAR_DIVINE')) or pPlayer.getBuildingClassCount(CvUtil.getInfoTypeForStringCached('BUILDINGCLASS_ALTAR_OF_THE_LUONNOTAR_EXALTED')) or pPlayer.getBuildingClassCount(CvUtil.getInfoTypeForStringCached('BUILDINGCLASS_ALTAR_OF_THE_LUONNOTAR_FINAL'))):
			if unit.isAlive():
				unit.setHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_BLESSED'), True)

		if unit.isHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_DRAGON')):
			unit.changeExtraCombatStr(gc.getDefineINT('STR_MODIFIER_DRAGON'))
			unit.changeExtraCombatDefense(gc.getDefineINT('STR_MODIFIER_DRAGON'))

		if pPlayer.isHuman():
			unit.changeExtraCombatStr(gc.getDefineINT('STR_MODIFIER_HUMAN'))
			unit.changeExtraCombatDefense(gc.getDefineINT('STR_MODIFIER_HUMAN'))
		else:
			unit.changeExtraCombatStr(gc.getDefineINT('STR_MODIFIER_AI'))
			unit.changeExtraCombatDefense(gc.getDefineINT('STR_MODIFIER_AI'))

		if unit.isHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_DWARF')):
			if city.hasBuilding(CvUtil.getInfoTypeForStringCached('BUILDING_BREWERY')):
				unit.changeExperience(2,-1)

		if unit.isHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_DEMON')):
			if city.hasBuilding(CvUtil.getInfoTypeForStringCached('BUILDING_DEMONS_ALTAR')):
				unit.changeExperience(2,-1)

		if city.hasBuilding(CvUtil.getInfoTypeForStringCached('BUILDING_STIGMATA_ON_THE_UNBORN')):
			if unit.isAlive():
				unit.setHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_STIGMATA'), True)

		if unit.isAlive():
			if (unit.isHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_THE_ORDER')) == False and unit.isHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_RUNES_OF_KILMORPH')) == False and unit.isHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_FELLOWSHIP_OF_LEAVES')) == False and unit.isHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_OCTOPUS_OVERLORDS')) == False and unit.isHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_THE_ASHEN_VEIL')) == False and unit.isHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_CULT_OF_THE_DRAGON')) == False):
				listReligions = []
				for i in range(gc.getNumReligionInfos()):
					if city.isHasReligion(i):
						listReligions.append(i)
				if len(listReligions) > 0:
					iReligion = listReligions[CyGame().getSorenRandNum(len(listReligions), "Pick Religion")]
					i = 20
					if pPlayer.getStateReligion() == iReligion:
						i = i + 10
					if city.isHolyCityByType(iReligion):
						i = i + 10
					if (CyGame().getSorenRandNum(100, "Give Religion") <= i and unit.getUnitCombatType() != CvUtil.getInfoTypeForStringCached('UNITCOMBAT_DISCIPLE')):
						if iReligion == CvUtil.getInfoTypeForStringCached('RELIGION_THE_ORDER'):
							unit.setHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_THE_ORDER'), True)
						if iReligion == CvUtil.getInfoTypeForStringCached('RELIGION_RUNES_OF_KILMORPH'):
							unit.setHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_RUNES_OF_KILMORPH'), True)
						if iReligion == CvUtil.getInfoTypeForStringCached('RELIGION_FELLOWSHIP_OF_LEAVES'):
							unit.setHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_FELLOWSHIP_OF_LEAVES'), True)
						if iReligion == CvUtil.getInfoTypeForStringCached('RELIGION_OCTOPUS_OVERLORDS'):
							unit.setHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_OCTOPUS_OVERLORDS'), True)
						if iReligion == CvUtil.getInfoTypeForStringCached('RELIGION_THE_ASHEN_VEIL'):
							unit.setHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_THE_ASHEN_VEIL'), True)
						if iReligion == CvUtil.getInfoTypeForStringCached('RELIGION_CULT_OF_THE_DRAGON'):
							unit.setHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_CULT_OF_THE_DRAGON'), True)

		if (unit.getUnitType() == CvUtil.getInfoTypeForStringCached('UNIT_DISCIPLE_FELLOWSHIP_OF_LEAVES') or unit.getUnitType() == CvUtil.getInfoTypeForStringCached('UNIT_PRIEST_OF_LEAVES')):
			if CyGame().getSorenRandNum(100, "Elven") <= 25:
				unit.setHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_ELF'), True)

		if (unit.getUnitType() == CvUtil.getInfoTypeForStringCached('UNIT_DISCIPLE_RUNES_OF_KILMORPH') or unit.getUnitType() == CvUtil.getInfoTypeForStringCached('UNIT_PRIEST_OF_KILMORPH')):
			if CyGame().getSorenRandNum(100, "Dwarven") <= 25:
				unit.setHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_DWARF'), True)

		if unit.isHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_PROPHECY_MARK')):
			cf.FFHChangeArmageddonCounter(1, unit.getOwner())

		if (pPlayer.getCivilizationType() == CvUtil.getInfoTypeForStringCached('CIVILIZATION_KHAZAD') or pPlayer.getCivilizationType() == CvUtil.getInfoTypeForStringCached('CIVILIZATION_LUCHUIRP')):
			if (unit.isAlive() and cf.FFHHasRace(unit) == False and isWorldUnitClass(unit.getUnitClassType()) == False):
				unit.setHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_DWARF'), True)

		if (pPlayer.getCivilizationType() == CvUtil.getInfoTypeForStringCached('CIVILIZATION_LJOSALFAR') or pPlayer.getCivilizationType() == CvUtil.getInfoTypeForStringCached('CIVILIZATION_SVARTALFAR')):
			if (unit.isAlive() and cf.FFHHasRace(unit) == False and isWorldUnitClass(unit.getUnitClassType()) == False):
				unit.setHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_ELF'), True)
		if pPlayer.getCivilizationType() == CvUtil.getInfoTypeForStringCached('CIVILIZATION_INFERNAL'):
			if (unit.isAlive() and cf.FFHHasRace(unit) == False and isWorldUnitClass(unit.getUnitClassType()) == False):
				unit.setHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_DEMON'), True)

		if pPlayer.getCivilizationType() == CvUtil.getInfoTypeForStringCached('CIVILIZATION_CLAN_OF_EMBERS'):
			if (unit.isAlive() and cf.FFHHasRace(unit) == False and isWorldUnitClass(unit.getUnitClassType()) == False):
				unit.setHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_ORC'), True)


		if city.hasBonus(CvUtil.getInfoTypeForStringCached('BONUS_NIGHTMARE')):
			for i in range(4):
				if gc.getUnitInfo(unit.getUnitType()).getPrereqOrBonuses(i) == CvUtil.getInfoTypeForStringCached('BONUS_NIGHTMARE'):
					unit.setHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_NIGHTMARE'), true)

		if (unit.isAlive() and pPlayer.getNumAvailableBonuses(CvUtil.getInfoTypeForStringCached('BONUS_MANA_CHAOS')) > 0):
			if CyGame().getSorenRandNum(100, "Mutate") <= (pPlayer.getNumAvailableBonuses(CvUtil.getInfoTypeForStringCached('BONUS_MANA_CHAOS')) * gc.getDefineINT('MANA_CHAOS_MUTATE_CHANCE')):
				cf.FFHMutate(unit)

		if (unit.getUnitType() == CvUtil.getInfoTypeForStringCached('UNIT_ABASHI') or unit.getUnitType() == CvUtil.getInfoTypeForStringCached('UNIT_EURABATRES')):
			if CyGame().isReligionFounded(CvUtil.getInfoTypeForStringCached('RELIGION_CULT_OF_THE_DRAGON')) == False:
				city.setHasReligion(CvUtil.getInfoTypeForStringCached('RELIGION_CULT_OF_THE_DRAGON'), true, true, true)
				CyGame().setHolyCity(CvUtil.getInfoTypeForStringCached('RELIGION_CULT_OF_THE_DRAGON'), city, True)

		CvAdvisorUtils.unitBuiltFeats(city, unit)

		if (not self.__LOG_UNITBUILD):
			return
		CvUtil.pyPrint('%s was finished by Player %d Civilization %s' 
			%(PyInfo.UnitInfo(unit.getUnitType()).getDescription(), player.getID(), player.getCivilizationName()))
	
	def onUnitKilled(self, argsList):
		'Unit Killed'
		unit, iAttacker = argsList
		player = PyPlayer(unit.getOwner())
		attacker = PyPlayer(iAttacker)
		pPlot = unit.plot()
		pPlayer = gc.getPlayer(unit.getOwner())

		if unit.isHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_IMMORTAL')):
			cf.FFHImmortal(unit)

		if unit.isHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_SPIRIT_GUIDE')):
			if unit.getExperience() > 0:
				iRnd = CyGame().getSorenRandNum(pPlayer.getNumUnits(), "Spirit Guide")
				pUnit = pPlayer.getUnit(iRnd)
				iXP = unit.getExperience() / 2
				pUnit.changeExperience(iXP, -1)
				unit.changeExperience(iXP * -1, -1)
				CyInterface().addMessage(unit.getOwner(),True,25,CyTranslator().getText("TXT_KEY_MESSAGE_SPIRIT_GUIDE",()),'AS2D_DISCOVERBONUS',1,'Art/Interface/Buttons/Promotions/SpiritGuide.dds',ColorTypes(7),pUnit.getX(),pUnit.getY(),True,True)

		if unit.isAlive():
			iX = unit.getX()
			iY = unit.getY()
			for iiX in range(iX-1, iX+2, 1):
				for iiY in range(iY-1, iY+2, 1):
					pPlot2 = CyMap().plot(iiX,iiY)
					if pPlot2.isCity():
						pCity = pPlot2.getPlotCity()
						if pCity.hasBuilding(CvUtil.getInfoTypeForStringCached('BUILDING_SOUL_FORGE')):
							pCity.changeProduction(unit.getExperience() + 10)
							CyInterface().addMessage(pCity.getOwner(),True,25,CyTranslator().getText("TXT_KEY_MESSAGE_SOUL_FORGE",()),'AS2D_DISCOVERBONUS',1,'Art/Interface/Buttons/Buildings/Soulforge.dds',ColorTypes(7),pCity.getX(),pCity.getY(),True,True)
			pPlot2 = CyMap().plot(iX,iY)
			if pPlot2.isCity():
				pCity = pPlot2.getPlotCity()
				if (pCity.hasBuilding(CvUtil.getInfoTypeForStringCached('BUILDING_MOKKAS_CAULDRON')) and pCity.getOwner() == unit.getOwner()):
					iUnit = CvUtil.getInfoTypeForStringCached('UNIT_FLESH_GOLEM')
					newUnit = pPlayer.initUnit(iUnit, pCity.getX(), pCity.getY(), UnitAITypes.NO_UNITAI)
					newUnit.changeExtraCombatStr(unit.baseCombatStr(True) - newUnit.baseCombatStr(True))
					newUnit.changeExtraCombatDefense(unit.baseCombatStr(False) - newUnit.baseCombatStr(False))
					newUnit.setDamage(gc.getMAX_HIT_POINTS()/2, True)# Smarter Orcs defect fix.
					CyInterface().addMessage(unit.getOwner(),True,25,CyTranslator().getText("TXT_KEY_MESSAGE_MOKKAS_CAULDRON",()),'AS2D_DISCOVERBONUS',1,'Art/Interface/Buttons/Buildings/Mokkas Cauldron.dds',ColorTypes(7),pCity.getX(),pCity.getY(),True,True)
			if (unit.isHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_THE_ASHEN_VEIL')) or unit.isHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_OCTOPUS_OVERLORDS')) or unit.isHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_DEATH1')) or unit.isHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_ENTROPY1'))):
				cf.FFHGiftUnit(CvUtil.getInfoTypeForStringCached('UNIT_MANES'), CvUtil.getInfoTypeForStringCached('CIVILIZATION_INFERNAL'), 0, unit.plot(), unit.getOwner())
				cf.FFHGiftUnit(CvUtil.getInfoTypeForStringCached('UNIT_MANES'), CvUtil.getInfoTypeForStringCached('CIVILIZATION_INFERNAL'), 0, unit.plot(), unit.getOwner())
			if (unit.isHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_THE_ORDER')) or unit.isHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_RUNES_OF_KILMORPH')) or pPlayer.getCivilizationType() == CvUtil.getInfoTypeForStringCached('CIVILIZATION_MERCURIANS')):
				cf.FFHGiftUnit(CvUtil.getInfoTypeForStringCached('UNIT_ANGEL'), CvUtil.getInfoTypeForStringCached('CIVILIZATION_MERCURIANS'), unit.getExperience(), unit.plot(), unit.getOwner())

		if unit.getUnitType() == CvUtil.getInfoTypeForStringCached('UNIT_BASIUM'):
			cf.FFHChangeArmageddonCounter(5, iAttacker)
		if unit.getUnitType() == CvUtil.getInfoTypeForStringCached('UNIT_HYBOREM'):
			cf.FFHChangeArmageddonCounter(-5, iAttacker)
		if unit.getUnitType() == CvUtil.getInfoTypeForStringCached('UNIT_VALIN'):
			cf.FFHChangeArmageddonCounter(2, iAttacker)
		if unit.getUnitType() == CvUtil.getInfoTypeForStringCached('UNIT_SPHENER'):
			cf.FFHChangeArmageddonCounter(3, iAttacker)
		if unit.getUnitType() == CvUtil.getInfoTypeForStringCached('UNIT_ROSIER'):
			cf.FFHChangeArmageddonCounter(-2, iAttacker)
		if unit.getUnitType() == CvUtil.getInfoTypeForStringCached('UNIT_MARDERO'):
			cf.FFHChangeArmageddonCounter(-3, iAttacker)
		if unit.getUnitType() == CvUtil.getInfoTypeForStringCached('UNIT_WRATH'):
			cf.FFHChangeArmageddonCounter(-3, iAttacker)
		if unit.getUnitType() == CvUtil.getInfoTypeForStringCached('UNIT_MESHABBER'):
			cf.FFHChangeArmageddonCounter(-3, iAttacker)

		if unit.getUnitType() == CvUtil.getInfoTypeForStringCached('UNIT_STEPHANOS'):
			cf.FFHChangeArmageddonCounter(-2, iAttacker)
			if pPlot.getBonusType(-1) == -1:
				pPlot.setBonusType(CvUtil.getInfoTypeForStringCached('BONUS_MANA_MIND'))
		if unit.getUnitType() == CvUtil.getInfoTypeForStringCached('UNIT_BUBOES'):
			cf.FFHChangeArmageddonCounter(-2, iAttacker)
			if pPlot.getBonusType(-1) == -1:
				pPlot.setBonusType(CvUtil.getInfoTypeForStringCached('BONUS_MANA_SPIRIT'))
		if unit.getUnitType() == CvUtil.getInfoTypeForStringCached('UNIT_YERSINIA'):
			cf.FFHChangeArmageddonCounter(-2, iAttacker)
			if pPlot.getBonusType(-1) == -1:
				pPlot.setBonusType(CvUtil.getInfoTypeForStringCached('BONUS_MANA_BODY'))
		if unit.getUnitType() == CvUtil.getInfoTypeForStringCached('UNIT_ARS'):
			cf.FFHChangeArmageddonCounter(-2, iAttacker)
			if pPlot.getBonusType(-1) == -1:
				pPlot.setBonusType(CvUtil.getInfoTypeForStringCached('BONUS_MANA_DEATH'))

		if unit.isHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_PROPHECY_MARK')):
			cf.FFHChangeArmageddonCounter(-1, iAttacker)

		if (not self.__LOG_UNITKILLED):
			return
		CvUtil.pyPrint('Player %d Civilization %s Unit %s was killed by Player %d' 
			%(player.getID(), player.getCivilizationName(), PyInfo.UnitInfo(unit.getUnitType()).getDescription(), attacker.getID()))

	def onUnitLost(self, argsList):
		'Unit Lost'
		unit = argsList[0]
		player = PyPlayer(unit.getOwner())
		if (not self.__LOG_UNITLOST):
			return
		CvUtil.pyPrint('%s was lost by Player %d Civilization %s' 
			%(PyInfo.UnitInfo(unit.getUnitType()).getDescription(), player.getID(), player.getCivilizationName()))
	
	def onUnitPromoted(self, argsList):
		'Unit Promoted'
		pUnit, iPromotion = argsList
		player = pUnit.getOwner()
		pPlayer = gc.getPlayer(player)

		if (not self.__LOG_UNITPROMOTED):
			return
		CvUtil.pyPrint('Unit Promotion Event: %s - %s' %(player.getCivilizationName(), pUnit.getName(),))

	def onUnitSelected(self, argsList):
		'Unit Selected'
		unit = argsList[0]
		player = PyPlayer(unit.getOwner())
		if (not self.__LOG_UNITSELECTED):
			return
		CvUtil.pyPrint('%s was selected by Player %d Civilization %s' 
			%(PyInfo.UnitInfo(unit.getUnitType()).getDescription(), player.getID(), player.getCivilizationName()))
	
	def onUnitRename(self, argsList):
		'Unit is renamed'
		pUnit = argsList[0]
		if (pUnit.getOwner() == CyGame().getActivePlayer()):
			self.__eventEditUnitNameBegin(pUnit)

	def onGoodyReceived(self, argsList):
		'Goody received'
		iPlayer, pPlot, pUnit, iGoodyType = argsList
		if (not self.__LOG_GOODYRECEIVED):
			return
		CvUtil.pyPrint('%s received a goody' %(gc.getPlayer(iPlayer).getCivilizationDescription(0)),)
	
	def onGreatPersonBorn(self, argsList):
		'Unit Promoted'
		pUnit, iPlayer, pCity = argsList
		player = PyPlayer(iPlayer)
		if pUnit.isNone() or pCity.isNone():
			return
		if (not self.__LOG_GREATPERSON):
			return
		CvUtil.pyPrint('A %s was born for %s in %s' %(pUnit.getName(), player.getCivilizationName(), pCity.getName()))
	
	def onTechAcquired(self, argsList):
		'Tech Acquired'
		iTechType, iTeam, iPlayer, bAnnounce = argsList
		# Note that iPlayer may be NULL (-1) and not a refer to a player object

		# Show tech splash when applicable
		if (iPlayer > -1 and bAnnounce and not CyInterface().noTechSplash()):
			if (gc.getGame().isFinalInitialized() and not gc.getGame().GetWorldBuilderMode()):
				if ((not CyGame().isNetworkMultiPlayer()) and (iPlayer == CyGame().getActivePlayer())):
					popupInfo = CyPopupInfo()
					popupInfo.setButtonPopupType(ButtonPopupTypes.BUTTONPOPUP_PYTHON_SCREEN)
					popupInfo.setData1(iTechType)
					popupInfo.setText(u"showTechSplash")
					popupInfo.addPopup(iPlayer)

		if (iPlayer != -1 and iPlayer != gc.getBARBARIAN_PLAYER()):
			pPlayer = gc.getPlayer(iPlayer)
			iReligion = -1
			if iTechType == CvUtil.getInfoTypeForStringCached('TECH_CORRUPTION_OF_SPIRIT'):
				iUnit = CvUtil.getInfoTypeForStringCached('UNIT_DISCIPLE_THE_ASHEN_VEIL')
				iReligion = CvUtil.getInfoTypeForStringCached('RELIGION_THE_ASHEN_VEIL')
			if iTechType == CvUtil.getInfoTypeForStringCached('TECH_ORDERS_FROM_HEAVEN'):
				iUnit = CvUtil.getInfoTypeForStringCached('UNIT_DISCIPLE_THE_ORDER')
				iReligion = CvUtil.getInfoTypeForStringCached('RELIGION_THE_ORDER')
			if iTechType == CvUtil.getInfoTypeForStringCached('TECH_WAY_OF_THE_FORESTS'):
				iUnit = CvUtil.getInfoTypeForStringCached('UNIT_DISCIPLE_FELLOWSHIP_OF_LEAVES')
				iReligion = CvUtil.getInfoTypeForStringCached('RELIGION_FELLOWSHIP_OF_LEAVES')
			if iTechType == CvUtil.getInfoTypeForStringCached('TECH_WAY_OF_THE_EARTHMOTHER'):
				iUnit = CvUtil.getInfoTypeForStringCached('UNIT_DISCIPLE_RUNES_OF_KILMORPH')
				iReligion = CvUtil.getInfoTypeForStringCached('RELIGION_RUNES_OF_KILMORPH')
			if iTechType == CvUtil.getInfoTypeForStringCached('TECH_MESSAGE_FROM_THE_DEEP'):
				iUnit = CvUtil.getInfoTypeForStringCached('UNIT_DISCIPLE_OCTOPUS_OVERLORDS')
				iReligion = CvUtil.getInfoTypeForStringCached('RELIGION_OCTOPUS_OVERLORDS')
			if iReligion != -1:
				if CyGame().isReligionFounded(iReligion):
					cf.FFHGiftUnit(iUnit, pPlayer.getCivilizationType(), 0, -1, -1)

		iCount = 0
		for iTeam in range(gc.getMAX_TEAMS()):
			eTeam = gc.getTeam(iTeam)
			if eTeam.isHasTech(CvUtil.getInfoTypeForStringCached('TECH_INFERNAL_PACT')):
				iCount = iCount + 1

		if (iTechType == CvUtil.getInfoTypeForStringCached('TECH_INFERNAL_PACT') and iPlayer != -1 and iCount == 1):
			iInfernalPlayer = cf.FFHGetOpenPlayer()
			iCivType = CvUtil.getInfoTypeForStringCached('CIVILIZATION_INFERNAL')
			iLeader = CvUtil.getInfoTypeForStringCached('LEADER_HYBOREM')
			iTeam = -1
			bDoGive = true
			pBestPlot = -1
			iBestPlot = -1
			for i in range (CyMap().numPlots()):
				pPlot = CyMap().plotByIndex(i)
				iPlot = -1
				if pPlot.isWater() == False:
					if pPlot.getNumUnits() == 0:
						if pPlot.isCity() == False:
							if pPlot.isImpassable() == False:
								iPlot = CyGame().getSorenRandNum(500, "Place Hyborem")
								iPlot = iPlot + (pPlot.area().getNumTiles() * 2)
								iPlot = iPlot + (pPlot.area().getNumUnownedTiles() * 10)
								if pPlot.isOwned() == False:
									iPlot = iPlot + 500
								if pPlot.getOwner() == iPlayer:
									iPlot = iPlot + 200
				if iPlot > iBestPlot:
					iBestPlot = iPlot
					pBestPlot = pPlot
			if (iPlayer != -1 and pBestPlot != -1):
#				CyMessageControl().sendApplyEvent(5015, EventContextTypes.EVENTCONTEXT_ALL, (iPlayer, iCivType, iLeader, iTeam, bDoGive, pPlot.getX(), pPlot.getY()))
				CyGame().createNewPlayer(iInfernalPlayer, iCivType, iLeader, iTeam, bDoGive, pBestPlot.getX(), pBestPlot.getY())
				iInfernalTeam = gc.getPlayer(iInfernalPlayer).getTeam()
				iFounderTeam = gc.getPlayer(iPlayer).getTeam()
				eInfernalTeam = gc.getTeam(gc.getPlayer(iInfernalPlayer).getTeam())
				eFounderTeam = gc.getTeam(gc.getPlayer(iPlayer).getTeam())
				for iTech in range(gc.getNumTechInfos()):
					if eFounderTeam.isHasTech(iTech):
						eInfernalTeam.setHasTech(iTech, true, iInfernalPlayer, true, false)
				eFounderTeam.signOpenBorders(iFounderTeam)
				eInfernalTeam.signOpenBorders(iInfernalTeam)
				iBarbTeam = gc.getPlayer(gc.getBARBARIAN_PLAYER()).getTeam()
				eInfernalTeam.makePeace(iBarbTeam)
				newUnit1 = gc.getPlayer(iInfernalPlayer).initUnit(CvUtil.getInfoTypeForStringCached('UNIT_HYBOREM'), pBestPlot.getX(), pBestPlot.getY(), UnitAITypes.NO_UNITAI)
				newUnit1.setHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_IMMORTAL'), true)
				newUnit2 = gc.getPlayer(iInfernalPlayer).initUnit(CvUtil.getInfoTypeForStringCached('UNIT_LONGBOWMAN_INFERNAL'), pBestPlot.getX(), pBestPlot.getY(), UnitAITypes.NO_UNITAI)
				newUnit2.setHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_MOBILITY1'), true)
				newUnit3 = gc.getPlayer(iInfernalPlayer).initUnit(CvUtil.getInfoTypeForStringCached('UNIT_LONGBOWMAN_INFERNAL'), pBestPlot.getX(), pBestPlot.getY(), UnitAITypes.NO_UNITAI)
				newUnit3.setHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_MOBILITY1'), true)
				newUnit4 = gc.getPlayer(iInfernalPlayer).initUnit(CvUtil.getInfoTypeForStringCached('UNIT_SECT_OF_FLIES'), pBestPlot.getX(), pBestPlot.getY(), UnitAITypes.NO_UNITAI)
				newUnit4.setHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_IRON_WEAPONS'), true)
				newUnit4.setHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_MOBILITY1'), true)
				newUnit5 = gc.getPlayer(iInfernalPlayer).initUnit(CvUtil.getInfoTypeForStringCached('UNIT_SECT_OF_FLIES'), pBestPlot.getX(), pBestPlot.getY(), UnitAITypes.NO_UNITAI)
				newUnit5.setHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_IRON_WEAPONS'), true)
				newUnit5.setHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_MOBILITY1'), true)
				newUnit6 = gc.getPlayer(iInfernalPlayer).initUnit(CvUtil.getInfoTypeForStringCached('UNIT_WORKER'), pBestPlot.getX(), pBestPlot.getY(), UnitAITypes.NO_UNITAI)
				newUnit6.setHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_DEMON'), true)
				newUnit7 = gc.getPlayer(iInfernalPlayer).initUnit(CvUtil.getInfoTypeForStringCached('UNIT_IMP'), pBestPlot.getX(), pBestPlot.getY(), UnitAITypes.NO_UNITAI)
				newUnit7.setHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_MOBILITY1'), true)
				newUnit8 = gc.getPlayer(iInfernalPlayer).initUnit(CvUtil.getInfoTypeForStringCached('UNIT_MANES'), pBestPlot.getX(), pBestPlot.getY(), UnitAITypes.NO_UNITAI)
				newUnit9 = gc.getPlayer(iInfernalPlayer).initUnit(CvUtil.getInfoTypeForStringCached('UNIT_MANES'), pBestPlot.getX(), pBestPlot.getY(), UnitAITypes.NO_UNITAI)
				newUnit10 = gc.getPlayer(iInfernalPlayer).initUnit(CvUtil.getInfoTypeForStringCached('UNIT_MANES'), pBestPlot.getX(), pBestPlot.getY(), UnitAITypes.NO_UNITAI)
				newUnit11 = gc.getPlayer(iInfernalPlayer).initUnit(CvUtil.getInfoTypeForStringCached('UNIT_SETTLER'), pBestPlot.getX(), pBestPlot.getY(), UnitAITypes.NO_UNITAI)
				newUnit11.setHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_MOBILITY1'), true)
				newUnit11.setHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_SENTRY'), true)
				newUnit11.setHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_SENTRY2'), true)
				newUnit11.setHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_DEMON'), true)
				if gc.getPlayer(iPlayer).isHuman():
					popupInfo = CyPopupInfo()
					popupInfo.setButtonPopupType(ButtonPopupTypes.BUTTONPOPUP_PYTHON)
					popupInfo.setText(CyTranslator().getText("TXT_KEY_POPUP_CONTROL_INFERNAL",()))
					popupInfo.setData1(iPlayer)
					popupInfo.setData2(iInfernalPlayer)
					popupInfo.addPythonButton(CyTranslator().getText("TXT_KEY_POPUP_YES", ()), "")
					popupInfo.addPythonButton(CyTranslator().getText("TXT_KEY_POPUP_NO", ()), "")
					popupInfo.setOnClickedPythonCallback("reassignPlayer")
					popupInfo.addPopup(iPlayer)

		if (not self.__LOG_TECH):
			return
		CvUtil.pyPrint('%s was finished by Team %d' 
			%(PyInfo.TechnologyInfo(iTechType).getDescription(), iTeam))
	
	def onTechSelected(self, argsList):
		'Tech Selected'
		iTechType, iPlayer = argsList
		if (not self.__LOG_TECH):
			return
		CvUtil.pyPrint('%s was selected by Player %d' %(PyInfo.TechnologyInfo(iTechType).getDescription(), iPlayer))
	
	def onReligionFounded(self, argsList):
		'Religion Founded'
		iReligion, iFounder = argsList
		player = PyPlayer(iFounder)

		iCityId = gc.getGame().getHolyCity(iReligion).getID()
		if (gc.getGame().isFinalInitialized() and not gc.getGame().GetWorldBuilderMode()):
			if ((not CyGame().isNetworkMultiPlayer()) and (iFounder == CyGame().getActivePlayer())):
				popupInfo = CyPopupInfo()
				popupInfo.setButtonPopupType(ButtonPopupTypes.BUTTONPOPUP_PYTHON_SCREEN)
				popupInfo.setData1(iReligion)
				popupInfo.setData2(iCityId)
				popupInfo.setData3(1)
				popupInfo.setText(u"showWonderMovie")
				popupInfo.addPopup(iFounder)

		if (not self.__LOG_RELIGION):
			return
		CvUtil.pyPrint('Player %d Civilization %s has founded %s'
			%(iFounder, player.getCivilizationName(), gc.getReligionInfo(iReligion).getDescription()))

	def onReligionSpread(self, argsList):
		'Religion Has Spread to a City'
		iReligion, iOwner, pSpreadCity = argsList
		iOrder = CvUtil.getInfoTypeForStringCached('RELIGION_THE_ORDER')
		iVeil = CvUtil.getInfoTypeForStringCached('RELIGION_THE_ASHEN_VEIL')
		player = PyPlayer(iOwner)
		pPlayer = gc.getPlayer(iOwner)
		if iReligion == iVeil:
			if pSpreadCity.isHolyCityByType(iVeil):
				cf.FFHChangeArmageddonCounter(5, iOwner)
			if (pSpreadCity.isHasReligion(iOrder) and pSpreadCity.isHolyCityByType(iVeil) == False):
				pSpreadCity.setHasReligion(iReligion, False, False, False)
				CyInterface().addMessage(iOwner,True,25,'The Ashen Veil was removed from a city.','AS2D_ORDER_DINK',1,'Art/Interface/Buttons/Religions/Order.dds',ColorTypes(8),pSpreadCity.getX(),pSpreadCity.getY(),True,True)
			else:
				if pSpreadCity.getOccupationTimer() <= 0:
					cf.FFHChangeArmageddonCounter(1, iOwner)

		if iReligion == iOrder and CyGame().getGameTurn() != CyGame().getStartTurn():
			if (pSpreadCity.isHasReligion(iVeil) and pSpreadCity.isHolyCityByType(iOrder) == False):
				pSpreadCity.setHasReligion(iReligion, False, False, False)
				CyInterface().addMessage(iOwner,True,25,'The Order was removed from a city.','AS2D_ASHEN_DINK',1,'Art/Interface/Buttons/Religions/Ashen.dds',ColorTypes(8),pSpreadCity.getX(),pSpreadCity.getY(),True,True)
			if (pPlayer.getStateReligion() == iOrder and pSpreadCity.isHasReligion(iVeil) == False and pSpreadCity.getOccupationTimer() <= 0):
				iRnd = CyGame().getSorenRandNum(100, "Bob")
				if (iRnd <= gc.getDefineINT('ORDER_CRUSADER_SPAWN_CHANCE')):
					eTeam = gc.getTeam(pPlayer.getTeam())
					if eTeam.isHasTech(CvUtil.getInfoTypeForStringCached('TECH_FANATICISM')):
						iUnit = CvUtil.getInfoTypeForStringCached('UNIT_CRUSADER')
						CyInterface().addMessage(iOwner,True,25,'A Crusader has joined your cause.','AS2D_UNIT_BUILD_UNIT',1,'Art/Interface/Buttons/Units/Crusader.dds',ColorTypes(8),pSpreadCity.getX(),pSpreadCity.getY(),True,True)
					else:
						iUnit = CvUtil.getInfoTypeForStringCached('UNIT_DISCIPLE_THE_ORDER')
						CyInterface().addMessage(iOwner,True,25,'An Acolyte has joined your cause.','AS2D_UNIT_BUILD_UNIT',1,'Art/Interface/Buttons/Units/Discipleorder.dds',ColorTypes(8),pSpreadCity.getX(),pSpreadCity.getY(),True,True)
					newUnit = pPlayer.initUnit(iUnit, pSpreadCity.getX(), pSpreadCity.getY(), UnitAITypes.NO_UNITAI)

		if (not self.__LOG_RELIGIONSPREAD):
			return
		CvUtil.pyPrint('%s has spread to Player %d Civilization %s city of %s'
			%(gc.getReligionInfo(iReligion).getDescription(), iOwner, player.getCivilizationName(), pSpreadCity.getName()))
				
	def onGoldenAge(self, argsList):
		'Golden Age'
		iPlayer = argsList[0]
		player = PyPlayer(iPlayer)
		if (not self.__LOG_GOLDENAGE):
			return
		CvUtil.pyPrint('Player %d Civilization %s has begun a golden age'
			%(iPlayer, player.getCivilizationName()))

	def onEndGoldenAge(self, argsList):
		'End Golden Age'
		iPlayer = argsList[0]
		player = PyPlayer(iPlayer)
		if (not self.__LOG_ENDGOLDENAGE):
			return
		CvUtil.pyPrint('Player %d Civilization %s golden age has ended'
			%(iPlayer, player.getCivilizationName()))

	def onChangeWar(self, argsList):
		'War Status Changes'
		bIsWar = argsList[0]
		iPlayer = argsList[1]
		iRivalTeam = argsList[2]
		if (not self.__LOG_WARPEACE):
			return
		if (bIsWar):
			strStatus = "declared war"
		else:
			strStatus = "declared peace"
		CvUtil.pyPrint('Player %d Civilization %s has %s on Team %d'
			%(iPlayer, gc.getPlayer(iPlayer).getCivilizationName(), strStatus, eRivalTeam))
	
	def onChat(self, argsList):
		'Chat Message Event'
		chatMessage = "%s" %(argsList[0],)
		
	def onSetPlayerAlive(self, argsList):
		'Set Player Alive Event'
		iPlayerID = argsList[0]
		bNewValue = argsList[1]

		if bNewValue == False:
			pPlayer = gc.getPlayer(iPlayerID)
			if pPlayer.getAlignment() == 1:
				cf.FFHChangeArmageddonCounter(5, -1)
			if pPlayer.getAlignment() == -1:
				cf.FFHChangeArmageddonCounter(-5, -1)
			if pPlayer.getLeaderType() == CvUtil.getInfoTypeForStringCached('LEADER_ALEXIS'):
				cf.FFHAddPopup(CyTranslator().getText("TXT_KEY_POPUP_DEFEATED_CALABIM",()),'art/interface/popups/Alexis.dds')
			if pPlayer.getLeaderType() == CvUtil.getInfoTypeForStringCached('LEADER_AMELANCHIER'):
				cf.FFHAddPopup(CyTranslator().getText("TXT_KEY_POPUP_DEFEATED_LJOSALFAR",()),'art/interface/popups/Amelanchier.dds')
			if pPlayer.getLeaderType() == CvUtil.getInfoTypeForStringCached('LEADER_ARENDEL'):
				cf.FFHAddPopup(CyTranslator().getText("TXT_KEY_POPUP_DEFEATED_LJOSALFAR",()),'art/interface/popups/Arendel.dds')
			if pPlayer.getLeaderType() == CvUtil.getInfoTypeForStringCached('LEADER_ARTURUS'):
				cf.FFHAddPopup(CyTranslator().getText("TXT_KEY_POPUP_DEFEATED_KHAZAD",()),'art/interface/popups/Arturus.dds')
			if pPlayer.getLeaderType() == CvUtil.getInfoTypeForStringCached('LEADER_AURIC'):
				cf.FFHAddPopup(CyTranslator().getText("TXT_KEY_POPUP_DEFEATED_ILLIANS",()),'art/interface/popups/Auric.dds')
			if pPlayer.getLeaderType() == CvUtil.getInfoTypeForStringCached('LEADER_BASIUM'):
				cf.FFHAddPopup(CyTranslator().getText("TXT_KEY_POPUP_DEFEATED_MERCURIANS",()),'art/interface/popups/Basium.dds')
			if pPlayer.getLeaderType() == CvUtil.getInfoTypeForStringCached('LEADER_BEERI'):
				cf.FFHAddPopup(CyTranslator().getText("TXT_KEY_POPUP_DEFEATED_LUCHUIRP",()),'art/interface/popups/Beeri.dds')
			if pPlayer.getLeaderType() == CvUtil.getInfoTypeForStringCached('LEADER_CAPRIA'):
				cf.FFHAddPopup(CyTranslator().getText("TXT_KEY_POPUP_DEFEATED_BANNOR",()),'art/interface/popups/Capria.dds')
			if pPlayer.getLeaderType() == CvUtil.getInfoTypeForStringCached('LEADER_CARDITH'):
				cf.FFHAddPopup(CyTranslator().getText("TXT_KEY_POPUP_DEFEATED_KURIOTATES",()),'art/interface/popups/Cardith.dds')
			if pPlayer.getLeaderType() == CvUtil.getInfoTypeForStringCached('LEADER_CASSIEL'):
				cf.FFHAddPopup(CyTranslator().getText("TXT_KEY_POPUP_DEFEATED_GRIGORI",()),'art/interface/popups/Cassiel.dds')
			if pPlayer.getLeaderType() == CvUtil.getInfoTypeForStringCached('LEADER_CHARADON'):
				cf.FFHAddPopup(CyTranslator().getText("TXT_KEY_POPUP_DEFEATED_DOVIELLO",()),'art/interface/popups/Charadon.dds')
			if pPlayer.getLeaderType() == CvUtil.getInfoTypeForStringCached('LEADER_DAIN'):
				cf.FFHAddPopup(CyTranslator().getText("TXT_KEY_POPUP_DEFEATED_AMURITES",()),'art/interface/popups/Dain.dds')
			if pPlayer.getLeaderType() == CvUtil.getInfoTypeForStringCached('LEADER_EINION'):
				cf.FFHAddPopup(CyTranslator().getText("TXT_KEY_POPUP_DEFEATED_ELOHIM",()),'art/interface/popups/Einion.dds')
			if pPlayer.getLeaderType() == CvUtil.getInfoTypeForStringCached('LEADER_ETHNE'):
				cf.FFHAddPopup(CyTranslator().getText("TXT_KEY_POPUP_DEFEATED_ELOHIM",()),'art/interface/popups/Ethne.dds')
			if pPlayer.getLeaderType() == CvUtil.getInfoTypeForStringCached('LEADER_FAERYL'):
				cf.FFHAddPopup(CyTranslator().getText("TXT_KEY_POPUP_DEFEATED_SVARTALFAR",()),'art/interface/popups/Faeryl.dds')
			if pPlayer.getLeaderType() == CvUtil.getInfoTypeForStringCached('LEADER_FALAMAR'):
				cf.FFHAddPopup(CyTranslator().getText("TXT_KEY_POPUP_DEFEATED_LANUN",()),'art/interface/popups/Falamar.dds')
			if pPlayer.getLeaderType() == CvUtil.getInfoTypeForStringCached('LEADER_FLAUROS'):
				cf.FFHAddPopup(CyTranslator().getText("TXT_KEY_POPUP_DEFEATED_CALABIM",()),'art/interface/popups/Flauros.dds')
			if pPlayer.getLeaderType() == CvUtil.getInfoTypeForStringCached('LEADER_GARRIM'):
				cf.FFHAddPopup(CyTranslator().getText("TXT_KEY_POPUP_DEFEATED_LUCHUIRP",()),'art/interface/popups/Garrim.dds')
			if pPlayer.getLeaderType() == CvUtil.getInfoTypeForStringCached('LEADER_HANNAH'):
				cf.FFHAddPopup(CyTranslator().getText("TXT_KEY_POPUP_DEFEATED_HANNAH",()),'art/interface/popups/Hannah.dds')
			if pPlayer.getLeaderType() == CvUtil.getInfoTypeForStringCached('LEADER_HYBOREM'):
				cf.FFHAddPopup(CyTranslator().getText("TXT_KEY_POPUP_DEFEATED_INFERNAL",()),'art/interface/popups/Hyborem.dds')
			if pPlayer.getLeaderType() == CvUtil.getInfoTypeForStringCached('LEADER_JONAS'):
				cf.FFHAddPopup(CyTranslator().getText("TXT_KEY_POPUP_DEFEATED_CLAN_OF_EMBERS",()),'art/interface/popups/Jonus.dds')
			if pPlayer.getLeaderType() == CvUtil.getInfoTypeForStringCached('LEADER_KANDROS'):
				cf.FFHAddPopup(CyTranslator().getText("TXT_KEY_POPUP_DEFEATED_KHAZAD",()),'art/interface/popups/Kandros.dds')
			if pPlayer.getLeaderType() == CvUtil.getInfoTypeForStringCached('LEADER_KEELYN'):
				cf.FFHAddPopup(CyTranslator().getText("TXT_KEY_POPUP_DEFEATED_KEELYN",()),'art/interface/popups/Keelyn.dds')
			if pPlayer.getLeaderType() == CvUtil.getInfoTypeForStringCached('LEADER_MAHALA'):
				cf.FFHAddPopup(CyTranslator().getText("TXT_KEY_POPUP_DEFEATED_DOVIELLO",()),'art/interface/popups/Mahala.dds')
			if pPlayer.getLeaderType() == CvUtil.getInfoTypeForStringCached('LEADER_MORGOTH'):
				cf.FFHAddPopup(CyTranslator().getText("TXT_KEY_POPUP_DEFEATED_SIDAR",()),'art/interface/popups/Morgoth.dds')
			if pPlayer.getLeaderType() == CvUtil.getInfoTypeForStringCached('LEADER_OS-GABELLA'):
				cf.FFHAddPopup(CyTranslator().getText("TXT_KEY_POPUP_DEFEATED_SHEAIM",()),'art/interface/popups/Os-Gabella.dds')
			if pPlayer.getLeaderType() == CvUtil.getInfoTypeForStringCached('LEADER_PERPENTACH'):
				cf.FFHAddPopup(CyTranslator().getText("TXT_KEY_POPUP_DEFEATED_BALSERAPHS",()),'art/interface/popups/Perpentach.dds')
			if pPlayer.getLeaderType() == CvUtil.getInfoTypeForStringCached('LEADER_RHOANNA'):
				cf.FFHAddPopup(CyTranslator().getText("TXT_KEY_POPUP_DEFEATED_HIPPUS",()),'art/interface/popups/Rhoanna.dds')
			if pPlayer.getLeaderType() == CvUtil.getInfoTypeForStringCached('LEADER_SABATHIEL'):
				cf.FFHAddPopup(CyTranslator().getText("TXT_KEY_POPUP_DEFEATED_BANNOR",()),'art/interface/popups/Sabathiel.dds')
			if pPlayer.getLeaderType() == CvUtil.getInfoTypeForStringCached('LEADER_SHEELBA'):
				cf.FFHAddPopup(CyTranslator().getText("TXT_KEY_POPUP_DEFEATED_CLAN_OF_EMBERS",()),'art/interface/popups/Sheelba.dds')
			if pPlayer.getLeaderType() == CvUtil.getInfoTypeForStringCached('LEADER_TASUNKE'):
				cf.FFHAddPopup(CyTranslator().getText("TXT_KEY_POPUP_DEFEATED_HIPPUS",()),'art/interface/popups/Tasunke.dds')
			if pPlayer.getLeaderType() == CvUtil.getInfoTypeForStringCached('LEADER_TEBRYN'):
				cf.FFHAddPopup(CyTranslator().getText("TXT_KEY_POPUP_DEFEATED_SHEAIM",()),'art/interface/popups/Tebryn.dds')
			if pPlayer.getLeaderType() == CvUtil.getInfoTypeForStringCached('LEADER_THESSA'):
				cf.FFHAddPopup(CyTranslator().getText("TXT_KEY_POPUP_DEFEATED_LJOSALFAR",()),'art/interface/popups/Thessa.dds')
			if pPlayer.getLeaderType() == CvUtil.getInfoTypeForStringCached('LEADER_VALLEDIA'):
				cf.FFHAddPopup(CyTranslator().getText("TXT_KEY_POPUP_DEFEATED_AMURITES",()),'art/interface/popups/Valledia.dds')
			if pPlayer.getLeaderType() == CvUtil.getInfoTypeForStringCached('LEADER_VARN'):
				cf.FFHAddPopup(CyTranslator().getText("TXT_KEY_POPUP_DEFEATED_MALAKIM",()),'art/interface/popups/Varn.dds')

		CvUtil.pyPrint("Player %d's alive status set to: %d" %(iPlayerID, int(bNewValue)))
		
	def onCityBuilt(self, argsList):
		'City Built'
		city = argsList[0]
		pPlot = city.plot()
		pPlayer = gc.getPlayer(city.getOwner())

		if pPlot.getBonusType(-1) == CvUtil.getInfoTypeForStringCached('BONUS_MANA'):
			pPlot.setBonusType(-1)

		cf.FFHNewCity(city)

		if pPlayer.getCivilizationType() == CvUtil.getInfoTypeForStringCached('CIVILIZATION_INFERNAL'):
			city.setHasReligion(CvUtil.getInfoTypeForStringCached('RELIGION_THE_ASHEN_VEIL'), True, True, True)
			city.setPopulation(3)
			city.setHasRealBuilding(CvUtil.getInfoTypeForStringCached('BUILDING_ELDER_COUNCIL'), True)
			city.setHasRealBuilding(CvUtil.getInfoTypeForStringCached('BUILDING_TRAINING_YARD'), True)
			city.setHasRealBuilding(CvUtil.getInfoTypeForStringCached('BUILDING_OBSIDIAN_GATE'), True)
			city.setHasRealBuilding(CvUtil.getInfoTypeForStringCached('BUILDING_FORGE'), True)
			city.setHasRealBuilding(CvUtil.getInfoTypeForStringCached('BUILDING_MAGE_GUILD'), True)
		if pPlayer.getCivilizationType() == CvUtil.getInfoTypeForStringCached('CIVILIZATION_BARBARIAN'):
			newUnit = pPlayer.initUnit(CvUtil.getInfoTypeForStringCached('UNIT_ORC_SPEARMAN'), city.getX(), city.getY(), UnitAITypes.NO_UNITAI)
			newUnit = pPlayer.initUnit(CvUtil.getInfoTypeForStringCached('UNIT_ORC_SPEARMAN'), city.getX(), city.getY(), UnitAITypes.NO_UNITAI)

		if (city.getOwner() == CyGame().getActivePlayer()):
			self.__eventEditCityNameBegin(city, False)	
		CvUtil.pyPrint('City Built Event: %s' %(city.getName()))
		
	def onCityRazed(self, argsList):
		'City Razed'
		city, iPlayer = argsList
		
		owner = PyPlayer(city.getOwner())
		razor = PyPlayer(iPlayer)
		pPlayer = gc.getPlayer(iPlayer)
		i = 0

		if city.isHasReligion(CvUtil.getInfoTypeForStringCached('RELIGION_THE_ASHEN_VEIL')):
			i = i - 1
		else:
			i = i + (city.getPopulation() / 4) + 1
		if city.isHolyCityByType(CvUtil.getInfoTypeForStringCached('RELIGION_THE_ASHEN_VEIL')):
			i = i - 5
		if city.isHolyCityByType(CvUtil.getInfoTypeForStringCached('RELIGION_THE_ORDER')):
			i = i + 5
		if city.isHolyCityByType(CvUtil.getInfoTypeForStringCached('RELIGION_RUNES_OF_KILMORPH')):
			i = i + 2
		if city.isHolyCityByType(CvUtil.getInfoTypeForStringCached('RELIGION_OCTOPUS_OVERLORDS')):
			i = i - 2
		if city.hasBuilding(CvUtil.getInfoTypeForStringCached('BUILDING_STIGMATA_ON_THE_UNBORN')):
			i = i - 5
		if city.hasBuilding(CvUtil.getInfoTypeForStringCached('BUILDING_PROPHECY_OF_RAGNAROK')):
			i = i - 5
		if i != 0:
			cf.FFHChangeArmageddonCounter(i, iPlayer)
	
		if gc.getPlayer(city.getOriginalOwner()).getAlignment() == -1:
			for i in range(city.getPopulation()):
				cf.FFHGiftUnit(CvUtil.getInfoTypeForStringCached('UNIT_MANES'), CvUtil.getInfoTypeForStringCached('CIVILIZATION_INFERNAL'), 0, city.plot(), city.getOwner())

		if gc.getPlayer(city.getOriginalOwner()).getAlignment() == 0:
			for i in range((city.getPopulation() / 4) + 1):
				cf.FFHGiftUnit(CvUtil.getInfoTypeForStringCached('UNIT_MANES'), CvUtil.getInfoTypeForStringCached('CIVILIZATION_INFERNAL'), 0, city.plot(), city.getOwner())
				cf.FFHGiftUnit(CvUtil.getInfoTypeForStringCached('UNIT_MANES'), CvUtil.getInfoTypeForStringCached('CIVILIZATION_INFERNAL'), 0, city.plot(), city.getOwner())
				cf.FFHGiftUnit(CvUtil.getInfoTypeForStringCached('UNIT_ANGEL'), CvUtil.getInfoTypeForStringCached('CIVILIZATION_MERCURIANS'), 0, city.plot(), city.getOwner())

		if gc.getPlayer(city.getOriginalOwner()).getAlignment() == 1:
			for i in range((city.getPopulation() / 2) + 1):
				cf.FFHGiftUnit(CvUtil.getInfoTypeForStringCached('UNIT_ANGEL'), CvUtil.getInfoTypeForStringCached('CIVILIZATION_MERCURIANS'), 0, city.plot(), city.getOwner())

		if city.hasBuilding(CvUtil.getInfoTypeForStringCached('BUILDING_THE_DRAGONS_HORDE')):
			newUnit = pPlayer.initUnit(CvUtil.getInfoTypeForStringCached('UNIT_DRAGONS_HOARD'), city.getX(), city.getY(), UnitAITypes.UNITAI_SPELL)

		CvUtil.pyPrint('Player %d Civilization %s City %s was razed by Player %d' 
			%(owner.getID(), owner.getCivilizationName(), city.getName(), razor.getID()))
		CvUtil.pyPrint("City Razed Event: %s" %(city.getName(),))
	
	def onCityAcquired(self, argsList):
		'City Acquired'
		owner,playerType,city,bConquest,bTrade = argsList
		pPlayer = gc.getPlayer(city.getOwner())

		if (pPlayer.getCivics(CvUtil.getInfoTypeForStringCached('CIVICOPTION_ECONOMY')) == CvUtil.getInfoTypeForStringCached('CIVIC_CONQUEST')):
			iGold = CyGame().getSorenRandNum(50, "Bob") + 25 + pPlayer.getGold()
			pPlayer.setGold(iGold)
			if city.getPopulation() >= 2:
				city.setPopulation(city.getPopulation() - 1)

		city.setHasRealBuilding(CvUtil.getInfoTypeForStringCached('BUILDING_DWARVEN_VAULT_EMPTY'), False)
		city.setHasRealBuilding(CvUtil.getInfoTypeForStringCached('BUILDING_DWARVEN_VAULT_LOW'), False)
		city.setHasRealBuilding(CvUtil.getInfoTypeForStringCached('BUILDING_DWARVEN_VAULT'), False)
		city.setHasRealBuilding(CvUtil.getInfoTypeForStringCached('BUILDING_DWARVEN_VAULT_ABUNDANT'), False)
		city.setHasRealBuilding(CvUtil.getInfoTypeForStringCached('BUILDING_DWARVEN_VAULT_STOCKED'), False)
		city.setHasRealBuilding(CvUtil.getInfoTypeForStringCached('BUILDING_DWARVEN_VAULT_FULL'), False)
		city.setHasRealBuilding(CvUtil.getInfoTypeForStringCached('BUILDING_DWARVEN_VAULT_OVERFLOWING'), False)
		city.setHasRealBuilding(CvUtil.getInfoTypeForStringCached('BUILDING_SETTLEMENT'), False)
		city.setHasRealBuilding(CvUtil.getInfoTypeForStringCached('BUILDING_CITY'), False)
		if pPlayer.isSprawling():
			if pPlayer.getBuildingClassCount(CvUtil.getInfoTypeForStringCached("BUILDINGCLASS_CITY")) >= cf.FFHGetMaxSprawlingCities():
				city.setHasRealBuilding(CvUtil.getInfoTypeForStringCached('BUILDING_SETTLEMENT'), True)
			else:
				city.setHasRealBuilding(CvUtil.getInfoTypeForStringCached('BUILDING_CITY'), True)

		CvUtil.pyPrint('City Acquired Event: %s' %(city.getName()))
	
	def onCityLost(self, argsList):
		'City Lost'
		city = argsList[0]
		player = PyPlayer(city.getOwner())
		if (not self.__LOG_CITYLOST):
			return
		CvUtil.pyPrint('City %s was lost by Player %d Civilization %s' 
			%(city.getName(), player.getID(), player.getCivilizationName()))
	
	def onCultureExpansion(self, argsList):
		'City Culture Expansion'
		pCity = argsList[0]
		iPlayer = argsList[1]
		CvUtil.pyPrint("City %s's culture has expanded" %(pCity.getName(),))
	
	def onCityGrowth(self, argsList):
		'City Population Growth'
		pCity = argsList[0]
		iPlayer = argsList[1]
		CvUtil.pyPrint("%s has grown" %(pCity.getName(),))

	def onCityDoTurn(self, argsList):
		'City Production'
		pCity = argsList[0]
		iPlayer = argsList[1]
		pPlot = pCity.plot()
		pPlayer = gc.getPlayer(pCity.getOwner())

		if pCity.hasBuilding(CvUtil.getInfoTypeForStringCached('BUILDING_HOPE')):
			bKeep = False
			for i in range(pPlot.getNumUnits()):
				pUnit = pPlot.getUnit(i)
				if (pUnit.isHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_SORCERY')) and pUnit.isHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_SPIRIT2'))):
					bKeep = True
			pCity.setHasRealBuilding(CvUtil.getInfoTypeForStringCached('BUILDING_HOPE'), bKeep)

		if pCity.hasBuilding(CvUtil.getInfoTypeForStringCached('BUILDING_INSPIRATION')):
			bKeep = False
			for i in range(pPlot.getNumUnits()):
				pUnit = pPlot.getUnit(i)
				if (pUnit.isHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_SORCERY')) and pUnit.isHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_MIND2'))):
					bKeep = True
			pCity.setHasRealBuilding(CvUtil.getInfoTypeForStringCached('BUILDING_INSPIRATION'), bKeep)

		if pCity.hasBuilding(CvUtil.getInfoTypeForStringCached('BUILDING_UNYIELDING_ORDER')):
			bKeep = False
			for i in range(pPlot.getNumUnits()):
				pUnit = pPlot.getUnit(i)
				if (pUnit.isHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_SORCERY')) and pUnit.isHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_LAW3'))):
					bKeep = True
			pCity.setHasRealBuilding(CvUtil.getInfoTypeForStringCached('BUILDING_UNYIELDING_ORDER'), bKeep)

		if pCity.hasBuilding(CvUtil.getInfoTypeForStringCached('BUILDING_WALL_OF_STONE')):
			bKeep = False
			for i in range(pPlot.getNumUnits()):
				pUnit = pPlot.getUnit(i)
				if (pUnit.isHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_CHANNELING1')) and pUnit.isHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_EARTH1'))):
					bKeep = True
			pCity.setHasRealBuilding(CvUtil.getInfoTypeForStringCached('BUILDING_WALL_OF_STONE'), bKeep)

		if pCity.hasBuilding(CvUtil.getInfoTypeForStringCached('BUILDING_CITADEL_OF_LIGHT')):
			bAlert = False
			iX = pCity.getX()
			iY = pCity.getY()
			eTeam = pPlayer.getTeam()
			for iiX in range(iX-2, iX+3, 1):
				for iiY in range(iY-2, iY+3, 1):
					pPlot = CyMap().plot(iiX,iiY)
					for i in range(pPlot.getNumUnits()):
						pUnit = pPlot.getUnit(i)
						p2Player = gc.getPlayer(pUnit.getOwner())
						e2Team = gc.getTeam(p2Player.getTeam())
						if e2Team.isAtWar(eTeam) == True:
							bAlert = True
			if bAlert == True:
				newUnit = pPlayer.initUnit(CvUtil.getInfoTypeForStringCached('UNIT_FIREBALL'), pCity.getX(), pCity.getY(), UnitAITypes.UNITAI_SPELL)
				newUnit.setHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_SUMMONED2'), True)

		if pCity.hasBuilding(CvUtil.getInfoTypeForStringCached('BUILDING_AQUAE_SUCELLUS')):
			for i in range(pPlot.getNumUnits()):
				pUnit = pPlot.getUnit(i)
				pUnit.setHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_DISEASED'), False)
				pUnit.setHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_WITHERED'), False)
				pUnit.setHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_PLAGUED'), False)

		if (pCity.hasBuilding(CvUtil.getInfoTypeForStringCached('BUILDING_FORGE')) or pCity.hasBuilding(CvUtil.getInfoTypeForStringCached('BUILDING_DWARVEN_FORGE'))):
			for i in range(pPlot.getNumUnits()):
				pUnit = pPlot.getUnit(i)
				pUnit.setHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_RUSTED'), False)

		if (pCity.isHasReligion(CvUtil.getInfoTypeForStringCached('RELIGION_CULT_OF_THE_DRAGON')) and pPlayer.getCivilizationType() != CvUtil.getInfoTypeForStringCached('CIVILIZATION_KURIOTATES') and pPlayer.getCivilizationType() != CvUtil.getInfoTypeForStringCached('CIVILIZATION_SHEAIM')):
			pCity.changeCulture(pCity.getOwner(), -2, True)

		if pCity.hasBuilding(CvUtil.getInfoTypeForStringCached('BUILDING_CARNIVAL')):
			if CyGame().getSorenRandNum(100000, "Bob") <= gc.getDefineINT('EVENT_CHANCE_CITY'):
				p2Plot = cf.FFHFindClearPlot(-1, pPlot)
				if p2Plot != -1:
					cf.FFHEventCarnivalEscape(pCity, p2Plot)

		if pCity.hasBuilding(CvUtil.getInfoTypeForStringCached('BUILDING_EYES_AND_EARS_NETWORK')):
			eTeam = gc.getTeam(pPlayer.getTeam())
			listTeams = []
			for iPlayer2 in range(gc.getMAX_PLAYERS()):
				pPlayer2 = gc.getPlayer(iPlayer2)
				if (pPlayer2.isAlive() and iPlayer2 != iPlayer):
					iTeam2 = pPlayer2.getTeam()
					if eTeam.isOpenBorders(iTeam2):
						listTeams.append(gc.getTeam(iTeam2))
			if len(listTeams) >= 3:
				for iTech in range(gc.getNumTechInfos()):
					if (iTech != CvUtil.getInfoTypeForStringCached('TECH_ARETE') and iTech != CvUtil.getInfoTypeForStringCached('TECH_MIND_STAPLING') and iTech != CvUtil.getInfoTypeForStringCached('TECH_UNQUESTIONING_OBEDIENCE') and iTech != CvUtil.getInfoTypeForStringCached('TECH_HIDDEN_PATHS') and iTech != CvUtil.getInfoTypeForStringCached('TECH_INFERNAL_PACT') and iTech != CvUtil.getInfoTypeForStringCached('TECH_SEAFARING') and iTech != CvUtil.getInfoTypeForStringCached('TECH_AWAKEN_THE_ANCIENTS')):
						if eTeam.isHasTech(iTech) == False:
							iCount = 0
							for i in range(len(listTeams)):
								if listTeams[i].isHasTech(iTech):
									iCount = iCount + 1
							if iCount >= 3:
								eTeam.setHasTech(iTech, True, iPlayer, False, True)
								CyInterface().addMessage(iPlayer,True,25,'You received a free Tech from the Eyes and Ears Network.','AS2D_TECH_DING',1,'Art/Interface/Buttons/Buildings/Eyesandearsnetwork.dds',ColorTypes(8),pCity.getX(),pCity.getY(),True,True)

		pCity.setHasRealBuilding(CvUtil.getInfoTypeForStringCached('BUILDING_GYPSY_CAMP'), False)
		for i in range(pPlot.getNumUnits()):
			pUnit = pPlot.getUnit(i)
			if (pUnit.getUnitType() == CvUtil.getInfoTypeForStringCached('UNIT_GYPSY_WAGON') or pUnit.getUnitType() == CvUtil.getInfoTypeForStringCached('UNIT_LOKI')):
				if pUnit.getOwner() != pCity.getOwner():
					pCity.setHasRealBuilding(CvUtil.getInfoTypeForStringCached('BUILDING_GYPSY_CAMP'), True)
					gc.getPlayer(pUnit.getOwner()).changeGold(1)
					if (pUnit.getUnitType() == CvUtil.getInfoTypeForStringCached('UNIT_LOKI') and CyGame().getSorenRandNum(100, "Loki") <= gc.getDefineINT('LOKI_UNREST_CHANCE')):
						pCity.changeHurryAngerTimer(2)
						CyInterface().addMessage(pUnit.getOwner(),True,25,'Loki has caused unrest in an opponents city.','AS2D_REVOLTSTART',1,'Art/Interface/Buttons/Units/Loki.dds',ColorTypes(8),pCity.getX(),pCity.getY(),True,True)
						CyInterface().addMessage(pCity.getOwner(),True,25,'Loki has caused unrest in one of your cities.','AS2D_REVOLTSTART',1,'Art/Interface/Buttons/Units/Loki.dds',ColorTypes(8),pCity.getX(),pCity.getY(),True,True)
						if pCity.getCulture(pCity.getOwner()) < 1:
							gc.getPlayer(pUnit.getOwner()).acquireCity(pCity,false,false)

		if pCity.hasBuilding(CvUtil.getInfoTypeForStringCached('BUILDING_PLANAR_GATE')):
			iMax = 1
			iMult = 1
			if CyGame().getArmageddonCounter() >= 50:
				iMax = 2
				iMult = 1.5
			if CyGame().getArmageddonCounter() >= 75:
				iMax = 3
				iMult = 2
			if CyGame().getArmageddonCounter() == 100:
				iMax = 4
				iMult = 2.5
			if CyGame().getSorenRandNum(10000, "Planar Gate") <= gc.getDefineINT('PLANAR_GATE_CHANCE') * iMult:
				listUnits = []
				iMax = iMax * pPlayer.countNumBuildings(CvUtil.getInfoTypeForStringCached('BUILDING_PLANAR_GATE'))
				if pCity.hasBuilding(CvUtil.getInfoTypeForStringCached('BUILDING_GAMBLING_HOUSE')):
					if pPlayer.getUnitClassCount(CvUtil.getInfoTypeForStringCached('UNITCLASS_REVELERS')) < iMax:
						listUnits.append(CvUtil.getInfoTypeForStringCached('UNIT_REVELERS'))
				if pCity.hasBuilding(CvUtil.getInfoTypeForStringCached('BUILDING_MAGE_GUILD')):
					if pPlayer.getUnitClassCount(CvUtil.getInfoTypeForStringCached('UNITCLASS_MOBIUS_WITCH')) < iMax:
						listUnits.append(CvUtil.getInfoTypeForStringCached('UNIT_MOBIUS_WITCH'))
				if pCity.hasBuilding(CvUtil.getInfoTypeForStringCached('BUILDING_CARNIVAL')):
					if pPlayer.getUnitClassCount(CvUtil.getInfoTypeForStringCached('UNITCLASS_CHAOS_MARAUDER')) < iMax:
						listUnits.append(CvUtil.getInfoTypeForStringCached('UNIT_CHAOS_MARAUDER'))
				if pCity.hasBuilding(CvUtil.getInfoTypeForStringCached('BUILDING_LARGE_ANIMAL_STABLES')):
					if pPlayer.getUnitClassCount(CvUtil.getInfoTypeForStringCached('UNITCLASS_MANTICORE')) < iMax:
						listUnits.append(CvUtil.getInfoTypeForStringCached('UNIT_MANTICORE'))
				if pCity.hasBuilding(CvUtil.getInfoTypeForStringCached('BUILDING_PUBLIC_BATHS')):
					if pPlayer.getUnitClassCount(CvUtil.getInfoTypeForStringCached('UNITCLASS_SUCCUBUS')) < iMax:
						listUnits.append(CvUtil.getInfoTypeForStringCached('UNIT_SUCCUBUS'))
				if pCity.hasBuilding(CvUtil.getInfoTypeForStringCached('BUILDING_WEAPONSMITH')):
					if pPlayer.getUnitClassCount(CvUtil.getInfoTypeForStringCached('UNITCLASS_MINOTAUR')) < iMax:
						listUnits.append(CvUtil.getInfoTypeForStringCached('UNIT_MINOTAUR'))
				if pCity.hasBuilding(CvUtil.getInfoTypeForStringCached('BUILDING_TEMPLE_OF_THE_VEIL')):
					if pPlayer.getUnitClassCount(CvUtil.getInfoTypeForStringCached('UNITCLASS_TAR_DEMON')) < iMax:
						listUnits.append(CvUtil.getInfoTypeForStringCached('UNIT_TAR_DEMON'))
				if len(listUnits) > 0:
					iUnit = listUnits[CyGame().getSorenRandNum(len(listUnits), "Planar Gate")]
					newUnit = pPlayer.initUnit(iUnit, pPlot.getX(), pPlot.getY(), UnitAITypes.NO_UNITAI)
					CyInterface().addMessage(iPlayer,True,25,'Creature entered your city through the Planar Gate','AS2D_DISCOVERBONUS',1,gc.getUnitInfo(newUnit.getUnitType()).getButton(),ColorTypes(8),pCity.getX(),pCity.getY(),True,True)
					if iUnit == CvUtil.getInfoTypeForStringCached('UNIT_MOBIUS_WITCH'):
						promotions = [ 'PROMOTION_AIR1','PROMOTION_BODY1','PROMOTION_CHAOS1','PROMOTION_DEATH1','PROMOTION_DIMENSIONAL1','PROMOTION_EARTH1','PROMOTION_ENCHANTMENT1','PROMOTION_ENTROPY1','PROMOTION_FIRE1','PROMOTION_LAW1','PROMOTION_LIFE1','PROMOTION_MIND1','PROMOTION_NATURE1','PROMOTION_SPIRIT1','PROMOTION_WATER1' ]
						newUnit.setLevel(4)
						newUnit.setExperience(14, -1)
						for i in promotions:
							if CyGame().getSorenRandNum(10, "Bob") == 1:
								newUnit.setHasPromotion(CvUtil.getInfoTypeForStringCached(i), True)

		if pCity.hasBuilding(CvUtil.getInfoTypeForStringCached('BUILDING_RIDE_OF_THE_NINE_KINGS')):
			for i in range(pPlot.getNumUnits()):
				pUnit = pPlot.getUnit(i)
				pUnit.setHasPromotion(CvUtil.getInfoTypeForStringCached('PROMOTION_MORALE'), True)

		if gc.getPlayer(pCity.getOwner()).getCivilizationType() == CvUtil.getInfoTypeForStringCached('CIVILIZATION_INFERNAL'):
			if pCity.isHasReligion(CvUtil.getInfoTypeForStringCached('RELIGION_THE_ORDER')):
				pCity.setHasReligion(CvUtil.getInfoTypeForStringCached('RELIGION_THE_ORDER'), False, True, True)
			if pCity.isHasReligion(CvUtil.getInfoTypeForStringCached('RELIGION_THE_ASHEN_VEIL')) == False:
				pCity.setHasReligion(CvUtil.getInfoTypeForStringCached('RELIGION_THE_ASHEN_VEIL'), True, True, True)

		if gc.getPlayer(pCity.getOwner()).getCivilizationType() == CvUtil.getInfoTypeForStringCached('CIVILIZATION_MERCURIANS'):
			if pCity.isHasReligion(CvUtil.getInfoTypeForStringCached('RELIGION_THE_ASHEN_VEIL')):
				pCity.setHasReligion(CvUtil.getInfoTypeForStringCached('RELIGION_THE_ASHEN_VEIL'), False, True, True)

		CvAdvisorUtils.cityAdvise(pCity, iPlayer)

	def onCityBuildingUnit(self, argsList):
		'City begins building a unit'
		pCity = argsList[0]
		iUnitType = argsList[1]
		if (not self.__LOG_CITYBUILDING):
			return
		CvUtil.pyPrint("%s has begun building a %s" %(pCity.getName(),gc.getUnitInfo(iUnitType).getDescription()))
	
	def onCityBuildingBuilding(self, argsList):
		'City begins building a Building'
		pCity = argsList[0]
		iBuildingType = argsList[1]
		if (not self.__LOG_CITYBUILDING):
			return
		CvUtil.pyPrint("%s has begun building a %s" %(pCity.getName(),gc.getBuildingInfo(iBuildingType).getDescription()))
	
	def onCityRename(self, argsList):
		'City is renamed'
		pCity = argsList[0]
		if (pCity.getOwner() == CyGame().getActivePlayer()):
			self.__eventEditCityNameBegin(pCity, True)	

	def onVictory(self, argsList):
		'Victory'
		iTeam, iVictory = argsList

		for iPlayer in range(gc.getMAX_PLAYERS()):
			pPlayer = gc.getPlayer(iPlayer)
			if pPlayer.isAlive():
				if pPlayer.isHuman():
					if pPlayer.getTeam() == iTeam:
						if pPlayer.getCivilizationType() == CvUtil.getInfoTypeForStringCached('CIVILIZATION_AMURITES'):
							Trophies.addTrophy("TROPHY_VICTORY_AMURITES")
						if pPlayer.getCivilizationType() == CvUtil.getInfoTypeForStringCached('CIVILIZATION_BALSERAPHS'):
							Trophies.addTrophy("TROPHY_VICTORY_BALSERAPHS")
						if pPlayer.getCivilizationType() == CvUtil.getInfoTypeForStringCached('CIVILIZATION_BANNOR'):
							Trophies.addTrophy("TROPHY_VICTORY_BANNOR")
						if pPlayer.getCivilizationType() == CvUtil.getInfoTypeForStringCached('CIVILIZATION_CALABIM'):
							Trophies.addTrophy("TROPHY_VICTORY_CALABIM")
						if pPlayer.getCivilizationType() == CvUtil.getInfoTypeForStringCached('CIVILIZATION_CLAN_OF_EMBERS'):
							Trophies.addTrophy("TROPHY_VICTORY_CLAN_OF_EMBERS")
						if pPlayer.getCivilizationType() == CvUtil.getInfoTypeForStringCached('CIVILIZATION_DOVIELLO'):
							Trophies.addTrophy("TROPHY_VICTORY_DOVIELLO")
						if pPlayer.getCivilizationType() == CvUtil.getInfoTypeForStringCached('CIVILIZATION_ELOHIM'):
							Trophies.addTrophy("TROPHY_VICTORY_ELOHIM")
						if pPlayer.getCivilizationType() == CvUtil.getInfoTypeForStringCached('CIVILIZATION_GRIGORI'):
							Trophies.addTrophy("TROPHY_VICTORY_GRIGORI")
						if pPlayer.getCivilizationType() == CvUtil.getInfoTypeForStringCached('CIVILIZATION_HIPPUS'):
							Trophies.addTrophy("TROPHY_VICTORY_HIPPUS")
						if pPlayer.getCivilizationType() == CvUtil.getInfoTypeForStringCached('CIVILIZATION_ILLIANS'):
							Trophies.addTrophy("TROPHY_VICTORY_ILLIANS")
						if pPlayer.getCivilizationType() == CvUtil.getInfoTypeForStringCached('CIVILIZATION_INFERNAL'):
							Trophies.addTrophy("TROPHY_VICTORY_INFERNAL")
						if pPlayer.getCivilizationType() == CvUtil.getInfoTypeForStringCached('CIVILIZATION_KHAZAD'):
							Trophies.addTrophy("TROPHY_VICTORY_KHAZAD")
						if pPlayer.getCivilizationType() == CvUtil.getInfoTypeForStringCached('CIVILIZATION_KURIOTATES'):
							Trophies.addTrophy("TROPHY_VICTORY_KURIOTATES")
						if pPlayer.getCivilizationType() == CvUtil.getInfoTypeForStringCached('CIVILIZATION_LANUN'):
							Trophies.addTrophy("TROPHY_VICTORY_LANUN")
						if pPlayer.getCivilizationType() == CvUtil.getInfoTypeForStringCached('CIVILIZATION_LJOSALFAR'):
							Trophies.addTrophy("TROPHY_VICTORY_LJOSALFAR")
						if pPlayer.getCivilizationType() == CvUtil.getInfoTypeForStringCached('CIVILIZATION_LUCHUIRP'):
							Trophies.addTrophy("TROPHY_VICTORY_LUCHUIRP")
						if pPlayer.getCivilizationType() == CvUtil.getInfoTypeForStringCached('CIVILIZATION_MALAKIM'):
							Trophies.addTrophy("TROPHY_VICTORY_MALAKIM")
						if pPlayer.getCivilizationType() == CvUtil.getInfoTypeForStringCached('CIVILIZATION_MERCURIANS'):
							Trophies.addTrophy("TROPHY_VICTORY_MERCURIANS")
						if pPlayer.getCivilizationType() == CvUtil.getInfoTypeForStringCached('CIVILIZATION_SHEAIM'):
							Trophies.addTrophy("TROPHY_VICTORY_SHEAIM")
						if pPlayer.getCivilizationType() == CvUtil.getInfoTypeForStringCached('CIVILIZATION_SIDAR'):
							Trophies.addTrophy("TROPHY_VICTORY_SIDAR")
						if pPlayer.getCivilizationType() == CvUtil.getInfoTypeForStringCached('CIVILIZATION_SVARTALFAR'):
							Trophies.addTrophy("TROPHY_VICTORY_SVARTALFAR")
						if iVictory == CvUtil.getInfoTypeForStringCached('VICTORY_ALTAR_OF_THE_LUONNOTAR'):
							Trophies.addTrophy("TROPHY_VICTORY_ALTAR_OF_THE_LUONNOTAR")
						if iVictory == CvUtil.getInfoTypeForStringCached('VICTORY_TOWER_OF_MASTERY'):
							Trophies.addTrophy("TROPHY_VICTORY_TOWER_OF_MASTERY")

		if (iVictory >= 0 and iVictory < gc.getNumVictoryInfos()):
			victoryInfo = gc.getVictoryInfo(int(iVictory))
			CvUtil.pyPrint("Victory!  Team %d achieves a %s victory"
				%(iTeam, victoryInfo.getDescription()))
	
	def onGameUpdate(self, argsList):
		'sample generic event, called on each game turn slice'
		genericArgs = argsList[0][0]	# tuple of tuple of my args
		turnSlice = genericArgs[0]
	
	def onMouseEvent(self, argsList):
		'mouse handler - returns 1 if the event was consumed'
		eventType,mx,my,px,py,interfaceConsumed,screens = argsList
		print "onMouseEvent"
		print eventType
		if ( px!=-1 and py!=-1 ):
			if ( eventType == self.EventLButtonDown ):
				if (self.bAllowCheats and self.bCtrl and self.bAlt and CyMap().plot(px,py).isCity() and not interfaceConsumed):
					# Launch Edit City Event
					self.beginEvent( CvUtil.EventEditCity, (px,py) )
					return 1
				
				elif (self.bAllowCheats and self.bCtrl and self.bShift and not interfaceConsumed):
					# Launch Place Object Event
					self.beginEvent( CvUtil.EventPlaceObject, (px, py) )
					return 1
			
		if ( eventType == self.EventBack ):
			return CvScreensInterface.handleBack(screens)
		elif ( eventType == self.EventForward ):
			return CvScreensInterface.handleForward(screens)
		
		return 0
		

#################### TRIGGERED EVENTS ##################	
	def __eventEditCityNameBegin(self, city, bRename):
		popup = PyPopup.PyPopup(CvUtil.EventEditCityName, EventContextTypes.EVENTCONTEXT_ALL)
		popup.setUserData((city.getID(), bRename))
		popup.setHeaderString(localText.getText("TXT_KEY_NAME_CITY", ()))
		popup.setBodyString(localText.getText("TXT_KEY_SETTLE_NEW_CITY_NAME", ()))
		popup.createEditBox(city.getName())
		popup.setEditBoxMaxCharCount( 15 )
		popup.launch()
	
	def __eventEditCityNameApply(self, playerID, userData, popupReturn):	
		'Edit City Name Event'
		iCityID = userData[0]
		bRename = userData[1]
		player = gc.getPlayer(playerID)
		city = player.getCity(iCityID)
		cityName = popupReturn.getEditBoxString(0)
		if (len(cityName) > 30):
			cityName = cityName[:30]
		city.setName(cityName, not bRename)

	def __eventEditCityBegin(self, argsList):
		'Edit City Event'
		px,py = argsList
		CvWBPopups.CvWBPopups().initEditCity(argsList)
	
	def __eventEditCityApply(self, playerID, userData, popupReturn):
		'Edit City Event Apply'
		if (getChtLvl() > 0):
			CvWBPopups.CvWBPopups().applyEditCity( (popupReturn, userData) )

	def __eventPlaceObjectBegin(self, argsList):
		'Place Object Event'
		CvDebugTools.CvDebugTools().initUnitPicker(argsList)
	
	def __eventPlaceObjectApply(self, playerID, userData, popupReturn):
		'Place Object Event Apply'
		if (getChtLvl() > 0):
			CvDebugTools.CvDebugTools().applyUnitPicker( (popupReturn, userData) )

	def __eventAwardTechsAndGoldBegin(self, argsList):
		'Award Techs & Gold Event'
		CvDebugTools.CvDebugTools().cheatTechs()
	
	def __eventAwardTechsAndGoldApply(self, playerID, netUserData, popupReturn):
		'Award Techs & Gold Event Apply'
		if (getChtLvl() > 0):
			CvDebugTools.CvDebugTools().applyTechCheat( (popupReturn) )
	
	def __eventEditUnitNameBegin(self, argsList):
		pUnit = argsList
		popup = PyPopup.PyPopup(CvUtil.EventEditUnitName, EventContextTypes.EVENTCONTEXT_ALL)
		popup.setUserData((pUnit.getID(),))
		popup.setBodyString(localText.getText("TXT_KEY_RENAME_UNIT", ()))
		popup.createEditBox(pUnit.getNameNoDesc())
		popup.launch()

	def __eventEditUnitNameApply(self, playerID, userData, popupReturn):	
		'Edit Unit Name Event'
		iUnitID = userData[0]
		unit = gc.getPlayer(playerID).getUnit(iUnitID)
		newName = popupReturn.getEditBoxString(0)
		if (len(newName) > 25):
			newName = newName[:25]			
		unit.setName(newName)

	def __eventWBAllPlotsPopupBegin(self, argsList):
		CvScreensInterface.getWorldBuilderScreen().allPlotsCB()
		return
	def __eventWBAllPlotsPopupApply(self, playerID, userData, popupReturn):
		if (popupReturn.getButtonClicked() >= 0):
			CvScreensInterface.getWorldBuilderScreen().handleAllPlotsCB(popupReturn)
		return

	def __eventWBLandmarkPopupBegin(self, argsList):
		CvScreensInterface.getWorldBuilderScreen().setLandmarkCB("")
		#popup = PyPopup.PyPopup(CvUtil.EventWBLandmarkPopup, EventContextTypes.EVENTCONTEXT_ALL)
		#popup.createEditBox(localText.getText("TXT_KEY_WB_LANDMARK_START", ()))
		#popup.launch()
		return

	def __eventWBLandmarkPopupApply(self, playerID, userData, popupReturn):
		if (popupReturn.getEditBoxString(0)):
			szLandmark = popupReturn.getEditBoxString(0)
			if (len(szLandmark)):
				CvScreensInterface.getWorldBuilderScreen().setLandmarkCB(szLandmark)
		return

	def __eventWBScriptPopupBegin(self, argsList):
		popup = PyPopup.PyPopup(CvUtil.EventWBScriptPopup, EventContextTypes.EVENTCONTEXT_ALL)
		popup.setHeaderString(localText.getText("TXT_KEY_WB_SCRIPT", ()))
		popup.createEditBox(CvScreensInterface.getWorldBuilderScreen().getCurrentScript())
		popup.launch()
		return

	def __eventWBScriptPopupApply(self, playerID, userData, popupReturn):
		if (popupReturn.getEditBoxString(0)):
			szScriptName = popupReturn.getEditBoxString(0)
			CvScreensInterface.getWorldBuilderScreen().setScriptCB(szScriptName)
		return

	def __eventWBStartYearPopupBegin(self, argsList):
		popup = PyPopup.PyPopup(CvUtil.EventWBStartYearPopup, EventContextTypes.EVENTCONTEXT_ALL)
		popup.createSpinBox(0, "", gc.getGame().getStartYear(), 1, 5000, -5000)
		popup.launch()
		return

	def __eventWBStartYearPopupApply(self, playerID, userData, popupReturn):
		iStartYear = popupReturn.getSpinnerWidgetValue(int(0))
		print iStartYear
		CvScreensInterface.getWorldBuilderScreen().setStartYearCB(iStartYear)
		return

	def __eventHasTraitBegin(self, argslist):
		return 0
	
	def __eventHasTraitApply(self, playerID, userData, popupReturn):
		iPlayer, iTrait, bHas = userData
		pPlayer = gc.getPlayer(iPlayer)
		if bHas == True:
			pPlayer.addTrait(iTrait)
		if bHas == False:
			pPlayer.removeTrait(iTrait)
		return 0

	def __eventChangePlayerBegin(self, argslist):
		return 0
	
	def __eventChangePlayerApply(self, playerID, userData, popupReturn):
		iPlayer, iCivilization, iLeader, iTeam, bIsHuman, bChangeGraphics = userData
		CyGame().changePlayer(iPlayer, iCivilization, iLeader, iTeam, bIsHuman, bChangeGraphics)
		return 0

	def __eventReassignPlayerBegin(self, argslist):
		return 0
	
	def __eventReassignPlayerApply(self, playerID, userData, popupReturn):
		iOldPlayer, iNewPlayer = userData
		CyGame().reassignPlayer(iOldPlayer, iNewPlayer)
		return 0

	def __eventCreatePlayerBegin(self, argslist):
		return 0
	
	def __eventCreatePlayerApply(self, playerID, userData, popupReturn):
		iPlayer, iCivType, iLeader, iTeam, bDoGive, iX, iY = userData
		CyGame().createNewPlayer(iPlayer, iCivType, iLeader, iTeam, bDoGive, iX, iY)
		return 0
